@@ -62,7 +62,7 @@
 
 /* Some constants */
 #define UDM_LANGPERDOC				16		/* FIXME */
-#define UDM_USER_AGENT				"MnogoSearch"
+#define UDM_USER_AGENT				"MnoGoSearch/" VERSION
 #define UDM_MAXWORDPERQUERY			32
 
 /* Some sizes and others definitions */
@@ -83,10 +83,6 @@
 /* storage types */
 #define UDM_DBMODE_SINGLE	0
 #define UDM_DBMODE_MULTI	1
-#define UDM_DBMODE_SINGLE_CRC	2
-#define UDM_DBMODE_WORD2URL	3
-#define UDM_DBMODE_MULTI_CRC	4
-#define UDM_DBMODE_CACHE	5
 #define UDM_DBMODE_BLOB		6
 
 
@@ -150,6 +146,8 @@
 #define UDM_METHOD_CHECKMP3	5
 #define UDM_METHOD_CHECKMP3ONLY	6
 #define UDM_METHOD_VISITLATER	7
+#define UDM_METHOD_INDEX	8
+#define UDM_METHOD_NOINDEX	9
 
 /* Tags/categories defines */
 #define UDM_CATSIZE	8
@@ -165,7 +163,7 @@
 #define UDM_LOCK_CONF		0
 #define UDM_LOCK_WIN		1
 #define UDM_LOCK_THREAD         2
-#define UDM_LOCK_STORED         3
+#define UDM_LOCK_SQL            3
 #define UDM_LOCK_SEGMENTER      4
 #define UDM_LOCK_DB             5
 #define UDM_LOCK_MAX		6
@@ -178,6 +176,7 @@ typedef struct stat_struct {
 } UDM_STAT;
 
 typedef struct stat_list_struct{
+	time_t		time;
 	size_t		nstats;
 	UDM_STAT	*Stat;
 } UDM_STATLIST;
@@ -200,33 +199,6 @@ typedef struct udm_varlist_st {
 } UDM_VARLIST;
 
 
-/************** Language and charset guesser *************/
-
-
-#define UDM_LM_MAXGRAM		6
-#define UDM_LM_HASHMASK		0x0FFF
-#define UDM_LM_TOPCNT           200
-
-typedef struct {
-	size_t		count, index;
-	char		str[UDM_LM_MAXGRAM+1];
-} UDM_LANGITEM;
-
-typedef struct {
-	float		expectation;			/**< Average value   */
-        int             needsave;
-	char		*lang;				/**< Map Language    */
-	char		*charset;			/**< Map charset     */
-        char            *filename;                      /**< Filename to write updates, if need */
-	UDM_LANGITEM	memb[UDM_LM_HASHMASK+1];	/**< Items list      */
-} UDM_LANGMAP;
-
-typedef struct {
-	size_t		nmaps;
-	UDM_LANGMAP	*Map;
-} UDM_LANGMAPLIST;
-
-/*****************************************************/
 
 typedef struct {
 	char		*str;
@@ -309,7 +281,11 @@ typedef struct {
 	size_t		nwords;	/**< Real number of words in list             */
 	size_t		swords;	/**< Number of words in sorted list           */
 	size_t		wordpos;/**< For phrases, number of current word      */
+#ifdef TRIAL_VER
+	UDM_WORD	Word[256];
+#else
 	UDM_WORD	*Word;	/**< Word list  itself                        */
+#endif
 } UDM_WORDLIST;
 
 
@@ -358,6 +334,7 @@ typedef struct {
 	int		match_type;
 	int		nomatch;
 	int		case_sense;
+	char		*section;
 	char		*pattern;
 	regex_t		*reg;
 	char		*arg;
@@ -622,6 +599,7 @@ typedef struct {
 	char		*word;
 	int		*uword;
 	size_t		len;
+	size_t		ulen;
         int     	origin;
 } UDM_WIDEWORD;
 
@@ -644,7 +622,7 @@ typedef struct {
 } UDM_SYNONYMLIST;
 
 typedef struct udm_chinaword_struct {
-  char *word;
+  int *word;
   int  freq;
 } UDM_CHINAWORD;
 
@@ -742,6 +720,7 @@ typedef struct udm_config_struct {
 	UDM_MATCHLIST	ReverseAliases;	/**< Reverse aliases            */
 	UDM_MATCHLIST	MimeTypes;	/**< For AddType commands       */
 	UDM_MATCHLIST	Filters;	/**< Allow, Disallow,etc        */
+	UDM_MATCHLIST	SectionFilters; /**< IndexIf, NoIndexIf         */
 	
 	UDM_RESULT	Targets;	/**< Targets cache              */
 	
@@ -765,7 +744,8 @@ typedef struct udm_config_struct {
 	int		is_log_open;	/**< if UdmOpenLog is already called   */
 	FILE		*logFD;		/**< FILE structure, syslog descriptor */
         int             logs_only;      /**< Cache mode writes mode            */
-	
+        int             do_store;       /**< Compressed copies storage flag    */
+	int		CVS_ignore;	/**< Skip CVS directgories - for tests */
 	void (*ThreadInfo)(struct udm_indexer_struct *,const char * state,const char * str);
 	void (*LockProc)(struct udm_indexer_struct *,int command,int type,const char *fname,int lineno);
 	void (*RefInfo)(int code,const char *url, const char *ref);
@@ -898,10 +878,6 @@ typedef struct {
         const char *b;
         const char **lt;
         const char *s;
-        int socket;
-        int chunks;
-        char *Content;
-        int finished;
 	size_t	ntoks;
 	struct {
 		const char *name;
@@ -926,12 +902,6 @@ typedef struct udm_cfg_st {
 #define udm_min(a,b) (((a) < (b)) ? (a) : (b))
 #endif
 
-#ifdef HAVE_MERGESORT
-#define UdmSort mergesort
-#else
-#define UdmSort qsort
-#endif
-
 #ifdef DMALLOC
 #include <dmalloc.h>
 #endif
@@ -941,4 +911,18 @@ typedef struct udm_cfg_st {
 #define UDM_DT_RANGE   3
 #define UDM_DT_UNKNOWN 4
 
+enum udm_indcmd
+{
+  UDM_IND_INDEX,
+  UDM_IND_STAT,
+  UDM_IND_CREATE,
+  UDM_IND_DROP,
+  UDM_IND_DELETE,
+  UDM_IND_REFERERS,
+  UDM_IND_SQLMON,
+  UDM_IND_CHECKCONF,
+  UDM_IND_CONVERT,
+  UDM_IND_MULTI2BLOB
+};
+
 #endif
