@@ -18,6 +18,8 @@
 #ifndef _UDM_UTILS_H
 #define _UDM_UTILS_H
 
+#include <stdio.h>
+
 /* for time_t */
 #include <time.h>
 
@@ -26,9 +28,9 @@
 
 #include <stdlib.h>
 #include <limits.h>
+#include <sys/types.h>
 
 #include "udm_config.h"
-#include "udm_common.h"
 
 /* This is used in UdmTime_t2Str and in its callers */
 #define UDM_MAXTIMESTRLEN	35
@@ -36,7 +38,7 @@
 
 /* Some useful MACROs */
 #define UDM_STREND(s)		(s+strlen(s))
-#define UDM_FREE(x)		{if((x)!=NULL){free(x);x=NULL;}}
+#define UDM_FREE(x)		{if((x)!=NULL){UdmFree(x);x=NULL;}}
 #define UDM_SKIP(s,set)		while((*s)&&(strchr(set,*s)))s++;
 #define UDM_SKIPN(s,set)	while((*s)&&(!strchr(set,*s)))s++;
 
@@ -98,6 +100,8 @@ extern int UdmInitTZ(void);
 extern unsigned long UdmStartTimer(void);
 
 
+extern char *UdmStrStore (char *dest, const char *src);
+
 /* Probably string missing functions */
 
 #ifndef HAVE_BZERO
@@ -130,20 +134,19 @@ extern char *udm_strtok_r(char *s, const char *delim, char **save_ptr);
 #define strtok_r udm_rstok_r
 #endif
 
+#if !defined(HAVE_STRNDUP) || defined(EFENCE)
+extern char *UdmStrndup(const char *str, size_t len);
+#endif
 #ifndef HAVE_STRNDUP
-extern char *strndup(const char *str, size_t len);
+#define strndup UdmStrndup
 #endif
 
-#ifndef HAVE_HSTRERROR
-extern const char *h_errlist[];
-# define hstrerror(err)  ((err) <= 4 ? h_errlist[(err)] : "unknown error")
-#endif
-
-extern int UdmHex2Int(int h);
+extern __C_LINK int __UDMCALL UdmHex2Int(int h);
+extern __C_LINK int __UDMCALL UdmInt2Hex(int i);
 
 
 #define BASE64_LEN(len) (4 * (((len) + 2) / 3) +2)
-extern __C_LINK void __UDMCALL udm_base64_encode (const char *s, char *store, size_t length);
+extern __C_LINK size_t __UDMCALL udm_base64_encode (const char *s, char *store, size_t length);
 extern __C_LINK size_t __UDMCALL udm_base64_decode (char * dst, const char * src, size_t len);
 extern char * udm_rfc1522_decode(char * dst, const char *src);
 
@@ -160,8 +163,6 @@ extern void UdmUniPrint(int * ustr);
 extern ssize_t UdmRecvall(int s, void *buf, size_t len);
 extern ssize_t UdmSend(int s, const void *msg, size_t len, int flags);
 
-extern void (*UdmSignal(int signo, void (*handler)(int)))(int);
-
 extern void UdmWriteLock(int fd);
 extern void UdmUnLock(int fd);
 extern void UdmReadLock(int fd);
@@ -169,4 +170,19 @@ extern void UdmReadLockFILE(FILE *f);
 extern __C_LINK void __UDMCALL UdmWriteLockFILE(FILE *f);
 extern __C_LINK void __UDMCALL UdmUnLockFILE(FILE *f);
 
+/************************ Dynamic strings stuff ***********/
+typedef struct dstr_struct {
+	size_t size_total; /* Bytes allocated */
+	size_t size_data;  /* Bytes used by data */
+	size_t size_page;  /* Bytes to allocate on overflow */
+	char free;         /* Need free flag */
+	char *data;        /* null-terminated string */
+} UDM_DSTR;
+
+extern UDM_DSTR *UdmDSTRInit (UDM_DSTR *dstr, size_t size_page);
+extern void UdmDSTRFree (UDM_DSTR *dstr);
+extern size_t UdmDSTRAppend (UDM_DSTR *dstr, const char *data, size_t size_data);
+extern size_t UdmDSTRAppendSTR (UDM_DSTR *dstr, const char *data);
+extern int UdmDSTRAppendf (UDM_DSTR *dstr, const char *fmt, ...);
+
 #endif /* _UDM_UTILS_H */
