@@ -31,10 +31,10 @@
 
 __C_LINK UDM_AGENT * __UDMCALL UdmAgentInit(UDM_AGENT *result,UDM_ENV * Env, int handle){
 #ifdef USE_TRACE
-  char filename[1024];
+  char filename[PATH_MAX];
 #endif
 	if(!result){
-		result=(UDM_AGENT*)malloc(sizeof(UDM_AGENT));
+		result=(UDM_AGENT*)UdmMalloc(sizeof(UDM_AGENT));
 		bzero((void*)result, sizeof(*result));
 		result->freeme=1;
 	}else{
@@ -44,10 +44,10 @@ __C_LINK UDM_AGENT * __UDMCALL UdmAgentInit(UDM_AGENT *result,UDM_ENV * Env, int
 	result->Conf=Env;
 	result->handle=handle; /* Handle is used in multi-threaded version */
 	result->action = UDM_OK;
-	result->LangMap = (UDM_LANGMAP*)malloc(sizeof(UDM_LANGMAP));
+	result->LangMap = (UDM_LANGMAP*)UdmMalloc(sizeof(UDM_LANGMAP));
 	bzero((void*)result->LangMap, sizeof(UDM_LANGMAP));
 #ifdef USE_TRACE
-	udm_snprintf(filename, 1024, "/tmp/udm_agent.%d.trace", handle);
+	udm_snprintf(filename, sizeof(filename), "/tmp/udm_agent.%d.trace", handle);
 	result->TR = fopen(filename, "w");
 #endif
 	return(result);
