@@ -103,11 +103,14 @@ const char *UdmMethodStr(int method){
 		case UDM_METHOD_HEAD:		return "CheckOnly";
 		case UDM_METHOD_HREFONLY:	return "HrefOnly";
 		case UDM_METHOD_VISITLATER:	return "Skip";
+		case UDM_METHOD_INDEX:		return "IndexIf";
+		case UDM_METHOD_NOINDEX:	return "NoIndexIf";
 	}
 	return "<Unknown method>";
 }
 
 int UdmMethod(const char *s){
+  if (s == NULL) return UDM_METHOD_UNKNOWN;
 	if(!strcasecmp(s,"Disallow"))		return UDM_METHOD_DISALLOW;
 	if(!strcasecmp(s,"Allow"))		return UDM_METHOD_GET;
 	if(!strcasecmp(s,"CheckMP3Only"))	return UDM_METHOD_CHECKMP3ONLY;
@@ -115,6 +118,8 @@ int UdmMethod(const char *s){
 	if(!strcasecmp(s,"CheckOnly"))		return UDM_METHOD_HEAD;
 	if(!strcasecmp(s,"HrefOnly"))		return UDM_METHOD_HREFONLY;
 	if(!strcasecmp(s,"Skip"))		return UDM_METHOD_VISITLATER;
+	if(!strcasecmp(s,"IndexIf"))		return UDM_METHOD_INDEX;
+	if(!strcasecmp(s,"NoIndexIf"))		return UDM_METHOD_NOINDEX;
 	return UDM_METHOD_UNKNOWN;
 }
 
@@ -141,8 +146,14 @@ void UdmWeightFactorsInit(const char *wf, int *res){
 static size_t rel_etc_name(UDM_ENV *Env,char *res,size_t maxlen,const char *name){
 	size_t		n;
 	const char	*dir=UdmVarListFindStr(&Env->Vars,"ConfDir",UDM_CONF_DIR);
-	if(name[0]=='/')n = udm_snprintf(res,maxlen,name);
-	else		n = udm_snprintf(res,maxlen,"%s%s%s",dir,UDMSLASHSTR,name);
+	if(name[0]=='/')
+		n = udm_snprintf(res,maxlen,name);
+	else
+#ifdef WIN32
+		n = udm_snprintf(res,maxlen,"%s", name);
+#else
+		n = udm_snprintf(res,maxlen,"%s%s%s",dir,UDMSLASHSTR,name);
+#endif
 	res[maxlen]='\0';
 	return n;
 }
@@ -221,7 +232,7 @@ static int add_srv(void *Cfg, size_t ac,char **av){
 		if(!strcasecmp(av[i],"regex"))C->Srv->Match.match_type=UDM_MATCH_REGEX;
 		else{
 			if(!C->Srv->Match.pattern)
-				C->Srv->Match.pattern = (char*)strdup(av[i]);
+				C->Srv->Match.pattern = (char*)UdmStrdup(av[i]);
 			else
 			if(!has_alias){
 				has_alias=1;
@@ -234,7 +245,7 @@ static int add_srv(void *Cfg, size_t ac,char **av){
 	}
 	if(UDM_OK != UdmServerAdd(Indexer, C->Srv)) {
 		char * s_err;
-		s_err = (char*)strdup(Conf->errstr);
+		s_err = (char*)UdmStrdup(Conf->errstr);
 		sprintf(Conf->errstr,"%s",s_err);
 		UDM_FREE(s_err);
 		UDM_FREE(C->Srv->Match.pattern);
@@ -289,7 +300,7 @@ static int add_alias(void *Cfg, size_t ac,char **av){
 			if(!strcasecmp(av[0],"Alias"))L=&Conf->Aliases;
 			if(!strcasecmp(av[0],"ReverseAlias"))L=&Conf->ReverseAliases;
 			
-			if(UDM_OK!=UdmMatchListAdd(L,&Alias,err,sizeof(err))){
+			if(UDM_OK != UdmMatchListAdd(NULL, L, &Alias, err, sizeof(err), 0)) {
 				udm_snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"%s",err);
 				return UDM_ERROR;
 			}
@@ -309,6 +320,9 @@ static int add_filter(void *Cfg, size_t ac,char **av){
 	UDM_MATCH	M;
 	size_t		i;
 	
+	if(!(C->flags & UDM_FLAG_ADD_SERV))
+		return UDM_OK;
+	
 	UdmMatchInit(&M);
 	M.match_type=UDM_MATCH_WILD;
 	M.case_sense=1;
@@ -331,10 +345,10 @@ static int add_filter(void *Cfg, size_t ac,char **av){
 		else{
 			char		err[120]="";
 			
-			M.arg=av[0];
-			M.pattern=av[i];
+			M.arg = av[0];
+			M.pattern = av[i];
 			
-			if(UDM_OK!=UdmMatchListAdd(&Conf->Filters,&M,err,sizeof(err))){
+			if(UDM_OK != UdmMatchListAdd(C->Indexer, &Conf->Filters, &M, err, sizeof(err), ++C->ordre)) {
 				udm_snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"%s",err);
 				return UDM_ERROR;
 			}
@@ -343,6 +357,68 @@ static int add_filter(void *Cfg, size_t ac,char **av){
 	return UDM_OK;
 }
 
+static int add_section_filter(void *Cfg, size_t ac,char **av){
+	UDM_CFG		*C=(UDM_CFG*)Cfg;
+	UDM_ENV		*Conf=C->Indexer->Conf;
+	UDM_MATCH	M;
+	size_t		i;
+	char		ss = 0;
+	
+	if(!(C->flags & UDM_FLAG_ADD_SERV))
+		return UDM_OK;
+	
+	UdmMatchInit(&M);
+	M.match_type=UDM_MATCH_WILD;
+	M.case_sense=1;
+	
+	C->ordre++;
+	for(i=1; i<ac ; i++){
+		if(!strcasecmp(av[i],"case"))M.case_sense=1;
+		else
+		if(!strcasecmp(av[i],"nocase"))M.case_sense=0;
+		else
+		if(!strcasecmp(av[i],"regex"))M.match_type=UDM_MATCH_REGEX;
+		else
+		if(!strcasecmp(av[i],"regexp"))M.match_type=UDM_MATCH_REGEX;
+		else
+		if(!strcasecmp(av[i],"string"))M.match_type=UDM_MATCH_WILD;
+		else
+		if(!strcasecmp(av[i],"nomatch"))M.nomatch=1;
+		else
+		if(!strcasecmp(av[i],"match"))M.nomatch=0;
+		else
+		if (! ss) {
+/*
+			UDM_VAR *Sec = UdmVarListFind(&Conf->Sections, av[i]);
+			if (! Sec) {
+				udm_snprintf(Conf->errstr, sizeof(Conf->errstr) - 1, "Section '%s' not found for %s", av[i], av[0]);
+				return(UDM_ERROR);
+			}
+			ss = 1;
+			M.section = Sec->name;
+*/
+			ss = 1;
+			M.section = av[i];
+		}else{
+			char		err[120]="";
+
+			M.arg = av[0];
+			M.pattern = av[i];
+
+			if(UDM_OK != UdmMatchListAdd(C->Indexer, &Conf->SectionFilters, &M, err, sizeof(err), ++C->ordre)) {
+				udm_snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"%s",err);
+				return UDM_ERROR;
+			}
+		}
+	}
+
+	if (! ss) {
+		udm_snprintf(Conf->errstr, sizeof(Conf->errstr) - 1, "No section given for %s", av[0]);
+		return(UDM_ERROR);
+	}
+	return UDM_OK;
+}
+
 static int add_type(void *Cfg, size_t ac,char **av){
 	UDM_CFG		*C=(UDM_CFG*)Cfg;
 	UDM_ENV		*Conf=C->Indexer->Conf;
@@ -369,7 +445,7 @@ static int add_type(void *Cfg, size_t ac,char **av){
 			M.arg=av[i];
 		else{
 			M.pattern=av[i];
-			if(UDM_OK != (rc=UdmMatchListAdd(&Conf->MimeTypes,&M,err,sizeof(err)))){
+			if(UDM_OK != (rc = UdmMatchListAdd(NULL, &Conf->MimeTypes,&M,err,sizeof(err), 0))) {
 				udm_snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"%s",err);
 				return rc;
 			}
@@ -394,11 +470,20 @@ static int add_section(void *Cfg, size_t ac,char **av){
 	UDM_CFG	*C=(UDM_CFG*)Cfg;
 	UDM_ENV	*Conf=C->Indexer->Conf;
 	UDM_VAR S;
+
 	bzero((void*)&S, sizeof(S));
 	S.name=av[1];
 	S.section=atoi(av[2]);
 	S.maxlen=av[3]?atoi(av[3]):0;
+
+	if (ac == 5 && av[4]) {
+		char *str = UdmStrStore(NULL, "separator.");
+		str = UdmStrStore(str, av[1]);
+		UdmVarListAddStr(&Conf->Vars, str, av[4]);
+		UDM_FREE(str);
+	}
 	UdmVarListReplace(&Conf->Sections,&S);
+
 	return UDM_OK;
 }
 
@@ -467,7 +552,7 @@ static int add_langmap(void *Cfg, size_t ac,char **av){
 	if(C->flags&UDM_FLAG_LOAD_LANGMAP){
 		char	fname[1024];
 		rel_etc_name(Conf,fname,sizeof(fname)-1,av[1]);
-		res=UdmLoadLangMapFile(Conf,fname);
+		res = UdmLoadLangMapFile(&Conf->LangMaps, fname);
 	}
 	return res;
 }
@@ -485,20 +570,20 @@ static int add_synonym(void *Cfg, size_t ac,char **av){
 }
 
 static int add_chinese(void *Cfg, size_t ac,char **av){
-#ifdef HAVE_CHARSET_gb2312
 	UDM_CFG	*C=(UDM_CFG*)Cfg;
 	UDM_ENV	*Conf=C->Indexer->Conf;
 	if(C->flags & UDM_FLAG_ADD_SERV){
-		return UdmChineseListLoad(&Conf->Chi, (av[1] != NULL) ? av[1] : UDM_CONF_DIR "/mandarin.freq" );
+	  return UdmChineseListLoad(C->Indexer, &Conf->Chi, (av[1] != NULL) ? av[1] : "GB2312",
+				    (av[2] != NULL) ? av[2] : UDM_CONF_DIR "/mandarin.freq" );
 	}
-#endif
 	return UDM_OK;
 }
 static int add_thai(void *Cfg, size_t ac,char **av){
 	UDM_CFG	*C=(UDM_CFG*)Cfg;
 	UDM_ENV	*Conf=C->Indexer->Conf;
 	if(C->flags & UDM_FLAG_ADD_SERV){
-		return UdmChineseListLoad(&Conf->Thai, (av[1] != NULL) ? av[1] : UDM_CONF_DIR "/thai.freq" );
+	  return UdmChineseListLoad(C->Indexer, &Conf->Thai, (av[1] != NULL) ? av[1] : "tis-620",
+				    (av[2] != NULL) ? av[2] : UDM_CONF_DIR "/thai.freq" );
 	}
 	return UDM_OK;
 }
@@ -529,24 +614,20 @@ static int add_srv_table(void *Cfg, size_t ac,char **av){
 	UDM_CFG	*C=(UDM_CFG*)Cfg;
 	UDM_ENV	*Conf=C->Indexer->Conf;
 	int		res = UDM_OK;
-	UDM_AGENT	Agent;
 	UDM_DBLIST      dbl;
 	UDM_DB		*db;
 	
 	UdmDBListInit(&dbl);
 	UdmDBListAdd(&dbl, av[1], UDM_OPEN_MODE_READ);
 	db = &dbl.db[0];
-	UdmAgentInit(&Agent,Conf,0);
-	Agent.flags = C->flags;
 
 #ifdef HAVE_SQL
-	res = UdmSrvActionSQL(&Agent, &Conf->Servers, UDM_SRV_ACTION_TABLE, db);
+	res = UdmSrvActionSQL(C->Indexer, &Conf->Servers, UDM_SRV_ACTION_TABLE, db);
 #endif
 	if(res != UDM_OK){
 	  strcpy(Conf->errstr,db->errstr);
 	}
 
-	UdmAgentFree(&Agent);
 	UdmDBListFree(&dbl);
 	return res;
 }
@@ -560,7 +641,7 @@ static int add_limit(void *Cfg, size_t ac,char **av){
 	
 	if((sc = strchr(av[1],':'))){
 		*sc++='\0';
-		nm=(char*)malloc(strlen(av[1])+8);
+		nm=(char*)UdmMalloc(strlen(av[1])+8);
 		sprintf(nm,"Limit-%s",av[1]);
 		UdmVarListReplaceStr(&Conf->Vars, nm, sc);
 		UDM_FREE(nm);
@@ -634,7 +715,7 @@ static int srv_rpl_auth(void *Cfg, size_t ac,char **av){
 	name[sizeof(name)-1]='\0';
 	if(av[1]){
 		size_t	len=strlen(av[1]);
-		char	*auth=(char*)malloc(BASE64_LEN(strlen(av[1])));
+		char	*auth=(char*)UdmMalloc(BASE64_LEN(strlen(av[1])));
 		udm_base64_encode(av[1],auth,len);
 		UdmVarListReplaceStr(&C->Srv->Vars,name,auth);
 		UDM_FREE(auth);
@@ -644,21 +725,11 @@ static int srv_rpl_auth(void *Cfg, size_t ac,char **av){
 	return UDM_OK;
 }
 
-static char *str_store (char *dest, char *src) {
-	size_t d_size = (dest ? strlen(dest) : 0);
-	size_t s_size = strlen(src) + 1;
-	char *d = realloc(dest, d_size + s_size);
-
-	if (d) memcpy(d + d_size, src, s_size);
-	else UDM_FREE(dest);
-	return(d);
-}
-
-static char *ParseEnvVar (UDM_ENV *Conf, const char *str) {
+char *UdmParseEnvVar (UDM_ENV *Conf, const char *str) {
 	char *p1 = (char *)str;
 	char *p2 = (char *)str;
 	char *p3;
-	char *s;
+	const char *s;
 	char *o = NULL;
 
 	if (! str) return(NULL);
@@ -668,12 +739,12 @@ static char *ParseEnvVar (UDM_ENV *Conf, const char *str) {
 			continue;
 		}
 		*p1 = 0;
-		o = str_store(o, p2);
+		o = UdmStrStore(o, p2);
 		*p1 = '$';
 		if ((p3 = strchr(p1 + 2, ')'))) {
 			*p3 = 0;
-			s = (char *)UdmVarListFindStr(&Conf->Vars, p1 + 2, NULL);
-			if (s) o = str_store(o, s);
+			s = UdmVarListFindStr(&Conf->Vars, p1 + 2, NULL);
+			if (s) o = UdmStrStore(o, s);
 			*p3 = ')';
 			p2 = p1 = p3 + 1;
 		} else {
@@ -681,7 +752,7 @@ static char *ParseEnvVar (UDM_ENV *Conf, const char *str) {
 			return(NULL);
 		}
 	}
-	o = str_store(o, p2);
+	o = UdmStrStore(o, p2);
 	return(o);
 }
 
@@ -752,6 +823,8 @@ static int env_rpl_bool_var(void *Cfg, size_t ac,char **av){
 	UDM_ENV	*Conf=C->Indexer->Conf;
 	int res=!strcasecmp(av[1],"yes");
 	if(!strcasecmp(av[0],"LogsOnly"))Conf->logs_only=res;
+	else if(!strcasecmp(av[0], "DoStore")) Conf->do_store = res;
+	else if(!strcasecmp(av[0], "CVSIgnore")) Conf->CVS_ignore = res;
 	return UDM_OK;
 }
 
@@ -823,6 +896,7 @@ static UDM_CONFCMD commands[] =
 	{"ForceIISCharset1251",	1,1,	env_rpl_var},
 	{"GroupBySite",		1,1,	env_rpl_var},
 	{"Cache",		1,1,	env_rpl_var},
+	{"wf",   		1,1,	env_rpl_var},
 	{"PopRankSkipSameSite",	1,1,	env_rpl_var},
 	{"PopRankFeedBack",	1,1,	env_rpl_var},
 	{"PopRankUseTracking",	1,1,	env_rpl_var},
@@ -835,6 +909,9 @@ static UDM_CONFCMD commands[] =
 	{"URLSelectCacheSize",	1,1,	env_rpl_var},
 	{"MaxDocSize",		1,1,	env_rpl_var},
 	{"StoredFiles",		1,1,	env_rpl_var},
+	{"URLDataFiles",	1,1,	env_rpl_var},
+	{"WrdFiles",		1,1,	env_rpl_var},
+	{"BaseFiles",		1,1,	env_rpl_var},
 	{"OptimizeInterval",	1,1,	env_rpl_var},
 	{"OptimizeRatio",	1,1,	env_rpl_var},
 	{"ParserTimeOut",	1,1,	env_rpl_var},
@@ -849,6 +926,8 @@ static UDM_CONFCMD commands[] =
 	{"MinWordLength",	1,1,	env_rpl_num_var},
 	{"MaxWordLength",	1,1,	env_rpl_num_var},
 	{"LogsOnly",		1,1,	env_rpl_bool_var},
+	{"DoStore",		1,1,	env_rpl_bool_var},
+	{"CVSIgnore",		1,1,	env_rpl_bool_var},
 	{"LocalCharset",	1,1,	env_rpl_charset},
 	{"BrowserCharset",	1,1,	env_rpl_charset},
 	
@@ -867,6 +946,7 @@ static UDM_CONFCMD commands[] =
 	{"Robots",		1,1,	srv_rpl_bool_var},
 	{"DetectClones",	1,1,	srv_rpl_bool_var},
 	{"Index",		1,1,	srv_rpl_bool_var},
+	{"FollowSymLinks",	1,1,	srv_rpl_bool_var},
 	{"NetErrorDelayTime",	1,1,	srv_rpl_time_var},
 	{"ReadTimeOut",		1,1,	srv_rpl_time_var},
 	{"DocTimeOut",		1,1,	srv_rpl_time_var},
@@ -886,6 +966,9 @@ static UDM_CONFCMD commands[] =
 	{"CheckMP3",		1,100,	add_filter},
 	{"CheckOnly",		1,100,	add_filter},
 	{"HrefOnly",		1,100,	add_filter},
+
+	{"IndexIf",		1,100,	add_section_filter},
+	{"NoIndexIf",		1,100,	add_section_filter},
 	
 	{"Server",		1,100,	add_srv},
 	{"Realm",		1,100,	add_srv},
@@ -897,15 +980,15 @@ static UDM_CONFCMD commands[] =
 	
 	{"AddType",		1,100,	add_type},
 	{"Mime",		2,3,	add_parser},
-	{"Section",		2,3,	add_section},
+	{"Section",		2,4,	add_section},
 	{"Affix",		3,3,	add_affix},
 	{"Spell",		3,3,	add_spell},
 	{"StopwordFile",	1,1,	add_stoplist},
 	{"LangMapFile",		1,1,	add_langmap},
 	{"LangMapUpdate",	1,1,	env_rpl_var},
 	{"Synonym",		1,1,	add_synonym},
-	{"LoadChineseList",	0,1,	add_chinese},
-	{"LoadThaiList",        0,1,    add_thai},
+	{"LoadChineseList",	0,2,	add_chinese},
+	{"LoadThaiList",        0,2,    add_thai},
 	{"Limit",		1,1,	add_limit},
 	{"ServerTable",		1,1,	add_srv_table},
 	{"FlushServerTable",	0,0,	flush_srv_table},
@@ -941,7 +1024,7 @@ __C_LINK int __UDMCALL UdmEnvAddLine(UDM_CFG *C,char *str){
 			
 			for (i = 1; i < ac; i++) {
 				if (! av[i]) continue;
-				p = ParseEnvVar(Conf, av[i]);
+				p = UdmParseEnvVar(Conf, av[i]);
 				if (! p) {
 					sprintf(Conf->errstr, "An error occured while parsing '%s'", av[i]);
 					return UDM_ERROR;
@@ -972,7 +1055,7 @@ static int EnvLoad(UDM_CFG *Cfg,const char *cname){
 	int	rc=UDM_OK;
 	size_t	line = 0, str0len = 0, str1len, str0size = 4096;
 	
-	if ((str0 = (char*)malloc(str0size)) == NULL) {
+	if ((str0 = (char*)UdmMalloc(str0size)) == NULL) {
 		sprintf(Cfg->Indexer->Conf->errstr, "Can't alloc %d bytes at '%s': %d", str0size, __FILE__, __LINE__);
 		return UDM_ERROR;
 	}
@@ -999,7 +1082,7 @@ static int EnvLoad(UDM_CFG *Cfg,const char *cname){
 			*end=0;
 			if (str0len + str1len >= str0size) {
 			  str0size += 4096 + str1len;
-			  if ((str0 = (char*)realloc(str0, str0size)) == NULL) {
+			  if ((str0 = (char*)UdmRealloc(str0, str0size)) == NULL) {
 			    sprintf(Cfg->Indexer->Conf->errstr, "Can't realloc %d bytes at '%s': %d", str0size, __FILE__, __LINE__);
 			    return UDM_ERROR;
 			  }
@@ -1059,7 +1142,7 @@ __C_LINK  int __UDMCALL UdmEnvLoad(UDM_AGENT *Indexer, const char *cname, int lf
 		/* Sort synonyms */
 		UdmSynonymListSort(&Indexer->Conf->Synonyms);
 		
-		UdmVarListInsStr(&Indexer->Conf->Vars, "Request.User-Agent", UDM_USER_AGENT "/" VERSION);
+		UdmVarListInsStr(&Indexer->Conf->Vars, "Request.User-Agent", UDM_USER_AGENT);
 	}
 
 freeex:
