@@ -70,10 +70,8 @@
 #include "udm_host.h"
 #include "udm_hash.h"
 #include "udm_doc.h"
-#include "udm_cache.h"
 #include "udm_services.h"
 #include "udm_xmalloc.h"
-#include "udm_worddb.h"
 #include "udm_searchcache.h"
 
 #define UDM_THREADINFO(A,s,m)	if(A->Conf->ThreadInfo)A->Conf->ThreadInfo(A,s,m)
@@ -84,17 +82,19 @@ void *UdmDBInit(void *vdb){
 	size_t	nbytes=sizeof(UDM_DB);
 	
 	if(!db){
-		db=(UDM_DB*)malloc(nbytes);
+		db=(UDM_DB*)UdmMalloc(nbytes);
 		bzero((void*)db, nbytes);
 		db->freeme=1;
 	}else{
 		bzero((void*)db, nbytes);
 	}
-	db->numtables=32;
+	db->numtables=1;
 
-	UdmURLInit(&db->addr);
+#if HAVE_SQL
+	UdmMultiCacheInit(&db->MultiCache);
+#endif
 	
-#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB || HAVE_DB2)
+#if (HAVE_ODBC)
 	db->hDbc=SQL_NULL_HDBC;
 	db->hEnv=SQL_NULL_HENV;
 	db->hstmt=SQL_NULL_HSTMT;
@@ -111,84 +111,18 @@ void *UdmDBInit(void *vdb){
 
 void UdmDBFree(void *vdb){
 	UDM_DB	*db=vdb;
-	
-	UdmURLFree(&db->addr);
-	UDM_FREE(db->DBADDR);
+
 	UDM_FREE(db->DBName);
-	UDM_FREE(db->DBUser);
-	UDM_FREE(db->DBPass);
-	UDM_FREE(db->DBSock);
 	UDM_FREE(db->where);
 	UDM_FREE(db->from);
-
-	if (db->cached_sd) {
-	  shutdown(db->cached_sd, 2);
-	  closesocket(db->cached_sd);
-	}
-	if (db->stored_sd) { 
-	  const char *hello = "B\0";
-	  UdmSend(db->stored_sd, hello, 1, 0);
-	  shutdown(db->stored_sd, 2);
-	  closesocket(db->stored_sd);
-	}
+	
 	if (db->searchd) UdmSearchdClose(db);
 	
-	if(!db->connected)goto ret;
-	
-#if HAVE_MYSQL
-	if(db->DBDriver==UDM_DB_MYSQL){
-		UdmSQLClose(db);
-		goto ret;
-	}
+#if HAVE_SQL
+	if (db->connected) UdmSQLClose(db);
 #endif
 	
-#if HAVE_PGSQL
-	if(db->DBDriver==UDM_DB_PGSQL){
-		UdmSQLClose(db);
-		goto ret;
-	}
-#endif
-	
-#if HAVE_MSQL
-	if(db->DBDriver==UDM_DB_MSQL){
-		UdmSQLClose(db);
-		goto ret;
-	}
-#endif
-	
-#if HAVE_IBASE
-	if(db->DBDriver==UDM_DB_IBASE){
-		UdmSQLClose(db);
-		goto ret;
-	}
-#endif
-
-#if HAVE_ORACLE8
-	if(db->DBDriver==UDM_DB_ORACLE8){
-		UdmSQLClose(db);
-		goto ret;
-	}
-#endif
-	
-#if HAVE_ORACLE7
-	if(db->DBDriver==UDM_DB_ORACLE7){
-		UdmSQLClose(db);
-		goto ret;
-	}
-#endif
-	
-#if HAVE_CTLIB
-	if(db->DBDriver==UDM_DB_MSSQL){
-		UdmSQLClose(db);
-		goto ret;
-	}
-#endif
-	
-#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB || HAVE_DB2)
-	UdmSQLClose(db);
-	goto ret;
-#endif
-ret:
+	UdmVarListFree(&db->Vars);
 	if(db->freeme)UDM_FREE(vdb);
 	return;
 }
@@ -466,49 +400,56 @@ static int UdmDocUpdate(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc){
 	UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
 	maxsize = UdmVarListFindInt(&Indexer->Conf->Vars,"DocMemCacheSize",0) * 1024 * 1024;
 	UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
-	if(Doc){
-		/* Add document into cache */
-		I->Doc=(UDM_DOCUMENT*)realloc(I->Doc,(I->num_rows+1)*sizeof(UDM_DOCUMENT));
-		I->Doc[I->num_rows]=Doc[0];
-		I->num_rows++;
-		
-		I->memused+=sizeof(UDM_DOCUMENT);
+
+	if (maxsize > 0 && I->memused > 0) UdmLog(Indexer, UDM_LOG_EXTRA, "DocCacheSize: %d/%d", I->memused, maxsize);
+	if (Doc) {
+		I->memused += sizeof(UDM_DOCUMENT);
 		/* Aproximation for Words memory usage  */
-		I->memused+=Doc->Words.nwords*(sizeof(UDM_WORD)+10);
+		I->memused += Doc->Words.nwords * (sizeof(UDM_WORD) + 5);
 		/* Aproximation for CrossWords memory usage */
-		I->memused+=Doc->CrossWords.ncrosswords*(sizeof(UDM_CROSSWORD)+10+20);
+		I->memused += Doc->CrossWords.ncrosswords * (sizeof(UDM_CROSSWORD) + 35);
 		/* Aproximation for Sections memory usage */
-		for(sec=0;sec<Doc->Sections.nvars;sec++){
-			I->memused+=sizeof(UDM_VAR);
-			I->memused+=Doc->Sections.Var[sec].maxlen*9;
+		for(sec = 0; sec < Doc->Sections.nvars; sec++) {
+			I->memused += sizeof(UDM_VAR);
+			I->memused += Doc->Sections.Var[sec].maxlen * 3 + 10;
 		}
-	}
-	
-	UdmLog(Indexer,UDM_LOG_EXTRA,"DocCacheSize: %d/%d",I->memused,maxsize);
-	if(I->memused>=maxsize)flush=1;
-	if(I->num_rows>=1024)flush=1;
-	if(!Doc)flush=1;
-	
-	if(flush){
+		I->memused += (sizeof(UDM_HREF) + 35) * Doc->Hrefs.nhrefs;
+		if (I->memused >= maxsize) flush = 1;
+		if (I->num_rows >= 1024) flush = 1;
+	} else flush = 1;
+
+	if (flush) {
 		size_t	docnum;
+
+		if (I->num_rows)
+			UdmLog(Indexer, UDM_LOG_EXTRA, "Flush %d document(s)", I->num_rows + ((Doc != NULL) ? 1 : 0));
 		
-		if (Indexer->Indexed.num_rows > 1)
-			UdmLog(Indexer, UDM_LOG_EXTRA, "Flush %d document(s)", Indexer->Indexed.num_rows);
+		if (Doc) {
+			UDM_THREADINFO(Indexer, "Updating", UdmVarListFindStr(&Doc->Sections, "URL", ""));
+			if(UDM_OK != (rc = DocUpdate(Indexer, Doc))) return rc;
+			UdmDocFree(Doc);
+		}
 		
-		for (docnum=0;docnum<I->num_rows;docnum++){
+		for (docnum = 0; docnum < I->num_rows; docnum++) {
 			/* Flush all hrefs from cache in-memory    */
 			/* cache into database. Note, this must    */
 			/* be done before call of  StoreCrossWords */
 			/* because we need to know all new URL IDs */
 			
-			UDM_THREADINFO(Indexer,"Updating",UdmVarListFindStr(&I->Doc[docnum].Sections,"URL",""));
+			UDM_THREADINFO(Indexer, "Updating", UdmVarListFindStr(&I->Doc[docnum].Sections, "URL", ""));
 			if(UDM_OK != (rc = DocUpdate(Indexer, &I->Doc[docnum])))
 				return rc;
 		}
 /* this should be only if we use blob mode, otherwise built-in wan't work
 		if(UDM_OK != (rc = UdmResAction(Indexer, &Indexer->Indexed, UDM_RES_ACTION_INSWORDS)))
 			return rc;*/
-		UdmResultFree(&Indexer->Indexed);
+		if (Indexer->Indexed.num_rows) UdmResultFree(&Indexer->Indexed);
+	} else {
+		/* Add document into cache */
+		I->Doc=(UDM_DOCUMENT*)UdmRealloc(I->Doc, (I->num_rows + 1) * sizeof(UDM_DOCUMENT));
+		I->Doc[I->num_rows] = Doc[0];
+		I->Doc[I->num_rows].freeme = 0;
+		I->num_rows++;
 	}
 	return rc;
 }
@@ -544,10 +485,6 @@ __C_LINK int __UDMCALL UdmURLAction(UDM_AGENT *A, UDM_DOCUMENT *D, int cmd) {
 	    res = UdmSearchdURLAction(A, D, cmd, db);
 	    execflag = 1;
 	    break;
-	  case UDM_DB_CACHE:
-	    res = UdmURLActionCache(A, D, cmd, db);
-	    execflag = 1;
-	    break;
 #ifdef HAVE_SQL
 	  default:
 		res=UdmURLActionSQL(A,D,cmd,db);
@@ -678,7 +615,6 @@ static const int search_cache_size=1000;
 
 int UdmFindWords(UDM_AGENT *A, UDM_RESULT *Res) {
 	const char      *cache_mode  = UdmVarListFindStr(&A->Conf->Vars, "Cache", "no");
-	UDM_DB		*db;
 	size_t i, dbfrom = 0, dbto =  A->Conf->dbl.nitems;
 	int res = UDM_OK;
 	size_t		nwrd=0;
@@ -695,9 +631,12 @@ int UdmFindWords(UDM_AGENT *A, UDM_RESULT *Res) {
 		/* Let's do actual search now:        */
 		/* Get groupped by url_id words array */
 	
-	  for (i = dbfrom; i < dbto; i++) {
-	    db = &A->Conf->dbl.db[i];
-	    UdmLog(A, UDM_LOG_DEBUG, "UdmFind for %s", db->DBADDR);
+	  for (i = dbfrom; i < dbto; i++)
+	  {
+	    UDM_DB *db = &A->Conf->dbl.db[i];
+	    const char *dbaddr= UdmVarListFindStr(&db->Vars,"DBAddr","<noaddr>");
+	    
+	    UdmLog(A, UDM_LOG_DEBUG, "UdmFind for %s", dbaddr);
 	    Res->CoordList.Coords = NULL;
 	    Res->CoordList.Data = NULL;
 	    Res->CoordList.ncoords = 0;
@@ -708,14 +647,8 @@ int UdmFindWords(UDM_AGENT *A, UDM_RESULT *Res) {
 			break;
 #ifdef HAVE_SQL
 		default:
-		  if (db->DBMode == UDM_DBMODE_BLOB) {
-		    res = UdmFindWordsBlob(A, Res, db);
-		  } else if (db->DBMode == UDM_DBMODE_CACHE) {
-		    res = UdmFindWordsCache(A, Res, db);
-		  } else {
 		    res = UdmFindWordsSQL(A, Res, db);
-		  }
-			break;
+		    break;
 #endif
 	    }
 	    wrdX[i] = Res->CoordList.Coords;
@@ -728,9 +661,9 @@ int UdmFindWords(UDM_AGENT *A, UDM_RESULT *Res) {
 	  }
 
 	  if (nwrd > 0) {
-	    curwrd=wrd=(UDM_URL_CRD*)malloc(sizeof(*wrd)*nwrd);
-	    curudt=udt=(UDM_URLDATA*)malloc(sizeof(*udt)*nwrd);
-	    Res->PerSite = persite = curpersite = (size_t*)malloc(sizeof(size_t) * nwrd);
+	    curwrd=wrd=(UDM_URL_CRD*)UdmMalloc(sizeof(*wrd)*nwrd);
+	    curudt=udt=(UDM_URLDATA*)UdmMalloc(sizeof(*udt)*nwrd);
+	    Res->PerSite = persite = curpersite = (size_t*)UdmMalloc(sizeof(size_t) * nwrd);
 	    for (i = dbfrom; i < dbto; i++) {
 		if(wrdX[i]){
 /*			size_t j;*/
@@ -795,7 +728,7 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A) {
 	unsigned long	ticks=UdmStartTimer(), ticks_;
 	size_t i, dbfrom = 0, dbto =  A->Conf->dbl.nitems;
 	int		page_number = UdmVarListFindInt(&A->Conf->Vars, "np", 0);
-	int		page_size   = UdmVarListFindInt(&A->Conf->Vars, "ps", 20);
+	int		page_size   = UdmVarListFindInt(&A->Conf->Vars, "ps", 10);
 	char		str[128];
 	
 	UdmLog(A,UDM_LOG_DEBUG,"Start UdmFind");
@@ -821,7 +754,7 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A) {
 	Res->last=Res->first+Res->num_rows-1;
 
 	/* Allocate an array for documents information */
-	if (Res->num_rows > 0) Res->Doc = (UDM_DOCUMENT*)malloc(sizeof(UDM_DOCUMENT) * (Res->num_rows));
+	if (Res->num_rows > 0) Res->Doc = (UDM_DOCUMENT*)UdmMalloc(sizeof(UDM_DOCUMENT) * (Res->num_rows));
 	
 	/* Copy url_id and coord to result */
 	for(i=0;i<Res->num_rows;i++){
@@ -832,6 +765,10 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A) {
 		UdmVarListReplaceStr(&Res->Doc[i].Sections, "Score", str);
 		UdmVarListReplaceInt(&Res->Doc[i].Sections,"Order",(int)(i + Res->first + 1));
 		UdmVarListReplaceInt(&Res->Doc[i].Sections, "sdnum", (int)(score & 0xFF));
+		if (Res->PerSite)
+		{
+		  UdmVarListReplaceUnsigned(&Res->Doc[i].Sections, "PerSite", Res->PerSite[i + Res->first]);
+		}
 	}
 	
 	for (i = dbfrom; i < dbto; i++) {
@@ -856,7 +793,7 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A) {
 		for(i=0;i<num;i++){
 			UDM_RESULT *Cl = UdmCloneList(A, &Res->Doc[i]);
 			if(Cl){
-				Res->Doc=(UDM_DOCUMENT*)realloc(Res->Doc,sizeof(UDM_DOCUMENT)*(Res->num_rows+Cl->num_rows));
+				Res->Doc=(UDM_DOCUMENT*)UdmRealloc(Res->Doc,sizeof(UDM_DOCUMENT)*(Res->num_rows+Cl->num_rows));
 				memcpy(&Res->Doc[Res->num_rows],Cl->Doc,sizeof(UDM_DOCUMENT)*Cl->num_rows);
 				Res->num_rows+=Cl->num_rows;
 				UDM_FREE(Cl->Doc);
@@ -865,7 +802,10 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A) {
 		}
 	}
 	ticks_ = UdmStartTimer() - ticks_;
-	UdmLog(A, UDM_LOG_DEBUG, "Stop  Clones: %.2f", (float)ticks_/1000);
+	UdmLog(A, UDM_LOG_DEBUG, "Stop  Clones:\t\t%.2f", (float)ticks_/1000);
+
+	ticks_=UdmStartTimer();
+	UdmLog(A, UDM_LOG_DEBUG, "Start adding Order");
 
 	/* first and last begins from 0, make it begin from 1 */
 	Res->first++;
@@ -873,15 +813,22 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A) {
 	for(i=0;i<Res->num_rows;i++){
 		UdmVarListReplaceInt(&Res->Doc[i].Sections,"Order",(int)(Res->first+i));
 	}
-/******************************/
+	ticks_ = UdmStartTimer() - ticks_;
+	UdmLog(A, UDM_LOG_DEBUG, "Stop  Order:\t\t\t%.2f", (float)ticks_/1000);
 
+	ticks_=UdmStartTimer();
+	UdmLog(A, UDM_LOG_DEBUG, "Start UdmConvert");
 	UdmConvert(A->Conf, Res, A->Conf->lcs, A->Conf->bcs);
+	ticks_ = UdmStartTimer() - ticks_;
+	UdmLog(A, UDM_LOG_DEBUG, "Stop  UdmConvert:\t\t%.2f", (float)ticks_/1000);
+
 	Res->work_time=ticks=UdmStartTimer()-ticks;
 	UdmLog(A,UDM_LOG_DEBUG,"Done  UdmFind %.2f",(float)ticks/1000);
 
 	UdmTrack(A, Res);
 	
-	if(res!=UDM_OK){
+	if(res!=UDM_OK)
+	{
 		UdmResultFree(Res);
 		Res=NULL;
 	}
@@ -891,124 +838,186 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A) {
 
 static int UdmStr2DBMode(const char * str1){
 	int m = -1;
-	if(!strncasecmp(str1,"multi-crc",9))m=UDM_DBMODE_MULTI_CRC;
-	else
-	if(!strncasecmp(str1,"crc-multi",9))m=UDM_DBMODE_MULTI_CRC;
-	else
 	if(!strncasecmp(str1,"single",6))m=UDM_DBMODE_SINGLE;
 	else
-	if(!strncasecmp(str1,"crc",3))m=UDM_DBMODE_SINGLE_CRC;
-	else
 	if(!strncasecmp(str1,"multi",5))m=UDM_DBMODE_MULTI;
 	else
-	if(!strncasecmp(str1,"word2url",8))m=UDM_DBMODE_WORD2URL;
-	else
-	if(!strncasecmp(str1,"cache",5))m=UDM_DBMODE_CACHE;
-	else
 	if(!strncasecmp(str1,"blob",4))m=UDM_DBMODE_BLOB;
 	return(m);
 }
 
-int UdmDBSetAddr(UDM_DB *db, const char *dbaddr, int mode){
-	char	*s, *cport;
-	char *stored_host = NULL, *cached_host = NULL;
-	struct hostent *hp;
-	struct sockaddr_in server_addr;
-	int nport;
-	
-	if (!dbaddr) return UDM_ERROR;
+__C_LINK const char* __UDMCALL UdmDBTypeToStr(int dbtype)
+{
+  switch(dbtype)
+  {
+    case UDM_DB_MYSQL:   return "mysql";
+    case UDM_DB_PGSQL:   return "pgsql";
+    case UDM_DB_IBASE:   return "ibase";
+    case UDM_DB_MSSQL:   return "mssql";
+    case UDM_DB_ORACLE8: return "oracle";
+    case UDM_DB_SQLITE:  return "sqlite";
+    case UDM_DB_MIMER:   return "mimer";
+    case UDM_DB_VIRT:    return "virtuoso";
+    case UDM_DB_ACCESS:  return "access";
+    case UDM_DB_DB2:     return "db2";
+    case UDM_DB_CACHE:   return "cache";
+  }
+  return "unknown_dbtype";
+}
 
-	if(UdmURLParse(&db->addr,dbaddr))return UDM_ERROR;
-	UDM_FREE(db->DBADDR);
+__C_LINK const char* __UDMCALL UdmDBModeToStr(int dbmode) {
+  switch(dbmode) 
+  {
+    case UDM_DBMODE_SINGLE:	return "single";
+    case UDM_DBMODE_MULTI:	return "multi";
+    case UDM_DBMODE_BLOB:	return "blob";
+  }
+  return "unknown_dbmode";
+}
+
+static int UdmDBSetParam(UDM_DB *db, char *param)
+{
+  char *tok, *lt;
+  const char *v;
+  
+  for(tok = udm_strtok_r(param, "&",&lt) ; tok ; 
+      tok = udm_strtok_r(NULL,"&",&lt))
+  {
+    char * val;
+    if((val=strchr(tok,'=')))
+    {
+      *val++='\0';
+      UdmVarListReplaceStr(&db->Vars, tok, val);
+    }
+    else
+    {
+      UdmVarListReplaceStr(&db->Vars, tok, "");
+    }
+  }
+  
+  if((v= UdmVarListFindStr(&db->Vars,"numtables",NULL)))
+  {
+    db->numtables= atoi(v);
+    if(!db->numtables)
+      db->numtables=1;
+  }
+  
+  if((v= UdmVarListFindStr(&db->Vars,"dbmode",NULL)))
+  {
+    if ((db->DBMode=UdmStr2DBMode(v)) < 0) 
+      return UDM_ERROR;
+  }
+  
+  return UDM_OK;
+}
+
+
+int UdmDBSetAddr(UDM_DB *db, const char *dbaddr, int mode)
+{
+	UDM_URL addr;
+	char	*s;
+	
+	UdmVarListFree(&db->Vars);
 	UDM_FREE(db->DBName);
-	UDM_FREE(db->DBUser);
-	UDM_FREE(db->DBPass);
-	UDM_FREE(db->DBSock);
 	UDM_FREE(db->where);
 	UDM_FREE(db->from);
-
-	db->open_mode=mode;
-	db->DBADDR = (char*)strdup(dbaddr);
-
-	if (db->addr.schema == NULL) return UDM_ERROR;
-	if(!strcasecmp(db->addr.schema,"cached")){
-		db->DBType = UDM_DB_CACHED;
+	
+	UdmVarListReplaceStr(&db->Vars, "DBAddr", dbaddr);
+	
+	UdmURLInit(&addr);
+	
+	if((!dbaddr) || UdmURLParse(&addr, dbaddr) || (!addr.schema))
+	{
+		/* TODO: add better error message here */
+		return UDM_ERROR;
 	}
-	else if(!strcasecmp(db->addr.schema,"cache")){
-		db->DBType = UDM_DB_CACHE;
+	
+	if((s=strchr(UDM_NULL2EMPTY(addr.auth),':')))
+	{
+		*s=0;
+		UdmVarListReplaceStr(&db->Vars, "DBUser", addr.auth);
+		UdmVarListReplaceStr(&db->Vars, "DBPass", s+1);
+		*s=':';
 	}
-	else if(!strcasecmp(db->addr.schema,"searchd")){
+	else
+	{
+		UdmVarListReplaceStr(&db->Vars, "DBUser", addr.auth);
+	}
+	
+	UdmVarListReplaceStr(&db->Vars, "DBHost", addr.hostname);
+	if (addr.port)
+		UdmVarListReplaceInt(&db->Vars, "DBPort", addr.port);
+	
+	if((s = strchr(UDM_NULL2EMPTY(addr.filename), '?')))
+	{
+		*s++='\0';
+		if (UDM_OK != UdmDBSetParam(db, s))
+			return UDM_ERROR;
+		UdmVarListReplaceStr(&db->Vars, "filename", s);
+	}
+	else
+	{
+		UdmVarListReplaceStr(&db->Vars, "filename", addr.filename);
+	}
+	
+	if(!strcasecmp(addr.schema, "searchd"))
+	{
 		db->DBType=UDM_DB_SEARCHD;
+		db->DBDriver=UDM_DB_SEARCHD;
 		if (UDM_OK != UdmSearchdConnect(db)) return UDM_ERROR;
+		goto cont;
 	}
-#if (HAVE_MSQL||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"msql")){
-		db->DBType=UDM_DB_MSQL;
-		db->DBSQL_LIMIT=1;
-	}
-#endif
-#if (HAVE_SOLID||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"solid")){
-		db->DBType=UDM_DB_SOLID;
-		db->DBSQL_IN=1;
-		db->DBSQL_GROUP=1;
-	}
-#endif
-#if (HAVE_ORACLE7||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"oracle7")){
-		db->DBType=UDM_DB_ORACLE7;
-		db->DBSQL_IN=1;
-		db->DBSQL_GROUP=1;
-		db->DBSQL_TRUNCATE=1;
-		db->DBSQL_SUBSELECT = 1;
-	}
-#endif
-#if (HAVE_ORACLE8||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"oracle8")){
-		db->DBType=UDM_DB_ORACLE8;
-		db->DBSQL_IN=1;
-		db->DBSQL_GROUP=1;
-		db->DBSQL_TRUNCATE=1;
-		db->DBSQL_SUBSELECT = 1;
-	}
-	else if(!strcasecmp(db->addr.schema,"oracle")){
+	
+	/* Try native drivers first */
+	
+#if (HAVE_ORACLE8)
+	if(!strcasecmp(addr.schema,"oracle8")||
+	   !strcasecmp(addr.schema,"oracle")){
 		db->DBType=UDM_DB_ORACLE8;
+		db->DBDriver=UDM_DB_ORACLE8;
 		db->DBSQL_IN=1;
 		db->DBSQL_GROUP=1;
 		db->DBSQL_TRUNCATE=1;
 		db->DBSQL_SUBSELECT = 1;
+		goto cont;
 	}
 #endif
-#if (HAVE_CTLIB||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"mssql")){
+#if (HAVE_CTLIB)
+	if(!strcasecmp(addr.schema,"mssql")){
 		db->DBType=UDM_DB_MSSQL;
+		db->DBDriver=UDM_DB_MSSQL;
 		db->DBSQL_IN=1;
 		db->DBSQL_GROUP=1;
 		db->DBSQL_TRUNCATE=1;
+		goto cont;
 	}
 #endif
-#if (HAVE_MYSQL||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"mysql")){
+#if (HAVE_MYSQL)
+	if(!strcasecmp(addr.schema,"mysql")){
 		db->DBType=UDM_DB_MYSQL;
+		db->DBDriver=UDM_DB_MYSQL;
 		db->DBSQL_IN=1;
 		db->DBSQL_LIMIT=1;
 		db->DBSQL_GROUP=1;
+		goto cont;
 	}
 #endif
-#if (HAVE_PGSQL||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"pgsql")){
+#if (HAVE_PGSQL)
+	if(!strcasecmp(addr.schema,"pgsql")){
 		db->DBType=UDM_DB_PGSQL;
+		db->DBDriver=UDM_DB_PGSQL;
 		db->DBSQL_IN=1;
 		db->DBSQL_LIMIT=1;
 		db->DBSQL_GROUP=1;
 		db->DBSQL_SELECT_FROM_DELETE=0;
 		db->DBSQL_SUBSELECT = 1;
+		goto cont;
 	}
 #endif
-#if (HAVE_IBASE||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"ibase")){
+#if (HAVE_IBASE)
+	if(!strcasecmp(addr.schema,"ibase")){
 		db->DBType=UDM_DB_IBASE;
-		
+		db->DBDriver=UDM_DB_IBASE;
 		/* 
 		while indexing large sites and using the SQL in statement 
 		interbase will fail when the items in the in IN statements
@@ -1018,171 +1027,152 @@ int UdmDBSetAddr(UDM_DB *db, const char *dbaddr, int mode){
 		
 		/* db->DBSQL_IN=1; */ 
 		db->DBSQL_GROUP=1;
+		goto cont;
 	}
 #endif
 #if (HAVE_SQLITE)
-	else if (!strcasecmp(db->addr.schema,"sqlite")){
+	if (!strcasecmp(addr.schema,"sqlite")){
 		db->DBType=UDM_DB_SQLITE;
+		db->DBDriver=UDM_DB_SQLITE;
 		db->DBSQL_IN=1;
 		db->DBSQL_LIMIT=1;
 		db->DBSQL_GROUP=1;
+		goto cont;
 	}
 #endif
-#if (HAVE_SAPDB||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"sapdb")){
+
+
+	/* Connect via ODBC */
+
+#if (HAVE_ODBC)
+
+	/* This allow to use ODBC even if a native driver is compiled */
+	s= addr.schema + (strncasecmp(addr.schema,"odbc-",5) ? 0 : 5);
+	
+	if(!strcasecmp(s,"solid")){
+		db->DBType=UDM_DB_SOLID;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_GROUP=1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"sapdb")){
 		db->DBType=UDM_DB_SAPDB;
+		db->DBDriver=UDM_DB_ODBC;
 		db->DBSQL_IN=1;
 		db->DBSQL_GROUP=1;
+		goto cont;
 	}
-#endif
-#if (HAVE_DB2||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"db2")){
+	if(!strcasecmp(s,"db2")){
 		db->DBType=UDM_DB_DB2;
+		db->DBDriver=UDM_DB_ODBC;
 		db->DBSQL_IN=1;
 		db->DBSQL_GROUP=1;
+		goto cont;
 	}
-#endif
-#if (HAVE_DB_ACCESS||HAVE_ODBC)
-	else if(!strcasecmp(db->addr.schema,"access")){
+	if(!strcasecmp(s,"access")){
 		db->DBType=UDM_DB_ACCESS;
+		db->DBDriver=UDM_DB_ODBC;
 		db->DBSQL_IN=1;
 		db->DBSQL_GROUP=1;
 		db->DBSQL_SUBSELECT = 1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"mimer")){
+		db->DBType=UDM_DB_MIMER;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_GROUP=1;
+		db->DBSQL_SUBSELECT = 1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"cache")){
+		db->DBType=UDM_DB_CACHE;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_GROUP=1;
+		db->DBSQL_SUBSELECT = 1;
+		goto cont;
+	}	
+	if(!strcasecmp(s,"virtuoso")){
+		db->DBType=UDM_DB_VIRT;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_GROUP=1;
+		db->DBSQL_SUBSELECT = 1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"oracle8")||
+		!strcasecmp(addr.schema,"oracle")){
+		db->DBType=UDM_DB_ORACLE8;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_GROUP=1;
+		db->DBSQL_TRUNCATE=1;
+		db->DBSQL_SUBSELECT = 1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"mssql")){
+		db->DBType=UDM_DB_MSSQL;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_GROUP=1;
+		db->DBSQL_TRUNCATE=1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"mysql")){
+		db->DBType=UDM_DB_MYSQL;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_LIMIT=1;
+		db->DBSQL_GROUP=1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"pgsql")){
+		db->DBType=UDM_DB_PGSQL;
+		db->DBDriver=UDM_DB_ODBC;
+		db->DBSQL_IN=1;
+		db->DBSQL_LIMIT=1;
+		db->DBSQL_GROUP=1;
+		db->DBSQL_SELECT_FROM_DELETE=0;
+		db->DBSQL_SUBSELECT = 1;
+		goto cont;
+	}
+	if(!strcasecmp(s,"ibase")){
+		db->DBType=UDM_DB_IBASE;
+		db->DBDriver=UDM_DB_ODBC;
+		/* 
+		while indexing large sites and using the SQL in statement 
+		interbase will fail when the items in the in IN statements
+		are more then 1500. We'd better have to fix code to avoid 
+		big INs instead of hidding DBSQL_IN.
+		*/
+		
+		/* db->DBSQL_IN=1; */ 
+		db->DBSQL_GROUP=1;
+		goto cont;
 	}
 #endif
-	else {
-		return UDM_ERROR;
-	}
 	
-	db->DBDriver=db->DBType;
 	
-	if((s = strchr(UDM_NULL2EMPTY(db->addr.filename), '?'))) {
-		char * tok, *lt;
-		
-		*s++='\0';
-		tok = udm_strtok_r(s,"&",&lt);
-		while(tok){
-			char * val;
-			
-			if((val=strchr(tok,'='))){
-				*val++='\0';
-				if(!strcasecmp(tok,"socket")&&val[0]){
-					UDM_FREE(db->DBSock);
-					db->DBSock = (char*)strdup(val);
-				}else
-				if(!strcasecmp(tok,"numtables")&&val[0]){
-					db->numtables=atoi(val);
-					if(!db->numtables)
-						db->numtables=1;
-				}else
-				if(!strcasecmp(tok,"dbmode")&&val[0]){
-				  if ((db->DBMode=UdmStr2DBMode(val)) < 0) {
-				    return UDM_ERROR;
-				  }
-				}else
-				if(!strcasecmp(tok, "stored") && val[0]){
-				  stored_host = (char*)strdup(val);
-				}else
-				if(!strcasecmp(tok, "cached") && val[0]){
-				  cached_host = (char*)strdup(val);
-				}
-			}
-			if (!strcasecmp(tok, "trackquery")) {
-			  db->TrackQuery = 1;
-			}
-			tok = udm_strtok_r(NULL,"&",&lt);
-		}
-	}
+	/* No suitable drivers found*/
 	
-	if(db->DBType==UDM_DB_IBASE || db->DBType==UDM_DB_SQLITE){
+	return UDM_ERROR;
+	
+cont:
+	
+	if(db->DBDriver==UDM_DB_IBASE || db->DBDriver==UDM_DB_SQLITE)
+	{
 		/* Ibase is a special case        */
 		/* It's database name consists of */
 		/* full path and file name        */ 
-		db->DBName = (char*)strdup(UDM_NULL2EMPTY(db->addr.path));
-	}else{
-		db->DBName = (char*)strdup(UDM_NULL2EMPTY(db->addr.path));
-		sscanf(UDM_NULL2EMPTY(db->addr.path), "/%[^/]s", db->DBName);
+		db->DBName = (char*)UdmStrdup(UDM_NULL2EMPTY(addr.path));
 	}
-	if((s=strchr(UDM_NULL2EMPTY(db->addr.auth),':'))){
-		*s=0;
-		db->DBUser = (char*)strdup(db->addr.auth);
-		db->DBPass = (char*)strdup(s+1);
-		*s=':';
-	}else{
-		db->DBUser = (char*)strdup(UDM_NULL2EMPTY(db->addr.auth));
+	else
+	{
+		db->DBName = (char*)UdmStrdup(UDM_NULL2EMPTY(addr.path));
+		sscanf(UDM_NULL2EMPTY(addr.path), "/%[^/]s", db->DBName);
 	}
-
-	if (stored_host) {
-	  if((cport = strchr(stored_host, ':'))) {
-	    *cport = '\0';
-	    nport = atoi(cport + 1);
-	  } else {
-	    nport = UDM_STORED_PORT;
-	  }
-
-	  if ((hp = gethostbyname(stored_host)) == 0 ) {
-	    UDM_FREE(stored_host); UDM_FREE(cached_host);
-	    fprintf(stderr, "StoreOpen ERR gethostbyname: %s", hstrerror(h_errno));
-	    return UDM_ERROR;
-	  }
-	  bzero((void*)&server_addr, sizeof(server_addr));
-	  memmove(&server_addr.sin_addr, hp->h_addr, (size_t)hp->h_length);
-	  server_addr.sin_family = hp->h_addrtype;
-	  server_addr.sin_port = htons((u_short)nport);
-	  if((db->stored_sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
-	    UDM_FREE(stored_host); UDM_FREE(cached_host);
-	    fprintf(stderr, "StoreOpen ERR socket: %s", strerror(errno));
-	    return UDM_ERROR;
-	  }
-  
-	  if(connect(db->stored_sd, (struct sockaddr *)&server_addr, sizeof(server_addr)) == -1) {
-	    UDM_FREE(stored_host); UDM_FREE(cached_host);
-	    fprintf(stderr, "StoreOpen ERR connect to %s: %s", inet_ntoa(server_addr.sin_addr), strerror(errno));
-	    return UDM_ERROR;
-	  }
-	}
-
-	if (cached_host) {
-	  struct timeval tval;
-
-	  if((cport = strchr(cached_host, ':')) !=NULL ) {
-	    *cport = '\0';
-	    nport = atoi(cport + 1);
-	  } else {
-	    nport = UDM_LOGD_PORT;
-	  }
-
-	  if ((hp = gethostbyname(cached_host ? cached_host : db->addr.hostname)) == 0 ) {
-	    UDM_FREE(stored_host); UDM_FREE(cached_host);
-	    fprintf(stderr, "CachedOpen ERR gethostbyname: %s", hstrerror(h_errno));
-	    return UDM_ERROR;
-	  }
-	  bzero((void*)&server_addr, sizeof(server_addr));
-	  memmove(&server_addr.sin_addr, hp->h_addr, (size_t)hp->h_length);
-	  server_addr.sin_family = hp->h_addrtype;
-	  server_addr.sin_port = htons((u_short)nport);
-	  if((db->cached_sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
-	    UDM_FREE(stored_host); UDM_FREE(cached_host);
-	    fprintf(stderr, "CachedOpen ERR socket: %s", strerror(errno));
-	    return UDM_ERROR;
-	  }
-  
-	  tval.tv_sec = 300;
-	  tval.tv_usec = 0;
-#if !defined(sgi) && !defined(__sgi) && !defined(__irix__)
-	  if (setsockopt(db->cached_sd, SOL_SOCKET, SO_SNDTIMEO, (char *)&tval, sizeof(tval)) != 0){
-	        fprintf(stderr, "%s [%d] setsockopt error: %d (%s)\n", __FILE__, __LINE__, errno, strerror(errno));
-		return UDM_ERROR;
-	  }
-#endif
-	  if(connect(db->cached_sd, (struct sockaddr *)&server_addr, sizeof(server_addr)) == -1) {
-	    UDM_FREE(stored_host); UDM_FREE(cached_host);
-	    fprintf(stderr, "CachedOpen ERR connect to %s: %s", inet_ntoa(server_addr.sin_addr), strerror(errno));
-	    return UDM_ERROR;
-	  }
-	}
-
-	UDM_FREE(stored_host); UDM_FREE(cached_host);
 	return UDM_OK;
 }
 
@@ -1192,12 +1182,11 @@ __C_LINK int __UDMCALL UdmStatAction(UDM_AGENT *A, UDM_STATLIST *S){
 	int	res=UDM_ERROR;
 	size_t i, dbfrom = 0, dbto =  A->Conf->dbl.nitems;
 	
-	bzero((void*)S, sizeof(S[0]));
+	S->nstats = 0;
+	S->Stat = NULL;
 
 	for (i = dbfrom; i < dbto; i++) {
 	  db = &A->Conf->dbl.db[i];
-/*	  if(db->DBDriver == UDM_DB_CACHE)
-		res = UdmStatActionCache(A, S, db); */ /* FIXME: usr this after it implementation */
 #ifdef HAVE_SQL
 	  res = UdmStatActionSQL(A, S, db);
 #endif
@@ -1225,20 +1214,21 @@ unsigned int UdmGetCategoryId(UDM_ENV *Conf, char *category) {
 }
 
 
-int UdmTrack(UDM_AGENT * query, UDM_RESULT *Res) {
+int UdmTrack(UDM_AGENT * query, UDM_RESULT *Res)
+{
   int rc = UDM_OK;
-  size_t i, dbfrom = 0, dbto =  query->Conf->dbl.nitems; 
-  UDM_DB *db;
-
-	for (i = dbfrom; i < dbto; i++) {
-	  db = &query->Conf->dbl.db[i];
-	  if(db->TrackQuery) {
 #ifdef HAVE_SQL
-	    rc = UdmTrackSQL(query, Res, db);
+  size_t i, dbfrom = 0, dbto=  query->Conf->dbl.nitems; 
+  
+  for (i = dbfrom; i < dbto; i++)
+  {
+    const char *v;
+    UDM_DB *db = &query->Conf->dbl.db[i];
+    if((v= UdmVarListFindStr(&db->Vars,"trackquery",NULL)))
+      rc = UdmTrackSQL(query, Res, db);
+  }
 #endif
-	  }
-	}
-	  return rc;
+  return rc;
 }
 
 UDM_RESULT * UdmCloneList(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc) {
@@ -1263,10 +1253,38 @@ UDM_RESULT * UdmCloneList(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc) {
 	    }
 	    if (rc != UDM_OK) break;
 	  }
+	  if (Res->num_rows > 0) return Res;
+	  UdmResultFree(Res);
 	  return NULL;
 }
 
 
+int UdmCheckUrlid(UDM_AGENT *Agent, urlid_t id) {
+	size_t i, dbfrom = 0, dbto;
+	UDM_DB		*db;
+	int		rc = 0;
+
+	UDM_GETLOCK(Agent, UDM_LOCK_CONF);
+	dbto =  Agent->Conf->dbl.nitems;
+	UDM_RELEASELOCK(Agent, UDM_LOCK_CONF);
+
+	for (i = dbfrom; i < dbto; i++) {
+	    db = &Agent->Conf->dbl.db[i];
+	    UDM_GETLOCK(Agent, UDM_LOCK_DB); 
+	    switch(db->DBDriver) {
+#ifdef HAVE_SQL
+	    default:
+	      rc = UdmCheckUrlidSQL(Agent, db, id);
+	      break;
+#endif
+	    }
+	    UDM_RELEASELOCK(Agent, UDM_LOCK_DB);
+	    if (rc != 0) break;
+	}
+	return rc;
+}
+
+
 /********************************************************/
 
 
@@ -1278,7 +1296,7 @@ UDM_DBLIST * UdmDBListInit(UDM_DBLIST * List){
 size_t UdmDBListAdd(UDM_DBLIST *List, const char * addr, int mode){
 	UDM_DB	*db;
 	int res;
-	db=List->db=(UDM_DB*)realloc(List->db,(List->nitems+1)*sizeof(UDM_DB));
+	db=List->db=(UDM_DB*)UdmRealloc(List->db,(List->nitems+1)*sizeof(UDM_DB));
 	db+=List->nitems;
 	UdmDBInit(db);
 	res = UdmDBSetAddr(db, addr, mode);
@@ -1296,3 +1314,294 @@ void UdmDBListFree(UDM_DBLIST *List){
 	UDM_FREE(List->db);
 	UdmDBListInit(List);
 }
+
+/********************* MultiCache stuff *************************/
+UDM_MULTI_CACHE *UdmMultiCacheInit (UDM_MULTI_CACHE *cache) {
+	size_t i;
+
+	if (! cache) {
+		cache = UdmMalloc(sizeof(UDM_MULTI_CACHE));
+		if (! cache) return(NULL);
+		cache->free = 1;
+	} else {
+		cache->free = 0;
+	}
+	cache->nrecs = 0;
+
+	for (i = 0; i <= MULTI_DICTS; i++) {
+		cache->tables[i].nurls = 0;
+		cache->tables[i].urls = NULL;
+	}
+
+	cache->nurls = 0;
+	cache->urls = NULL;
+
+	return(cache);
+}
+
+void UdmMultiCacheFree (UDM_MULTI_CACHE *cache) {
+	size_t w, s, u, t;
+
+	if (! cache) return;
+	for (t = 0; t <= MULTI_DICTS; t++) {
+		UDM_MULTI_CACHE_TABLE *table = &cache->tables[t];
+		for (u = 0; u < table->nurls; u++) {
+			UDM_MULTI_CACHE_URL *url = &table->urls[u];
+			for (s = 0; s < url->nsections; s++) {
+				UDM_MULTI_CACHE_SECTION *section = &url->sections[s];
+				for (w = 0; w < section->nwords; w++) {
+					UDM_MULTI_CACHE_WORD *word = &section->words[w];
+					UdmFree(word->word);
+					UdmFree(word->intags);
+				}
+				UdmFree(section->words);
+			}
+			UdmFree(url->sections);
+		}
+		UdmFree(table->urls);
+		cache->tables[t].nurls = 0;
+		cache->tables[t].urls = NULL;
+	}
+
+	UdmFree(cache->urls);
+	cache->nurls = 0;
+	cache-> urls = NULL;
+	cache->nrecs = 0;
+
+	if (cache->free) UdmFree(cache);
+}
+
+int UdmMultiCacheFlush (UDM_AGENT *Indexer) {
+#ifdef HAVE_SQL
+	size_t i;
+
+	for (i = 0; i < Indexer->Conf->dbl.nitems; i++) {
+		UDM_DB *db = &Indexer->Conf->dbl.db[i];
+		UDM_GETLOCK(Indexer, UDM_LOCK_DB);
+		if (db->DBMode == UDM_DBMODE_MULTI) {
+			UdmMultiCacheWrite(Indexer, db, 0);
+		}
+		UDM_RELEASELOCK(Indexer, UDM_LOCK_DB);
+	}
+#endif
+	return(0);
+}
+
+int UdmMulti2Blob (UDM_AGENT *Indexer) {
+#ifdef HAVE_SQL
+	size_t i;
+	unsigned long ticks;
+
+	UdmLog(Indexer,UDM_LOG_ERROR,"Converting multi to blob");
+	ticks=UdmStartTimer();
+
+	for (i = 0; i < Indexer->Conf->dbl.nitems; i++) {
+		UDM_DB *db = &Indexer->Conf->dbl.db[i];
+		UDM_GETLOCK(Indexer, UDM_LOCK_DB);
+		if (db->DBMode == UDM_DBMODE_MULTI) UdmMulti2BlobSQL(Indexer, db);
+		UDM_RELEASELOCK(Indexer, UDM_LOCK_DB);
+	}
+
+	ticks=UdmStartTimer()-ticks;
+	UdmLog(Indexer,UDM_LOG_ERROR,"Converting multi to blob finished\t%.2f",(float)ticks/1000);
+#endif
+	return(0);
+}
+
+static size_t UdmMultiCacheAddWord (UDM_MULTI_CACHE_WORD *cache, uint4 coord) {
+	uint4 *tmp;
+
+	if (! cache) return(0);
+	tmp = UdmRealloc(cache->intags, (cache->nintags + 1) * sizeof(uint4));
+	if (! tmp) return(0);
+	cache->intags = tmp;
+	cache->intags[cache->nintags] = coord;
+	cache->nintags++;
+	return(1);
+}
+
+static size_t UdmMultiCacheAddSection (UDM_MULTI_CACHE_SECTION *cache, UDM_WORD *word) {
+	size_t i;
+	uint4 coord = UDM_WRDPOS(word->coord) & 0xFFFF;
+
+	if (! cache) return(0);
+
+	for (i = 0; i < cache->nwords; i++) {
+		if (! strcmp(cache->words[i].word, word->word))  break;
+	}
+
+	if (i == cache->nwords) {
+		UDM_MULTI_CACHE_WORD *tmp;
+		tmp = UdmRealloc(cache->words, (cache->nwords + 1) * sizeof(UDM_MULTI_CACHE_WORD));
+		if (! tmp) return(0);
+		cache->words = tmp;
+		cache->words[cache->nwords].word = (char *)UdmStrdup(word->word);
+		cache->words[cache->nwords].nintags = 0;
+		cache->words[cache->nwords].intags = NULL;
+		cache->nwords++;
+	}
+
+	return(UdmMultiCacheAddWord(&cache->words[i], coord));
+}
+
+static size_t UdmMultiCacheAddURL (UDM_MULTI_CACHE_URL *cache, UDM_WORD *word) {
+	size_t i;
+	unsigned char secno = UDM_WRDSEC(word->coord);
+
+	if (! cache) return(0);
+
+	for (i = 0; i < cache->nsections; i++) {
+		if (cache->sections[i].secno == secno) break;
+	}
+
+	if (i == cache->nsections) {
+		UDM_MULTI_CACHE_SECTION *tmp;
+		tmp = UdmRealloc(cache->sections, (cache->nsections + 1) * sizeof(UDM_MULTI_CACHE_SECTION));
+		if (! tmp) return(0);
+		cache->sections = tmp;
+		cache->sections[cache->nsections].secno = secno;
+		cache->sections[cache->nsections].nwords = 0;
+		cache->sections[cache->nsections].words = NULL;
+		cache->nsections++;
+	}
+
+	return(UdmMultiCacheAddSection(&cache->sections[i], word));
+}
+
+static size_t UdmMultiCacheAddTable (UDM_MULTI_CACHE_TABLE *cache, urlid_t url_id, unsigned char reindex, UDM_WORD *word) {
+	size_t i;
+
+	if (! cache) return(0);
+	for (i = 0; i < cache->nurls; i++) {
+		if (cache->urls[i].url_id == url_id) break;
+	}
+
+	if (i == cache->nurls) {
+		UDM_MULTI_CACHE_URL *tmp;
+		tmp = UdmRealloc(cache->urls, (cache->nurls + 1) * sizeof(UDM_MULTI_CACHE_URL));
+		if (! tmp) return(0);
+		cache->urls = tmp;
+		cache->urls[cache->nurls].url_id = url_id;
+		cache->urls[cache->nurls].reindex = reindex;
+		cache->urls[cache->nurls].nsections = 0;
+		cache->urls[cache->nurls].sections = NULL;
+		cache->nurls++;
+	}
+
+	return(UdmMultiCacheAddURL(&cache->urls[i], word));
+}
+
+size_t UdmMultiCacheAdd (UDM_MULTI_CACHE *cache, urlid_t url_id, unsigned char reindex, UDM_WORD *word) {
+	udmhash32_t table = UdmStrHash32(word->word) & MULTI_DICTS;
+	size_t i;
+	
+	if (! cache) return(0);
+
+	cache->nrecs++;
+
+	if (reindex) {
+		for (i = 0; i < cache->nurls; i++) {
+			if (cache->urls[i] == url_id) break;
+		}
+
+		if (i == cache->nurls) {
+			urlid_t *tmp;
+			tmp = UdmRealloc(cache->urls, (cache->nurls + 1) * sizeof(urlid_t));
+			if (! tmp) return(0);
+			cache->urls = tmp;
+			cache->urls[cache->nurls] = url_id;
+			cache->nurls++;
+		}
+	}
+
+	return(UdmMultiCacheAddTable(&cache->tables[table], url_id, reindex, word));
+}
+
+/******************** BlobCache stuff *******************/
+
+UDM_BLOB_CACHE *UdmBlobCacheInit (UDM_BLOB_CACHE *cache) {
+	if (! cache) {
+		cache = UdmMalloc(sizeof(UDM_BLOB_CACHE));
+		if (! cache) return(NULL);
+		cache->free = 1;
+	} else {
+		cache->free = 0;
+	}
+	cache->nwords = 0;
+	cache->awords = 0;
+	cache->words = NULL;
+
+	return(cache);
+}
+
+void UdmBlobCacheFree (UDM_BLOB_CACHE *cache) {
+	if (! cache) return;
+
+	UdmFree(cache->words);
+	cache->nwords = 0;
+	cache->awords = 0;
+	cache->words = NULL;
+
+	if (cache->free) UdmFree(cache);
+}
+
+size_t UdmBlobCacheAdd (UDM_BLOB_CACHE *cache, urlid_t url_id, unsigned char secno, const char *word, size_t nintags, const char *intag) {
+	if (! cache) {
+		fprintf(stderr, "Cache variable empty\n");
+		return(0);
+	}
+	if (! url_id) {
+		fprintf(stderr, "url_id variable empty\n");
+		return(0);
+	}
+	if (! secno) {
+		fprintf(stderr, "secno variable empty\n");
+		return(0);
+	}
+	if (! word) {
+		fprintf(stderr, "word variable empty\n");
+		return(0);
+	}
+	if (! nintags) {
+		fprintf(stderr, "nintags variable empty\n");
+		return(0);
+	}
+	if (! intag) {
+		fprintf(stderr, "intag variable empty\n");
+		return(0);
+	}
+
+	if (cache->nwords == cache->awords) {
+		UDM_BLOB_CACHE_WORD *tmp;
+		tmp = UdmRealloc(cache->words, (cache->awords + 256) * sizeof(UDM_BLOB_CACHE_WORD));
+		if (! tmp) {
+			fprintf(stderr, "Realloc failed while adding word\n");
+			return(0);
+		}
+		cache->words = tmp;
+		cache->awords += 256;
+	}
+	cache->words[cache->nwords].secno = secno;
+	cache->words[cache->nwords].word = word;
+	cache->words[cache->nwords].url_id = url_id;
+	cache->words[cache->nwords].nintags = nintags;
+	cache->words[cache->nwords].intags = intag;
+	cache->nwords++;
+
+	return(1);
+}
+
+static int bccmpwrd (UDM_BLOB_CACHE_WORD *s1, UDM_BLOB_CACHE_WORD *s2) {
+	int _ = strcmp(s1->word, s2->word);
+	if (! _) _ = s1->secno - s2->secno;
+	if (! _) {
+		if (s1->url_id > s2->url_id) _ = 1;
+		else if (s1->url_id < s2->url_id) _ = -1;
+		else _ = 0;
+	}
+	return(_);
+}
+
+void UdmBlobCacheSort (UDM_BLOB_CACHE *cache) {
+	UdmSort(cache->words, cache->nwords, sizeof(UDM_BLOB_CACHE_WORD), (qsort_cmp)bccmpwrd);
+}
