@@ -54,7 +54,7 @@ __C_LINK UDM_ENV * __UDMCALL UdmEnvInit(UDM_ENV *Conf){
 #endif
 
 	if(!Conf){
-		Conf=(UDM_ENV *)malloc(sizeof(UDM_ENV));
+		Conf=(UDM_ENV *)UdmMalloc(sizeof(UDM_ENV));
 		bzero((void*)Conf, sizeof(*Conf));
 		Conf->freeme=1;
 	}else{
@@ -91,6 +91,7 @@ __C_LINK void __UDMCALL UdmEnvFree(UDM_ENV * Env){
 	UdmMatchListFree(&Env->Aliases);
 	UdmMatchListFree(&Env->ReverseAliases);
 	UdmMatchListFree(&Env->Filters);
+	UdmMatchListFree(&Env->SectionFilters);
 	
 	UdmSynonymListFree(&Env->Synonyms);
 	UdmVarListFree(&Env->Sections);
@@ -113,7 +114,7 @@ __C_LINK char * __UDMCALL UdmEnvErrMsg(UDM_ENV * Conf){
   for(i = 0; i<Conf->dbl.nitems; i++){
     db = &Conf->dbl.db[i];
     if (db->errcode) {
-      char *oe = (char*)strdup(Conf->errstr);
+      char *oe = (char*)UdmStrdup(Conf->errstr);
       udm_snprintf(Conf->errstr, 2048, "DB err: %s - %s", db->errstr, oe);
       UDM_FREE(oe);
     }
