@@ -40,42 +40,41 @@
 
 /*#define DEBUG_GUESSER*/
 
-static UDM_LANGMAP *FindLangMap(UDM_ENV *Env, char *lang, char *charset, char *filename) {
+static UDM_LANGMAP *FindLangMap(UDM_LANGMAPLIST *L, char *lang, char *charset, char *filename) {
   UDM_LANGMAP *o = NULL;
   register size_t i;
 
-  for(i = 0; i < Env->LangMaps.nmaps; i++) {
+  for(i = 0; i < L->nmaps; i++) {
     if(
-       (!strcasecmp(Env->LangMaps.Map[i].charset, charset))
+       (!strcasecmp(L->Map[i].charset, charset))
        &&
-       (!strcasecmp(Env->LangMaps.Map[i].lang, lang))
+       (!strcasecmp(L->Map[i].lang, lang))
        ) {
-      return &Env->LangMaps.Map[i];
+      return &L->Map[i];
     }
   }
 
-  if(Env->LangMaps.nmaps == 0){
-    o = Env->LangMaps.Map=(UDM_LANGMAP*)malloc(sizeof(UDM_LANGMAP));
+  if(L->nmaps == 0){
+    o = L->Map = (UDM_LANGMAP*)UdmMalloc(sizeof(UDM_LANGMAP));
   }else{
-    Env->LangMaps.Map=(UDM_LANGMAP*)realloc(Env->LangMaps.Map,(Env->LangMaps.nmaps+1)*sizeof(UDM_LANGMAP));
-    o = &Env->LangMaps.Map[Env->LangMaps.nmaps];
+    L->Map = (UDM_LANGMAP*)UdmRealloc(L->Map, (L->nmaps + 1) * sizeof(UDM_LANGMAP));
+    o = &L->Map[L->nmaps];
   }
-  if (o == NULL || Env->LangMaps.Map == NULL) {
-    udm_snprintf(Env->errstr, sizeof(Env->errstr), "Can't alloc/realloc for language map (%s, %s), nmaps: %d (%d)", lang, charset, 
-		 Env->LangMaps.nmaps + 1, (Env->LangMaps.nmaps+1)*sizeof(UDM_LANGMAP) );
+  if (o == NULL || L->Map == NULL) {
+    fprintf(stderr, "Can't alloc/realloc for language map (%s, %s), nmaps: %d (%d)", lang, charset, 
+		 L->nmaps + 1, (L->nmaps + 1) * sizeof(UDM_LANGMAP) );
     return NULL;
   }
   bzero((void*)o, sizeof(UDM_LANGMAP));
   for (i = 0; i <= UDM_LM_HASHMASK; i++) o->memb[i].index = i;
-  o->charset = (char*)strdup(charset);
-  o->lang = (char*)strdup(lang);
-  o->filename = (filename == NULL) ? NULL : (char*)strdup(filename);
-  Env->LangMaps.nmaps++;
+  o->charset = (char*)UdmStrdup(charset);
+  o->lang = (char*)UdmStrdup(lang);
+  o->filename = (filename == NULL) ? NULL : (char*)UdmStrdup(filename);
+  L->nmaps++;
   return o;
-  
 }
 
-__C_LINK int __UDMCALL UdmLoadLangMapFile(UDM_ENV * Env, const char * filename){
+__C_LINK int __UDMCALL UdmLoadLangMapFile(UDM_LANGMAPLIST *L, const char * filename) {
      FILE * f;
      char str[1000];
      char *Ccharset = NULL, *Clanguage = NULL;
@@ -84,7 +83,7 @@ __C_LINK int __UDMCALL UdmLoadLangMapFile(UDM_ENV * Env, const char * filename){
 
      f=fopen(filename,"r");
      if(!f){
-          sprintf(Env->errstr,"Can't open LangMapFile '%s'\n",filename);
+          fprintf(stderr, "Can't open LangMapFile '%s'\n", filename);
           return UDM_ERROR;
      }
      while(fgets(str,sizeof(str),f)){
@@ -97,7 +96,7 @@ __C_LINK int __UDMCALL UdmLoadLangMapFile(UDM_ENV * Env, const char * filename){
                if((charset = udm_strtok_r(str+8," \t\n\r",&lasttok))){
                        const char *canon = UdmCharsetCanonicalName(charset);
                     if (canon) {
-                      Ccharset = (char*)strdup(canon);
+                      Ccharset = (char*)UdmStrdup(canon);
                     } else {
                       fprintf(stderr, "Charset: %s in %s not supported\n", charset, filename);
                       return UDM_ERROR;
@@ -108,7 +107,7 @@ __C_LINK int __UDMCALL UdmLoadLangMapFile(UDM_ENV * Env, const char * filename){
                char * lang, *lasttok;
                UDM_FREE(Clanguage);
                if((lang = udm_strtok_r(str+9," \t\n\r",&lasttok))){
-                    Clanguage = (char*)strdup(lang);
+                    Clanguage = (char*)UdmStrdup(lang);
                }
           }else{
                char *s;
@@ -116,19 +115,22 @@ __C_LINK int __UDMCALL UdmLoadLangMapFile(UDM_ENV * Env, const char * filename){
                
                if(!(s=strchr(str,'\t')))continue;
                if(Clanguage == NULL) {
-                 sprintf(Env->errstr, "No language definition in LangMapFile '%s'\n", filename);
+                 fprintf(stderr, "No language definition in LangMapFile '%s'\n", filename);
                  return UDM_ERROR;
                }
 
                if(Ccharset == NULL) {
-                 sprintf(Env->errstr, "No charset definition in LangMapFile '%s'\n", filename);
+                 fprintf(stderr, "No charset definition in LangMapFile '%s'\n", filename);
                  return UDM_ERROR;
                }
                if(!(cs = UdmGetCharSet(Ccharset))) {
-                 sprintf(Env->errstr, "Unknown charset '%s' in LangMapFile '%s'\n", Ccharset, filename);
+                 fprintf(stderr, "Unknown charset '%s' in LangMapFile '%s'\n", Ccharset, filename);
                  return UDM_ERROR;
                }
-               Cmap = FindLangMap(Env, Clanguage, Ccharset, filename);
+               if (Cmap == NULL) {
+		 Cmap = FindLangMap(L, Clanguage, Ccharset, filename);
+		 UdmSort(Cmap->memb, UDM_LM_HASHMASK + 1, sizeof(UDM_LANGITEM), (qsort_cmp)UdmLMcmpIndex);
+	       }
                if (Cmap == NULL) return UDM_ERROR;
                *s='\0';
 
@@ -142,7 +144,7 @@ __C_LINK int __UDMCALL UdmLoadLangMapFile(UDM_ENV * Env, const char * filename){
                     register int hindex;
                     hindex = ((unsigned int)UdmStrHash32(str)) & UDM_LM_HASHMASK;
                     Cmap->memb[hindex].count += count;
-                    strcpy(Env->LangMaps.Map[Env->LangMaps.nmaps-1].memb[hindex].str,str);
+                    strcpy(Cmap->memb[hindex].str, str);
                }
           }
      }
@@ -168,7 +170,7 @@ void UdmLangMapListFree(UDM_LANGMAPLIST *List){
 }
 
 void UdmLangMapListSave(UDM_LANGMAPLIST *List) {
-     size_t i, j;
+     size_t i, j, minv;
      FILE *out;
      UDM_LANGMAP *Cmap;
      char name[128];
@@ -188,9 +190,10 @@ void UdmLangMapListSave(UDM_LANGMAPLIST *List) {
          fprintf(out, "Language: %s\n", Cmap->lang);
          fprintf(out, "Charset:  %s\n", Cmap->charset);
          fprintf(out, "\n\n");
-         UdmSort(Cmap->memb, UDM_LM_HASHMASK + 1, sizeof(UDM_LANGITEM), &UdmLMcmpCount);
+         UdmSort(Cmap->memb, UDM_LM_HASHMASK + 1, sizeof(UDM_LANGITEM), (qsort_cmp)UdmLMcmpCount);
+	 minv = (Cmap->memb[UDM_LM_TOPCNT - 1].count > 1000) ? 1000 : Cmap->memb[UDM_LM_TOPCNT - 1].count;
          for(j = 0; j < UDM_LM_TOPCNT; j++) {
-           Cmap->memb[j].count -= Cmap->memb[UDM_LM_TOPCNT - 1].count - 1;
+           Cmap->memb[j].count -= Cmap->memb[UDM_LM_TOPCNT - 1].count - minv;
          }
          for(j = 0; j < UDM_LM_TOPCNT; j++) {
           char * s;
@@ -237,7 +240,7 @@ void UdmBuildLangMap(UDM_LANGMAP * map,const char * text,size_t textlen, int Str
                t++;
 
                buf[buflen]=code;
-               if (buflen == 0) continue;
+/*               if (buflen == 0) continue;*/
                buf[buflen+1]='\0';
 
                hindex = UdmHash32(buf, buflen + 1);
@@ -262,54 +265,51 @@ void UdmBuildLangMap(UDM_LANGMAP * map,const char * text,size_t textlen, int Str
 
 
 int UdmLMstatcmp(const void * i1, const void * i2){
-  register UDM_MAPSTAT *s1 = (UDM_MAPSTAT *)i1;
-  register UDM_MAPSTAT *s2 = (UDM_MAPSTAT *)i2;
+  register const UDM_MAPSTAT *s1 = (const UDM_MAPSTAT *)i1;
+  register const UDM_MAPSTAT *s2 = (const UDM_MAPSTAT *)i2;
   if (s1->hits > s2->hits) return -1;
   if (s1->hits < s2->hits) return 1;
   if (s1->miss < s2->miss) return -1;
   if (s1->miss > s2->miss) return 1;
+
   return 0;
 }
 
-int UdmLMcmpCount(const void * i1,const void * i2){
-     const UDM_LANGITEM * m1=i1;
-     const UDM_LANGITEM * m2=i2;
-     
-     if (m2->count > m1->count) return 1;
-     if (m2->count < m1->count) return -1;
-     return 0;
+int UdmLMcmpCount(UDM_LANGITEM *m1, UDM_LANGITEM *m2) {
+     if (m2->count > m1->count) return(1);
+     if (m2->count < m1->count) return(-1);
+     return(0);
 }
 
-int UdmLMcmpIndex(const void * i1,const void * i2) {
-     const UDM_LANGITEM * m1=i1;
-     const UDM_LANGITEM * m2=i2;
-     
-     if (m2->index > m1->index) return -1;
-     if (m2->index < m1->index) return 1;
-     return 0;
+int UdmLMcmpIndex(UDM_LANGITEM *m1, UDM_LANGITEM *m2) {
+     if (m1->index > m2->index) return(1);
+     if (m1->index < m2->index) return(-1);
+     return(0);
 }
 
-
 void UdmPrepareLangMap(UDM_LANGMAP * map){
-     int nmemb=UDM_LM_HASHMASK+1;
 
-     UdmSort(map->memb, UDM_LM_HASHMASK + 1, sizeof(UDM_LANGITEM), &UdmLMcmpCount);
-     UdmSort(map->memb, UDM_LM_TOPCNT, sizeof(UDM_LANGITEM), &UdmLMcmpIndex);
+     UdmSort(map->memb, UDM_LM_HASHMASK + 1, sizeof(UDM_LANGITEM), (qsort_cmp)UdmLMcmpCount);
+     UdmSort(map->memb, UDM_LM_TOPCNT, sizeof(UDM_LANGITEM), (qsort_cmp)UdmLMcmpIndex);
 }
 
 
 void UdmCheckLangMap(UDM_LANGMAP * map0, UDM_LANGMAP * map1, UDM_MAPSTAT *stat, size_t InfMiss) {
      register int i;
+     UDM_LANGITEM *HIT;
 
      stat->hits = stat->miss = 0;
      for (i = 0; i < UDM_LM_TOPCNT; i++) {
-       if (bsearch(&map1->memb[i], map0->memb, UDM_LM_TOPCNT, sizeof(UDM_LANGITEM), &UdmLMcmpIndex) == NULL) {
+       if ( (HIT = bsearch(&map1->memb[i], map0->memb, UDM_LM_TOPCNT, sizeof(UDM_LANGITEM), (qsort_cmp)UdmLMcmpIndex)) == NULL) {
 	 stat->miss++;
        } else {
-	 stat->hits++;
+	 stat->hits += UDM_LM_TOPCNT - abs(HIT - map0->memb);
        }
        if (stat->miss > InfMiss) break;
      }
+#if 0
+     printf("Hit: %d Miss:%d '%s''%s'\n",stat->hits,stat->miss,map0->lang,map0->charset);
+#endif
 }
 
 
@@ -329,7 +329,7 @@ int  UdmGuessCharSet(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc,UDM_LANGMAPLIST *Lis
      use_meta = !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars, "GuesserUseMeta", "yes"), "yes");
      update_lm = !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars, "LangMapUpdate", "no"), "yes");
      UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
-			
+
      if (*charset == '\0') {
        if (*server_charset != '\0' 
 	   && (!use_meta || *meta_charset == '\0' || !strcasecmp(server_charset, meta_charset) )) charset = server_charset;
@@ -374,7 +374,7 @@ int  UdmGuessCharSet(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc,UDM_LANGMAPLIST *Lis
        UdmPrepareLangMap(LangMap);
 
        /* Allocate memory for comparison statistics */
-       if ((mapstat=(UDM_MAPSTAT *)malloc(List->nmaps*sizeof(UDM_MAPSTAT))) == NULL) {
+       if ((mapstat=(UDM_MAPSTAT *)UdmMalloc((List->nmaps + 1) * sizeof(UDM_MAPSTAT))) == NULL) {
 	 UdmLog(Indexer, UDM_LOG_ERROR, "Can't alloc momory for UdmGuessCharSet (%d bytes)", List->nmaps*sizeof(UDM_MAPSTAT));
 	 return UDM_ERROR;
        }
@@ -401,10 +401,10 @@ int  UdmGuessCharSet(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc,UDM_LANGMAPLIST *Lis
        if (List->nmaps > 1) UdmSort(mapstat, List->nmaps, sizeof(UDM_MAPSTAT), &UdmLMstatcmp);
 
        /* Display results, best is shown first */
-/*
-       for (i = 0; i < 5; i++)
-	 fprintf(stderr, "Guesser: %dh:%dm %s-%s\n",mapstat[i].hits, mapstat[i].miss,mapstat[i].map->lang,mapstat[i].map->charset);
-*/        
+
+       for (i = 0; i < ((List->nmaps < 5) ? List->nmaps : 5); i++)
+	 UdmLog(Indexer,UDM_LOG_EXTRA, "Guesser: %dh:%dm %s-%s",mapstat[i].hits, mapstat[i].miss,mapstat[i].map->lang,mapstat[i].map->charset);
+
 
        if (*server_charset != '\0' || *meta_charset != '\0')
 	 for(i=0;i<List->nmaps;i++){
@@ -436,9 +436,9 @@ int  UdmGuessCharSet(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc,UDM_LANGMAPLIST *Lis
           if (mapstat[i].map->charset && *charset == '\0' && (!strcmp(lang, mapstat[i].map->lang)) ) {
                UdmVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
           }
-          
-/*          fprintf(stderr, "Guesser: %dh:%dm %s-%s\n",mapstat[i].hits, mapstat[i].miss,mapstat[i].map->lang,mapstat[i].map->charset);*/
-          
+#if 0
+          fprintf(stderr, "Guesser: %dh:%dm %s-%s\n",mapstat[i].hits, mapstat[i].miss,mapstat[i].map->lang,mapstat[i].map->charset);
+#endif
        }
        if (List->nmaps > 0 && mapstat[0].map->charset && (*charset == '\0') ) {
 	     UdmVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[0].map->charset);
@@ -453,8 +453,8 @@ int  UdmGuessCharSet(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc,UDM_LANGMAPLIST *Lis
          UDM_LANGMAP *Cmap;
 
 	 UDM_GETLOCK(Indexer,UDM_LOCK_CONF);
-         if ((Cmap = FindLangMap(Indexer->Conf, lang, charset, NULL)) != NULL) {
-	   UdmSort(Cmap->memb, UDM_LM_HASHMASK + 1, sizeof(UDM_LANGITEM), &UdmLMcmpIndex);
+         if ((Cmap = FindLangMap(&Indexer->Conf->LangMaps, lang, charset, NULL)) != NULL) {
+	   UdmSort(Cmap->memb, UDM_LM_HASHMASK + 1, sizeof(UDM_LANGITEM), (qsort_cmp)UdmLMcmpIndex);
            for (i = 0; i < UDM_LM_HASHMASK+1; i++) {
 	     if (Cmap->memb[i].count == 0) {
 	       strcpy(Cmap->memb[i].str, LangMap->memb[i].str);
