@@ -54,7 +54,7 @@ __C_LINK int __UDMCALL UdmHrefListAdd(UDM_HREFLIST * HrefList,UDM_HREF * Href){
 	/* Don't add empty or too long link */
 	len = strlen(Href->url);
 	if (len < 1) return 0;
-	ehref = (char*)malloc(len + 1);
+	ehref = (char*)UdmMalloc(len + 1);
 	if(ehref == NULL) return 0;
 
 	strcpy(ehref,Href->url);
@@ -67,7 +67,13 @@ __C_LINK int __UDMCALL UdmHrefListAdd(UDM_HREFLIST * HrefList,UDM_HREF * Href){
 	while(l<=r){
 		c=(l+r)/2;
 		if(!(res=strcmp(HrefList->Href[c].url,ehref))){
-			HrefList->Href[c].stored|=Href->stored;
+			HrefList->Href[c].stored = Href->stored;
+			HrefList->Href[c].referrer = Href->referrer;
+			HrefList->Href[c].hops = Href->hops;
+			HrefList->Href[c].method = Href->method;
+			HrefList->Href[c].stored = Href->stored;
+			HrefList->Href[c].site_id = Href->site_id;
+			HrefList->Href[c].server_id = Href->server_id;
 			UDM_FREE(ehref);
 			return(0);
 		}
@@ -79,16 +85,22 @@ __C_LINK int __UDMCALL UdmHrefListAdd(UDM_HREFLIST * HrefList,UDM_HREF * Href){
 	/* Find in unsorted part */
 	for(i=HrefList->shrefs;i<HrefList->nhrefs;i++){
 		if(!strcmp(HrefList->Href[i].url,ehref)){
-			HrefList->Href[i].stored|=Href->stored;
+			HrefList->Href[i].stored = Href->stored;
+			HrefList->Href[i].referrer = Href->referrer;
+			HrefList->Href[i].hops = Href->hops;
+			HrefList->Href[i].method = Href->method;
+			HrefList->Href[i].stored = Href->stored;
+			HrefList->Href[i].site_id = Href->site_id;
+			HrefList->Href[i].server_id = Href->server_id;
 			UDM_FREE(ehref);
 			return(0);
 		}
 	}
 	if(HrefList->nhrefs>=HrefList->mhrefs){
 		HrefList->mhrefs+=HSIZE;
-		HrefList->Href=(UDM_HREF *)realloc(HrefList->Href,HrefList->mhrefs*sizeof(UDM_HREF));
+		HrefList->Href=(UDM_HREF *)UdmRealloc(HrefList->Href,HrefList->mhrefs*sizeof(UDM_HREF));
 	}
-	HrefList->Href[HrefList->nhrefs].url = (char*)strdup(ehref);
+	HrefList->Href[HrefList->nhrefs].url = (char*)UdmStrdup(ehref);
 	HrefList->Href[HrefList->nhrefs].referrer=Href->referrer;
 	HrefList->Href[HrefList->nhrefs].hops=Href->hops;
 	HrefList->Href[HrefList->nhrefs].method=Href->method;
