@@ -94,7 +94,7 @@ void UdmParseHTTPResponse(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 	if(!Doc->Buf.content)return;
 	
 	/* Copy headers not to break them */
-	headers = (char*)strdup(Doc->Buf.buf);
+	headers = (char*)UdmStrdup(Doc->Buf.buf);
 	
 	/* Now lets parse response header lines */
 	token = udm_strtok_r(headers,"\r\n",&lt);
