@@ -101,9 +101,26 @@ __C_LINK int __UDMCALL UdmStoreHrefs(UDM_AGENT * Indexer) {
 
 	UdmDocInit(&Doc);
 
+	for (i = 0; i < Indexer->Hrefs.dhrefs; i++) {
+		UDM_HREF *H = &Indexer->Hrefs.Href[i];
+		if(H->stored == 0) {
+			if( 1 /*strlen(H->url) <= UDM_URLSIZE */ ) {  /* FIXME: replace this by config parameter chacking */
+				UdmVarListReplaceInt(&Doc.Sections, "Referrer-ID", H->referrer);
+				UdmVarListReplaceUnsigned(&Doc.Sections,"Hops", H->hops);
+				UdmVarListReplaceStr(&Doc.Sections,"URL",H->url?H->url:"");
+				UdmVarListReplaceInt(&Doc.Sections, "URL_ID", UdmStrHash32(H->url ? H->url : ""));
+				UdmVarListReplaceInt(&Doc.Sections,"Site_id", H->site_id);
+				UdmVarListReplaceInt(&Doc.Sections,"Server_id", H->server_id);
+				if(UDM_OK != (res = UdmURLAction(Indexer, &Doc, UDM_URL_ACTION_ADD_LINK))){
+					return(res);
+				}
+			}
+			H->stored=1;
+		}
+	}
 	for (i = Indexer->Hrefs.dhrefs; i < Indexer->Hrefs.nhrefs; i++) {
 		UDM_HREF *H = &Indexer->Hrefs.Href[i];
-		if(!(H->stored)){
+		if(H->stored == 0) {
 			if( 1 /*strlen(H->url) <= UDM_URLSIZE */ ) {  /* FIXME: replace this by config parameter chacking */
 				UdmVarListReplaceInt(&Doc.Sections, "Referrer-ID", H->referrer);
 				UdmVarListReplaceUnsigned(&Doc.Sections,"Hops", H->hops);
@@ -132,6 +149,7 @@ __C_LINK int __UDMCALL UdmStoreHrefs(UDM_AGENT * Indexer) {
 	/* FIXME: this is incorrect with both SQL and built-in compiled */
 	if(Indexer->Hrefs.nhrefs > MAXHSIZE)
 		UdmHrefListFree(&Indexer->Hrefs);
+
 	return UDM_OK;
 }
 
@@ -141,7 +159,7 @@ static void RelLink(UDM_URL *curURL, UDM_URL *newURL, char **str){
 	const char	*path = (newURL->path && newURL->path[0]) ? newURL->path : curURL->path;
 	const char	*fname = ((newURL->filename && newURL->filename[0]) || (newURL->path && newURL->path[0])) ? 
 	  newURL->filename : curURL->filename;
-	char		*pathfile = (char*)malloc(strlen(UDM_NULL2EMPTY(path)) + strlen(UDM_NULL2EMPTY(fname)) + 5);
+	char		*pathfile = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(path)) + strlen(UDM_NULL2EMPTY(fname)) + 5);
 		
 	if (pathfile == NULL) return;
 	sprintf(pathfile, "/%s%s",  UDM_NULL2EMPTY(path), UDM_NULL2EMPTY(fname));
@@ -150,13 +168,13 @@ static void RelLink(UDM_URL *curURL, UDM_URL *newURL, char **str){
 	
 
 	if (!strcasecmp(UDM_NULL2EMPTY(schema), "mailto") || !strcasecmp(UDM_NULL2EMPTY(schema), "javascript")) {
-	        *str = (char*)malloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(UDM_NULL2EMPTY(newURL->specific)) + 4);
+	        *str = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(UDM_NULL2EMPTY(newURL->specific)) + 4);
 		sprintf(*str, "%s:%s", UDM_NULL2EMPTY(schema), UDM_NULL2EMPTY(newURL->specific));
 	} else if(/*!strcasecmp(UDM_NULL2EMPTY(schema), "file") ||*/ !strcasecmp(UDM_NULL2EMPTY(schema), "htdb")) {
-	        *str = (char*)malloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(pathfile) + 4);
+	        *str = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(pathfile) + 4);
 		sprintf(*str, "%s:%s", UDM_NULL2EMPTY(schema), pathfile);
 	}else{
-	        *str = (char*)malloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(pathfile) + strlen(UDM_NULL2EMPTY(hostinfo)) + 8);
+	        *str = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(schema)) + strlen(pathfile) + strlen(UDM_NULL2EMPTY(hostinfo)) + 8);
 		sprintf(*str, "%s://%s%s", UDM_NULL2EMPTY(schema), UDM_NULL2EMPTY(hostinfo), pathfile);
 	}
 	
@@ -183,7 +201,7 @@ static int UdmDocBaseHref(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 		
 		if(!(parse_res=UdmURLParse(baseURL, basehref))) {
 			UdmURLParse(&Doc->CurURL,basehref);
-			UdmLog(Indexer,UDM_LOG_ERROR,"BASE HREF '%s'",basehref);
+			UdmLog(Indexer, UDM_LOG_DEBUG, "BASE HREF '%s'", basehref);
 		}else{
 			switch(parse_res){
 			case UDM_URL_LONG:
@@ -199,17 +217,45 @@ static int UdmDocBaseHref(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 	return UDM_OK;
 }
 
+#define NS 10
 static int UdmFilterFind(UDM_MATCHLIST *L,const char *newhref,char *reason){
-	UDM_MATCH_PART	P[10];
+	UDM_MATCH_PART	P[NS];
 	UDM_MATCH	*M;
 	int		res=UDM_METHOD_GET;
 	
-	if((M=UdmMatchListFind(L,newhref,10,P))){
+#ifdef USE_PARANOIA
+	void *paran = UdmViolationEnter();
+#endif
+	if((M=UdmMatchListFind(L, newhref, NS, P))) {
 		sprintf(reason,"%s %s %s '%s'",M->arg,UdmMatchTypeStr(M->match_type),M->case_sense?"Sensitive":"InSensitive",M->pattern);
 		res=UdmMethod(M->arg);
 	}else{
 		sprintf(reason,"Allow by default");
 	}
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
+	return res;
+}
+
+static int UdmSectionFilterFind(UDM_MATCHLIST *L, UDM_DOCUMENT *Doc, char *reason){
+	UDM_MATCH_PART	P[NS];
+	UDM_MATCH	*M;
+	int		res=UDM_METHOD_INDEX;
+	
+#ifdef USE_PARANOIA
+	void *paran = UdmViolationEnter();
+#endif
+
+	if((M=UdmMatchSectionListFind(L, Doc, NS, P))) {
+		sprintf(reason,"%s %s %s '%s'",M->arg,UdmMatchTypeStr(M->match_type),M->case_sense?"Sensitive":"InSensitive",M->pattern);
+		res=UdmMethod(M->arg);
+	}else{
+		sprintf(reason,"Allow by default");
+	}
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
 	return res;
 }
 
@@ -218,11 +264,15 @@ int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL, UDM_HREF *Href){
 	UDM_URL		*newURL = UdmURLInit(NULL);
 	char		*newhref = NULL;
 	UDM_MATCH	*Alias;
-	char		*alias;
+	char		*alias = NULL;
 	size_t		aliassize, nparts = 10;
 	UDM_MATCH_PART	Parts[10];
 	UDM_SERVER	*Srv;
 	char		reason[1024]="";
+
+#ifdef USE_PARANOIA
+	void *paran = UdmViolationEnter();
+#endif
 	
 	if((parse_res=UdmURLParse(newURL, Href->url))){
 		switch(parse_res){
@@ -236,19 +286,20 @@ int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL, UDM_HREF *Href){
 	}
 	
 	RelLink(CurURL, newURL, &newhref);
-	aliassize = 128 + 10 * strlen(newhref);
-	alias = (char*)malloc(aliassize);
-	if (alias == NULL) {
-	  UdmLog(Indexer, UDM_LOG_ERROR, "No memory. %s line %d", __FILE__, __LINE__);
-	  goto ret;
-	}
 	
 	UdmLog(Indexer,UDM_LOG_DEBUG,"Link '%s' %s",Href->url,newhref);
 	for(cascade = 0; ((Alias=UdmMatchListFind(&Indexer->Conf->ReverseAliases,newhref,nparts,Parts))) && (cascade < 1024); cascade++) {
+	        aliassize = strlen(Alias->arg) + strlen(Alias->pattern) + strlen(newhref) + 8;
+		alias = (char*)UdmRealloc(alias, aliassize);
+		if (alias == NULL) {
+		  UdmLog(Indexer, UDM_LOG_ERROR, "No memory (%d bytes). %s line %d", aliassize, __FILE__, __LINE__);
+		  goto ret;
+		}
 		UdmMatchApply(alias,aliassize,newhref,Alias->arg,Alias,nparts,Parts);
 		if(alias[0]){
 		  UdmLog(Indexer,UDM_LOG_DEBUG,"ReverseAlias%d: '%s'", cascade, alias);
-		  strcpy(newhref,alias);
+		  UDM_FREE(newhref);
+		  newhref = (char*)UdmStrdup(alias);
 		} else break;
 	}
 	
@@ -290,12 +341,15 @@ int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL, UDM_HREF *Href){
 	
 	/* FIXME: add MaxHops, Robots */
 	UDM_FREE(Href->url);
-	Href->url = (char*)strdup(newhref);
+	Href->url = (char*)UdmStrdup(newhref);
 	Href->server_id = Srv->site_id;
 ret:	
 	UDM_FREE(newhref);
 	UDM_FREE(alias);
 	UdmURLFree(newURL);
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
 	return UDM_OK;
 }
 
@@ -346,6 +400,10 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
 	int result,res;
 	UDM_URL *myurl = UdmURLInit(NULL);
 	UDM_HREF Href;
+
+#ifdef USE_PARANOIA
+	void *paran = UdmViolationEnter();
+#endif
 	
 	/* Read lines and clear/insert/check URLs                     */
 	/* We've already tested in main.c to make sure it can be read */
@@ -377,13 +435,21 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
 		case UDM_URL_FILE_REINDEX:
 			UdmVarListReplaceStr(&Indexer->Conf->Vars, "ul", str);
 			result = UdmURLAction(Indexer, NULL, UDM_URL_ACTION_EXPIRE);
-			if(result!=UDM_OK) { UdmURLFree(myurl); return(result); }
+			if(result!=UDM_OK) { UdmURLFree(myurl); 
+#ifdef USE_PARANOIA
+			UdmViolationExit(paran);
+#endif
+			return(result); }
 			UdmVarListDel(&Indexer->Conf->Vars, "ul");
 			break;
 		case UDM_URL_FILE_CLEAR:
 			UdmVarListReplaceStr(&Indexer->Conf->Vars, "ul", str);
 			result=UdmClearDatabase(Indexer);
-			if(result!=UDM_OK) { UdmURLFree(myurl); return(UDM_ERROR); }
+			if(result!=UDM_OK) { UdmURLFree(myurl); 
+#ifdef USE_PARANOIA
+			UdmViolationExit(paran);
+#endif
+			return(UDM_ERROR); }
 			UdmVarListDel(&Indexer->Conf->Vars, "ul");
 			break;
 		case UDM_URL_FILE_INSERT:
@@ -406,6 +472,9 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
 					UdmLog(Indexer,UDM_LOG_ERROR,"Error in URL: '%s'",str);
 				}
 				UdmURLFree(myurl);
+#ifdef USE_PARANOIA
+				UdmViolationExit(paran);
+#endif
 				return(UDM_ERROR);
 			}
 			break;
@@ -415,6 +484,9 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
 	if(url_file!=stdin)
 		fclose(url_file);
 	UdmURLFree(myurl);
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
 	return(UDM_OK);
 }
 
@@ -434,7 +506,7 @@ static int UdmDocAlias(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 	const char	*url=UdmVarListFindStr(&Doc->Sections,"URL","");
 
 	alstrlen = 256 + strlen(url);
-	if ((alstr = (char*)malloc(alstrlen)) == NULL) return UDM_ERROR;
+	if ((alstr = (char*)UdmMalloc(alstrlen)) == NULL) return UDM_ERROR;
 	alstr[0] = '\0';
 	if(alias_prog){
 		result = UdmAliasProg(Indexer, alias_prog, url, alstr, alstrlen - 1);
@@ -466,7 +538,7 @@ static int UdmDocLookupConn(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 		char *port;
 
 		UdmLog(Indexer, UDM_LOG_DEBUG, "Using Proxy: %s", proxy);
-		Doc->connp.hostname = (char*)strdup(proxy);
+		Doc->connp.hostname = (char*)UdmStrdup(proxy);
 		if((port=strchr(Doc->connp.hostname,':'))){
 			*port++='\0';
 			Doc->connp.port=atoi(port);
@@ -475,7 +547,7 @@ static int UdmDocLookupConn(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 		}
 	}else{
 		if (Doc->CurURL.hostname){
-			Doc->connp.hostname = (char*)strdup(Doc->CurURL.hostname);
+			Doc->connp.hostname = (char*)UdmStrdup(Doc->CurURL.hostname);
 			Doc->connp.port=Doc->CurURL.port?Doc->CurURL.port:Doc->CurURL.default_port;
 		}
 	}
@@ -551,7 +623,7 @@ static int UdmDocCheck(UDM_AGENT *Indexer,UDM_SERVER *CurSrv,UDM_DOCUMENT *Doc){
 }
 
 
-static void UdmAppendTarget(UDM_AGENT *Indexer, const char *url, const char *lang, const int hops) {
+static void UdmAppendTarget(UDM_AGENT *Indexer, const char *url, const char *lang, const int hops, int parent) {
   UDM_DOCUMENT *Doc, *Save;
   size_t i;
 
@@ -569,7 +641,7 @@ static void UdmAppendTarget(UDM_AGENT *Indexer, const char *url, const char *lan
     }
   }
   if ((Indexer->Conf->Targets.Doc = 
-       realloc(Save = Indexer->Conf->Targets.Doc, (Indexer->Conf->Targets.num_rows + 1) * sizeof(UDM_DOCUMENT))) == NULL) {
+       UdmRealloc(Save = Indexer->Conf->Targets.Doc, (Indexer->Conf->Targets.num_rows + 1) * sizeof(UDM_DOCUMENT))) == NULL) {
     Indexer->Conf->Targets.Doc = Save;
     UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
     UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
@@ -581,6 +653,7 @@ static void UdmAppendTarget(UDM_AGENT *Indexer, const char *url, const char *lan
   UdmVarListAddStr(&Doc->Sections, "URL", url);
   UdmVarListAddInt(&Doc->Sections, "Hops", hops);
   UdmVarListReplaceInt(&Doc->Sections, "URL_ID", UdmStrHash32(url));
+  UdmVarListReplaceInt(&Doc->Sections, "Referrer-ID", parent);
   UdmURLAction(Indexer, Doc, UDM_URL_ACTION_ADD);
   if (*lang != '\0') UdmVarListAddStr(&Doc->RequestHeaders, "Accept-Language", lang);
   UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
@@ -591,9 +664,10 @@ static void UdmAppendTarget(UDM_AGENT *Indexer, const char *url, const char *lan
 static int UdmDocProcessResponseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){  /* This function must have exclusive acces to Conf */
 	UDM_VAR		*var;
 	UDM_MATCH_PART	P[10];
-	size_t		sec;
 	const char      *content_type = UdmVarListFindStr(&Doc->Sections, "Content-Type", NULL);
 	const char      *vary = UdmVarListFindStr(&Doc->Sections, "Vary", NULL);
+	const int       content_length = UdmVarListFindInt(&Doc->Sections, "Content-Length", 0);
+	const int       parent = UdmVarListFindInt(&Doc->Sections, "Referrer-ID", 0);
 
 	if (vary != NULL) {
 	  if (strcasestr(vary, "accept-language") != NULL) {
@@ -609,13 +683,13 @@ static int UdmDocProcessResponseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 	    if (strcmp(UDM_NULL2EMPTY(newURL->filename), "robots.txt") != 0) {
 	      CL = UdmVarListFindStr(&Doc->Sections, "Content-Location", UDM_NULL2EMPTY(newURL->filename));
 	      urlen = 128 + strlen(UDM_NULL2EMPTY(newURL->hostinfo)) + strlen(UDM_NULL2EMPTY(newURL->path)) + strlen(CL);
-	      if ((url = (char*)malloc(urlen)) != NULL) {
+	      if ((url = (char*)UdmMalloc(urlen)) != NULL) {
 		snprintf(url, urlen, "%s://%s%s%s", UDM_NULL2EMPTY(newURL->schema), UDM_NULL2EMPTY(newURL->hostinfo), 
 			 UDM_NULL2EMPTY(newURL->path), CL );
-		UdmAppendTarget(Indexer,  url, "", hops);
+		UdmAppendTarget(Indexer,  url, "", hops, parent);
 		tok = udm_strtok_r((char*)VaryLang, " ,\t", &lt);
 		while (tok != NULL) {
-		  UdmAppendTarget(Indexer, ourl, tok, hops );
+		  UdmAppendTarget(Indexer, ourl, tok, hops, parent );
 		  tok = udm_strtok_r(NULL, " ,\t", &lt);
 		}
 		UDM_FREE(url);
@@ -625,6 +699,10 @@ static int UdmDocProcessResponseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 	  }
 	}
 
+	if ((size_t)content_length > Doc->Buf.maxsize) {
+	  UdmVarListReplaceInt(&Doc->Sections, "Status", UDM_HTTP_STATUS_PARTIAL_OK);
+	}
+
 	if (content_type != NULL) {
 	  char *p;
 	  if ((p = strstr(content_type, "charset=")) != NULL) {
@@ -657,12 +735,9 @@ static int UdmDocProcessResponseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 		UdmVarListAddStr(&Doc->Sections,"Content-Type","application/octet-stream");
 	}
 
-	for(sec=0;sec<Doc->Sections.nvars;sec++){
-		var=&Doc->Sections.Var[sec];
-		
-		if(!strcasecmp(var->name,"Location")){
-			UDM_URL *newURL = UdmURLInit(NULL);
-			switch(UdmURLParse(newURL, var->val)) {
+	if ((var = UdmVarListFind(&Doc->Sections, "Location"))) {
+	        UDM_URL *newURL = UdmURLInit(NULL);
+		switch(UdmURLParse(newURL, var->val)) {
 				case UDM_URL_OK:
 					if (UDM_NULL2EMPTY(newURL->schema) != NULL) {
 						UDM_HREF Href;
@@ -682,20 +757,79 @@ static int UdmDocProcessResponseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 				case UDM_URL_BAD:
 				default:
 					UdmLog(Indexer,UDM_LOG_ERROR,"Error in redirect URL: '%s'",var->val);
-			}
-			UdmURLFree(newURL);
 		}
+		UdmURLFree(newURL);
 	}
 	return UDM_OK;
 }
 
 
+#ifdef HAVE_ZLIB
+#include <zlib.h>
+
+static int UdmPutCachedCopy(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc)
+{
+  int rc= UDM_OK;
+  z_stream zstream;
+  char *Zbuf = NULL;
+  urlid_t rec_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
+  size_t content_size=Doc->Buf.size - (Doc->Buf.content-Doc->Buf.buf);
+  size_t Zlen;
+  size_t Blen;
+  UDM_VAR *Sec; 
+  
+  if (!(Sec=UdmVarListFind(&Doc->Sections, "CachedCopy")))
+    return rc;
+  
+  zstream.zalloc = Z_NULL;
+  zstream.zfree = Z_NULL;
+  zstream.opaque = Z_NULL;
+  zstream.next_in = (Byte*)Doc->Buf.content;
+  
+  if (deflateInit2(&zstream, 9, Z_DEFLATED, 15, 9, Z_DEFAULT_STRATEGY) != Z_OK)
+  {
+    rc= UDM_ERROR;
+    goto ex;
+  }
+  
+  zstream.avail_in = content_size;
+  zstream.avail_out = 2 * content_size + 256;
+  if (!(Zbuf=  UdmMalloc(zstream.avail_out)))
+  {
+    rc= UDM_ERROR;
+    goto ex;
+  }
+  zstream.next_out= (Byte*) Zbuf;
+  deflate(&zstream, Z_FINISH);
+  deflateEnd(&zstream);
+  Zlen= (size_t)zstream.total_out;
+  Blen= BASE64_LEN(Zlen);
+  if (Blen > Sec->maxlen)
+    goto ex;
+  
+  UdmLog(Indexer, UDM_LOG_DEBUG, "Stored rec_id: %d Size: %d Ratio: %5.2f%%",
+				  rec_id, content_size, (double)100.0 * zstream.total_out / content_size);
+  
+  if (!(Sec->val= UdmRealloc(Sec->val, Blen)))
+  {
+    rc= UDM_ERROR;
+    goto ex;
+  }
+  Sec->curlen= udm_base64_encode(Zbuf,Sec->val,Zlen);
+  
+ex:
+  UDM_FREE(Zbuf);
+  
+  return rc;
+}
+#endif
 
 
 static int UdmDocParseContent(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
 	
 #ifdef USE_PARSER
 	UDM_PARSER	*Parser;
+	const int       status = UdmVarListFindInt(&Doc->Sections, "Status", 0);
 #endif
 	const char	*real_content_type=NULL;
 	const char	*url=UdmVarListFindStr(&Doc->Sections,"URL","");
@@ -731,33 +865,40 @@ static int UdmDocParseContent(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
 	
 #ifdef USE_PARSER
 	/* Let's try to start external parser for this Content-Type */
-	
-	if((Parser=UdmParserFind(&Indexer->Conf->Parsers,ct))){
+
+	if((Parser = UdmParserFind(&Indexer->Conf->Parsers, ct))) {
 		UdmLog(Indexer,UDM_LOG_DEBUG,"Found external parser '%s' -> '%s'",
 			Parser->from_mime?Parser->from_mime:"NULL",
 			Parser->to_mime?Parser->to_mime:"NULL");
 	}
-	if(Parser && UdmParserExec(Indexer,Parser,Doc)){
+	if(Parser && UdmParserExec(Indexer, Parser, Doc)) {
+	  if (status == UDM_HTTP_STATUS_OK) {
 		char *to_charset;
 		real_content_type=Parser->to_mime?Parser->to_mime:"unknown";
-		UdmLog(Indexer,UDM_LOG_DEBUG,"Parser-Content-Type: %s",real_content_type);
+		UdmLog(Indexer,UDM_LOG_DEBUG,"Parser.Content-Type: %s",real_content_type);
 		if((to_charset=strstr(real_content_type,"charset="))){
 		        const char *cs = UdmCharsetCanonicalName(UdmTrim(to_charset + 8, " \t"));
 			UdmVarListReplaceStr(&Doc->Sections, "Server-Charset", cs);
-			UdmLog(Indexer,UDM_LOG_DEBUG, "to_charset='%s'", cs);
+			UdmLog(Indexer,UDM_LOG_DEBUG, "Parser.Charset: %s", cs);
+			UdmVarListAddStr(&Doc->Sections,"Parser.Charset",cs);
 		}
 #ifdef DEBUG_PARSER
 		fprintf(stderr,"content='%s'\n",Doc->content);
 #endif
+	  } else {
+	    UdmLog(Indexer, UDM_LOG_WARN, "Parser not executed, document status: %d", status);
+	  }
 	}
 #endif
 	
 	if(!real_content_type)real_content_type=ct;
-	UdmVarListAddStr(&Doc->Sections,"Parser-Content-Type",real_content_type);
+	UdmVarListAddStr(&Doc->Sections,"Parser.Content-Type",real_content_type);
 	
 #ifdef HAVE_ZLIB
-	if ( strncmp(real_content_type, "text/", 5) == 0) {
-	  UdmStoreDoc(Indexer, Doc);
+	if (UDM_OK!=(result = UdmPutCachedCopy(Indexer, Doc)))
+	{
+	  UdmLog(Indexer,UDM_LOG_ERROR,"Failed to create cached copy");
+	  return result;
 	}
 #endif
 	
@@ -780,11 +921,9 @@ static int UdmDocParseContent(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
 		if(!strncasecmp(real_content_type,"text/html",9)){
 			UdmHTMLParse(Indexer,Doc);
 		}else
-#ifdef USE_XML
 		if(!strncasecmp(real_content_type, "text/xml",8)) {
 			UdmXMLParse(Indexer, Doc);
 		}else
-#endif
 		if(!strncasecmp(real_content_type, "audio/mpeg",10)) {
 			UdmMP3Parse(Indexer,Doc);
 		}else{
@@ -793,6 +932,7 @@ static int UdmDocParseContent(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
 			UdmVarListReplaceInt(&Doc->Sections,"Status",UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE);
 		}
 	}
+
 	return result;
 }
 
@@ -964,23 +1104,39 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 	UDM_DOCUMENT	Doc;
 	const char	*url, *alias;
 	char		*origurl = NULL, *aliasurl = NULL;
+	UDM_SERVER	*Server = NULL;
+	char		reason[1024] = "";
+#ifdef USE_PARANOIA
+	void * paran = UdmViolationEnter();
+#endif
 	
 	UdmDocInit(&Doc);
 	
 	UDM_THREADINFO(Indexer,"Selecting","");
 
 	if(UDM_OK==(result=UdmStoreHrefs(Indexer))) {
-	        if (Indexer->action != UDM_OK) return Indexer->action;
+	        if (Indexer->action != UDM_OK) {
+#ifdef USE_PARANOIA
+		  UdmViolationExit(paran);
+#endif
+		  return Indexer->action;
+		}
 		result = UdmNextTarget(Indexer, &Doc);
 	}
 	
 	if(result==UDM_NOTARGET){
 		UdmDocFree(&Doc);	/* To free Doc.connp->connp */
+#ifdef USE_PARANOIA
+		UdmViolationExit(paran);
+#endif
 		return result;
 	}
 	
 	if(result!=UDM_OK){
 		UdmDocFree(&Doc);
+#ifdef USE_PARANOIA
+		UdmViolationExit(paran);
+#endif
 		return result;
 	}
 	
@@ -1003,7 +1159,7 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 	        UDM_GETLOCK(Indexer,UDM_LOCK_CONF);
 		Doc.Buf.maxsize=(size_t)UdmVarListFindInt(&Indexer->Conf->Vars,"MaxDocSize",UDM_MAXDOCSIZE);
 		UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
-		if ((Doc.Buf.buf=(char*)malloc(Doc.Buf.maxsize + 1)) == NULL) {
+		if ((Doc.Buf.buf=(char*)UdmMalloc(Doc.Buf.maxsize + 1)) == NULL) {
 		  UdmLog(Indexer, UDM_LOG_ERROR, "Out of memory.");
 		  return UDM_ERROR;
 		}
@@ -1018,7 +1174,6 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 	if ((Doc.CurURL.filename != NULL) && (!strcmp(Doc.CurURL.filename, "robots.txt"))) {
 		Doc.method = UDM_METHOD_DISALLOW;
 	}else{
-		UDM_SERVER	*Server;
 		char		*alstr = NULL;
 		
 		/* Find correspondent Server */
@@ -1052,8 +1207,8 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 
 			if((alias = UdmVarListFindStr(&Doc.Sections, "Alias", NULL))) {
 			  const char *u = UdmVarListFindStr(&Doc.Sections, "URL", NULL);
-			  origurl = (char*)strdup(u);
-			  aliasurl = (char*)strdup(alias);
+			  origurl = (char*)UdmStrdup(u);
+			  aliasurl = (char*)UdmStrdup(alias);
 			  UdmLog(Indexer,UDM_LOG_EXTRA,"Alias: '%s'", alias);
 			  UdmVarListReplaceStr(&Doc.Sections, "URL", alias);
 			  UdmVarListReplaceInt(&Doc.Sections, "URL_ID", UdmStrHash32(alias));
@@ -1071,6 +1226,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 	
 	if(result!=UDM_OK){
 		UdmDocFree(&Doc);
+#ifdef USE_PARANOIA
+		UdmViolationExit(paran);
+#endif
 		return result;
 	}
 	
@@ -1081,7 +1239,7 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 				UdmLog(Indexer,UDM_LOG_WARN, "robots.txt support is disallowed for '%s'", 
 				       UDM_NULL2EMPTY(Doc.CurURL.hostinfo));
 				UDM_GETLOCK(Indexer,UDM_LOCK_CONF);
-				result = UdmRobotParse(Indexer, NULL, UDM_NULL2EMPTY(Doc.CurURL.hostinfo));
+				result = UdmRobotParse(Indexer, NULL, NULL, UDM_NULL2EMPTY(Doc.CurURL.hostinfo));
 				UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
 			}else{
 				UDM_ROBOT_RULE	*rule;
@@ -1093,7 +1251,7 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 				UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
 				
 				if(take_robots){
-				        UDM_SERVER	*Server;
+				        UDM_SERVER	*rServer;
 					char		*rurl;
 					UDM_DOCUMENT	rDoc;
 					size_t          rurlen;
@@ -1102,12 +1260,12 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 					UdmDocInit(&rDoc);
 					rDoc.Buf.maxsize=Doc.Buf.maxsize;
 					rDoc.Spider=Doc.Spider;
-					rDoc.Buf.buf=(char*)malloc(Doc.Buf.maxsize + 1);
+					rDoc.Buf.buf=(char*)UdmMalloc(Doc.Buf.maxsize + 1);
 					rDoc.Buf.buf[0]='\0';
 
 					rurlen = 32 + strlen(UDM_NULL2EMPTY(Doc.CurURL.schema)) + 
 					  strlen(UDM_NULL2EMPTY(Doc.CurURL.hostinfo));
-					rurl = (char*)malloc(rurlen);
+					rurl = (char*)UdmMalloc(rurlen);
 					snprintf(rurl, rurlen, "%s://%s/robots.txt", UDM_NULL2EMPTY(Doc.CurURL.schema),
 						 UDM_NULL2EMPTY(Doc.CurURL.hostinfo));
 					UdmVarListAddStr(&rDoc.Sections,"URL",rurl);
@@ -1116,16 +1274,14 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 					UdmLog(Indexer,UDM_LOG_INFO,"ROBOTS: %s",rurl);
 
 					UDM_GETLOCK(Indexer,UDM_LOCK_CONF);
-					Server = UdmServerFind(Indexer->Conf, &Indexer->Conf->Servers, rurl, NULL);
+					rServer = UdmServerFind(Indexer->Conf, &Indexer->Conf->Servers, rurl, NULL);
 					UdmDocAddDocExtraHeaders(&rDoc);
 					UdmDocAddConfExtraHeaders(Indexer->Conf,&rDoc);
-					UdmRobotAddEmpty(&Indexer->Conf->Robots, UDM_NULL2EMPTY(Doc.CurURL.hostinfo));
 					UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
 
-					if (Server != NULL) {
-					  UdmVarListReplaceLst(&rDoc.Sections, &Server->Vars, NULL, "*");
-					  UdmDocAddServExtraHeaders(Server, &rDoc);
-					}
+					if (rServer == NULL) rServer = Server;
+					UdmVarListReplaceLst(&rDoc.Sections, &rServer->Vars, NULL, "*");
+					UdmDocAddServExtraHeaders(rServer, &rDoc);
 					UdmVarListLog(Indexer,&rDoc.RequestHeaders,UDM_LOG_DEBUG,"Request");
 					
 					UDM_THREADINFO(Indexer,"Getting",rurl);
@@ -1136,13 +1292,18 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 					UdmVarListLog(Indexer,&rDoc.Sections,UDM_LOG_DEBUG,"Response");
 					UDM_FREE(rurl);
 
-					if ((status = UdmVarListFindInt(&rDoc.Sections, "Status", 0)) == 200 ) {
+					if ((status = UdmVarListFindInt(&rDoc.Sections, "Status", 0)) == UDM_HTTP_STATUS_OK) {
 					
 					        UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
-						result = UdmRobotParse(Indexer, rDoc.Buf.content, UDM_NULL2EMPTY(rDoc.CurURL.hostinfo));
+						result = UdmRobotParse(Indexer, rServer, rDoc.Buf.content, 
+								       UDM_NULL2EMPTY(rDoc.CurURL.hostinfo));
 						UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
 					} else if (status >= 500) {
 						Doc.method = UDM_METHOD_VISITLATER;
+					} else {
+					        UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
+						UdmRobotAddEmpty(&Indexer->Conf->Robots, UDM_NULL2EMPTY(Doc.CurURL.hostinfo));
+						UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
 					}
 
 					UdmDocFree(&rDoc);
@@ -1172,6 +1333,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 	if(result!=UDM_OK){
 	        UDM_FREE(origurl); UDM_FREE(aliasurl);
 		UdmDocFree(&Doc);
+#ifdef USE_PARANOIA
+		UdmViolationExit(paran);
+#endif
 		return result;
 	}
 
@@ -1224,6 +1388,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 			if(result!=UDM_OK){
 			        UDM_FREE(origurl); UDM_FREE(aliasurl);
 				UdmDocFree(&Doc);
+#ifdef USE_PARANOIA
+				UdmViolationExit(paran);
+#endif
 				return result;
 			}
 			
@@ -1254,6 +1421,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 		if(result!=UDM_OK){
 		        UDM_FREE(origurl); UDM_FREE(aliasurl);
 			UdmDocFree(&Doc);
+#ifdef USE_PARANOIA
+			UdmViolationExit(paran);
+#endif
 			return result;
 		}
 		
@@ -1261,8 +1431,8 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 		Indexer->nbytes+=Doc.Buf.size;
 		Indexer->ndocs++;
 		
-		if((!Doc.Buf.content) && (status<500)) {
-			UdmLog(Indexer,UDM_LOG_ERROR,"Illegal HTTP headers in response");
+		if((!Doc.Buf.content) && (status < 500)) {
+			UdmLog(Indexer, UDM_LOG_ERROR, "No data received");
 			status=UDM_HTTP_STATUS_SERVICE_UNAVAILABLE;
 			UdmVarListReplaceInt(&Doc.Sections,"Status",status);
 		}
@@ -1277,6 +1447,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 			if(result!=UDM_OK){
 			        UDM_FREE(origurl); UDM_FREE(aliasurl);
 				UdmDocFree(&Doc);
+#ifdef USE_PARANOIA
+				UdmViolationExit(paran);
+#endif
 				return result;
 			}
 			
@@ -1303,8 +1476,47 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 			
 			UdmParseURLText(Indexer,&Doc);
 			UdmParseHeaders(Indexer,&Doc);
-			if (Doc.method!=UDM_METHOD_HREFONLY) {
-				UdmPrepareWords(Indexer,&Doc);
+
+			if (Doc.method != UDM_METHOD_HREFONLY)
+			{
+				size_t nfilters;
+				
+				UDM_GETLOCK(Indexer,UDM_LOCK_CONF);
+				nfilters= Indexer->Conf->SectionFilters.nmatches;
+				UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
+				
+				if (nfilters)
+				{
+					UDM_VARLIST TmpList;
+					int sec_method;
+					
+					/* Save sections into TmpList */
+					UdmVarListInit(&TmpList);
+					UdmVarListReplaceLst(&TmpList,&Doc.Sections,NULL,"*");
+					UdmPrepareWords(Indexer,&Doc);
+					
+					UDM_GETLOCK(Indexer,UDM_LOCK_CONF);
+					sec_method= UdmSectionFilterFind(&Indexer->Conf->SectionFilters, &Doc, reason);
+					UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
+					
+					if (sec_method != UDM_METHOD_INDEX)
+					{
+						/* Restore saved sections */
+						UdmVarListFree(&Doc.Sections);
+						UdmVarListReplaceLst(&Doc.Sections,&TmpList,NULL,"*");
+						
+						/* Remove unnesessary data */
+						UdmVarListDel(&Doc.Sections, "CachedCopy");
+						UdmWordListFree(&Doc.Words);
+						UdmCrossListFree(&Doc.CrossWords);
+					}
+					UdmVarListFree(&TmpList);
+					UdmLog(Indexer,UDM_LOG_DEBUG,"SectionFilter: %s",reason);
+				}
+				else
+				{
+					UdmPrepareWords(Indexer,&Doc);
+				}
 			}
 			
 			/* Remove StopWords */
@@ -1333,6 +1545,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 			}
 			UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
 			if(UDM_OK != (result = UdmURLAction(Indexer, &Doc, UDM_URL_ACTION_LINKS_DELETE))) {
+#ifdef USE_PARANOIA
+			  UdmViolationExit(paran);
+#endif
 			  return result;
 			}
 		}
@@ -1346,6 +1561,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 	if(result!=UDM_OK){
 	        UDM_FREE(origurl); UDM_FREE(aliasurl);
 		UdmDocFree(&Doc);
+#ifdef USE_PARANOIA
+		UdmViolationExit(paran);
+#endif
 		return result;
 	}
 	
@@ -1362,5 +1580,8 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 	UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
 	if (result == UDM_OK) result = Indexer->action;
 	UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
 	return result;
 }
