@@ -15,10 +15,6 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */
 
-#ifdef WIN32
-#include "stdafx.h"
-#endif
-
 #include "udm_config.h"
 
 #include <stdio.h>
@@ -28,6 +24,11 @@
 #include <sys/types.h>
 #include <fcntl.h>
 #include <signal.h>
+#include <assert.h>
+
+#ifdef WIN32
+#include <process.h>
+#endif
 
 #ifdef HAVE_READLINE
 #include <readline/readline.h>
@@ -83,7 +84,7 @@ UDM_AGENT **ThreadIndexers;
 HANDLE *threads;
 
 /* CallBack function for Thread information */
-extern __C_LINK void __UDMCALL UdmThreadProc(UDM_AGENT* A, const char *state, const char* str);
+extern __C_LINK void __UDMCALL UdmShowInfo(UDM_AGENT* A, const char *state, const char* str);
 
 #else
 
@@ -93,180 +94,74 @@ pthread_t *threads;
 
 #endif
 
-enum udm_indcmd
-{
-  UDM_IND_INDEX,
-  UDM_IND_STAT,
-  UDM_IND_CREATE,
-  UDM_IND_DROP,
-  UDM_IND_DELETE,
-  UDM_IND_REFERERS,
-  UDM_IND_SQLMON,
-  UDM_IND_CHECKCONF
-};
 
-#define UDM_DISPLAY_FIELDS    1
-
-static int UdmDisplaySQLQuery(UDM_AGENT *A, UDM_DB *db,const char *q, FILE *outfile, int dflags) {
+static int UdmDisplaySQLQuery(UDM_SQLMON_PARAM *prm, UDM_SQLRES *sqlres) {
      int       res = UDM_OK;
 #ifdef HAVE_SQL     
-     UDM_SQLRES     sqlres;
      size_t         i,j;
 
-     bzero((void*)&sqlres,sizeof(sqlres));
-     UDM_GETLOCK(A, UDM_LOCK_DB);
-     res=UdmSQLQuery(db,&sqlres,q);
-     
-     if (res!=UDM_OK){
-          fprintf(outfile,"%s\n",db->errstr);
-          UDM_RELEASELOCK(A, UDM_LOCK_DB);
-          return res;
-     }
-     
-     if (dflags & UDM_DISPLAY_FIELDS)
+     if (prm->flags & UDM_SQLMON_DISPLAY_FIELDS)
      {
-          for (i=0;i<sqlres.nCols;i++){
-               if(i>0)fprintf(outfile,"\t");
-               fprintf(outfile,"%s",sqlres.Fields ? sqlres.Fields[i].sqlname : "<NONAME>");
-               if(i+1==sqlres.nCols)fprintf(outfile,"\n");
+          for (i=0;i<sqlres->nCols;i++){
+               if(i>0)fprintf(prm->outfile,"\t");
+               fprintf(prm->outfile,"%s",sqlres->Fields ? sqlres->Fields[i].sqlname : "<NONAME>");
+               if(i+1==sqlres->nCols)fprintf(prm->outfile,"\n");
           }
      }
      
-     for (i=0;i<sqlres.nRows;i++){
-          for(j=0;j<sqlres.nCols;j++){
-               const char *v=UdmSQLValue(&sqlres,i,j);
-               if(j>0)fprintf(outfile,"\t");
-               fprintf(outfile,"%s",v?v:NULL);
-               if(j+1==sqlres.nCols)fprintf(outfile,"\n");
+     for (i=0;i<sqlres->nRows;i++){
+          for(j=0;j<sqlres->nCols;j++){
+               const char *v=UdmSQLValue(sqlres,i,j);
+               if(j>0)fprintf(prm->outfile,"\t");
+               if (j < 10 && (prm->colflags[j] & 1))
+               {
+                 const char* s;
+                 fprintf(prm->outfile,"0x");
+                 for (s=v; s[0]; s++)
+                   fprintf(prm->outfile,"%02X",(int)(unsigned char)s[0]);
+               }
+               else
+               {
+                 fprintf(prm->outfile,"%s",v?v:"NULL");
+               }
+               if(j+1==sqlres->nCols)fprintf(prm->outfile,"\n");
           }
      }
-     UdmSQLFree(&sqlres);
-     UDM_RELEASELOCK(A, UDM_LOCK_DB);
+     
 #endif
      return res;
 }
 
-
-
-static int SQLMonitor(UDM_AGENT *A, UDM_ENV *Env, FILE *infile, FILE *outfile){
-     char str[10*1024];
-     char *snd=str;
-     int  fields= UDM_DISPLAY_FIELDS;
-     
-     str[sizeof(str)-1]='\0';
-     while(1){
-          int  sqlres;
-          int  exec=0;
-          char *send0;
-          size_t  rbytes= (sizeof(str)-1) - (snd - str);
-          
-#ifdef HAVE_READLINE
-          if ((infile == stdin) && isatty(0))
-          {
-               char prompt[]="SQL>";
-               char *line= readline(prompt);
-               if (!line)break;
-               if (*line) add_history(line);
-               strncpy(snd, line, rbytes);
-          }
-          else
-#endif
-          {
-               fprintf(outfile,"SQL>");
-               if(!fgets(snd, rbytes, infile))
-                    break;
-          }
-          
-          if(snd[0]=='#' || !strncmp(snd, "--", 2)){
-               continue;
-          }
-          send0 = snd;
-          snd = snd + strlen(snd);
-          while(snd >= send0 && strchr(" \r\n\t", snd[-1])){
-               *--snd='\0';
-          }
-          
-          if(snd[-1]==';'){
-               exec=1;
-               *--snd='\0';
-          }else
-          if(snd - 2 >= str && snd[-1]=='g' && snd[-2]=='\\'){
-               exec=1;
-               snd-=2;
-               *snd='\0';
-          }else
-          if(snd-2 >= str && strchr("oO", snd[-1]) && strchr("gG",snd[-2])){
-               exec=1;
-               snd-=2;
-               *snd='\0';
-          }else
-          if((size_t)(snd - str + 1) >= sizeof(str)){
-               exec=1;
-          }
-          
-          if(exec){
-               printf("'%s'\n",str);
-               if (!strncasecmp(str,"connection",10)){
-                    size_t newnum= atoi(str+10);
-                    if (newnum < Env->dbl.nitems){
-                         Env->dbl.currdbnum= newnum;
-                         printf("Connection changed to #%d\n",Env->dbl.currdbnum);
-                    }else{
-                         printf("Error: wrong connection number %d\n",newnum);
-                    }
-               }
-               else if (!strcasecmp(str,"fields=off")){
-                    fields=0;
-               }
-               else if (!strcasecmp(str,"fields=on")){
-                    fields=1;
-               }
-               else{
-                    sqlres=UdmDisplaySQLQuery(A, &Env->dbl.db[Env->dbl.currdbnum],str,outfile,fields);
-               }
-               snd=str;
-               str[0]='\0';
-          }else{
-               if(send0!=snd){
-                    *snd=' ';
-                    snd++;
-                    *snd='\0';
-               }
-          }
-     }
-     printf("\n");
-     return UDM_OK;
-}
-
-
-static const char* UdmDBTypeStr(int dbtype)
+static char* sqlmongets(UDM_SQLMON_PARAM *prm, char *str, size_t size)
 {
-  switch(dbtype)
+#ifdef HAVE_READLINE
+  if ((prm->infile == stdin) && isatty(0))
   {
-    case UDM_DB_MYSQL:   return "mysql";
-    case UDM_DB_PGSQL:   return "pgsql";
-    case UDM_DB_IBASE:   return "ibase";
-    case UDM_DB_MSSQL:   return "mssql";
-    case UDM_DB_ORACLE8: return "oracle";
-    case UDM_DB_SQLITE:  return "sqlite";
+     char prompt[]="SQL>";
+     char *line= readline(prompt);
+     if (!line)
+       return 0;
+     
+     if (*line) add_history(line);
+     strncpy(str, line, size);
   }
-  return "unknown_dbtype";
+  else
+#endif
+  {
+    prm->prompt(prm, UDM_SQLMON_MSG_PROMPT, "SQL>");
+    if(!fgets(str, size, prm->infile))
+      return 0;
+  }
+  return str;
 }
 
-static const char* UdmDBModeStr(int dbmode) {
-  switch(dbmode) {
-  case UDM_DBMODE_SINGLE:
-  case UDM_DBMODE_CACHE:
-    return "single";
-  case UDM_DBMODE_SINGLE_CRC: return "crc";
-  case UDM_DBMODE_MULTI:  return "multi";
-  case UDM_DBMODE_MULTI_CRC:  return "crc-multi";
-  }
-  return "unknown_dbmode";
+static int sqlmonprompt(UDM_SQLMON_PARAM *prm, int msqtype, const char *msg)
+{
+  fprintf(prm->outfile,"%s",msg);
+  return UDM_OK;
 }
 
-
-static const char* UdmIndCmdStr(enum udm_indcmd cmd)
+__C_LINK /*static*/ const char* __UDMCALL UdmIndCmdStr(enum udm_indcmd cmd)
 {
   switch(cmd)
   {
@@ -285,14 +180,15 @@ static int CreateOrDrop(UDM_AGENT *A, enum udm_indcmd cmd)
   char fname[1024];
   const char *sdir=UdmVarListFindStr(&Conf.Vars,"ShareDir",UDM_SHARE_DIR);
   UDM_DBLIST *L= &A->Conf->dbl;
+  UDM_SQLMON_PARAM prm;
   
   for (i=0; i<L->nitems; i++)
   {
     FILE *infile;
     UDM_DB *db= &L->db[i];
     udm_snprintf(fname,sizeof(fname),"%s%s%s%s%s.%s.sql",
-      sdir,UDMSLASHSTR, UdmDBTypeStr(db->DBType),UDMSLASHSTR,
-      UdmIndCmdStr(cmd),UdmDBModeStr(db->DBMode));
+      sdir,UDMSLASHSTR, UdmDBTypeToStr(db->DBType),UDMSLASHSTR,
+      UdmIndCmdStr(cmd),UdmDBModeToStr(db->DBMode));
     printf("'%s' dbtype=%d dbmode=%d\n",fname,db->DBType,db->DBMode);
     if(!(infile= fopen(fname,"r")))
     {
@@ -300,7 +196,16 @@ static int CreateOrDrop(UDM_AGENT *A, enum udm_indcmd cmd)
       return UDM_ERROR;
     }
     L->currdbnum= i;
-    SQLMonitor(A, A->Conf,infile,stdout);
+    bzero((void*)&prm,sizeof(prm));
+    prm.infile= infile;
+    prm.outfile= stdout;
+    prm.flags= UDM_SQLMON_DISPLAY_FIELDS;
+    prm.gets= sqlmongets;
+    prm.display= UdmDisplaySQLQuery;
+    prm.prompt= sqlmonprompt;
+    UdmSQLMonitor(A, A->Conf,&prm);
+    printf("%d queries sent, %d succeeded, %d failed\n",
+      prm.nqueries, prm.ngood, prm.nbad);
     fclose(infile);
   }
   return UDM_OK;
@@ -308,13 +213,39 @@ static int CreateOrDrop(UDM_AGENT *A, enum udm_indcmd cmd)
 
 static int ShowStatistics(UDM_AGENT *Indexer){
      int       res;
+     struct tm tm;
+     const char *stat_time;
+     char sbuf[32];
      UDM_STATLIST   Stats;
      size_t         snum;
      UDM_STAT  Total;
      
      bzero((void*)&Total, sizeof(Total));
+     Stats.time = time(NULL);
+     stat_time = UdmVarListFindStr(&Conf.Vars, "stat_time", "0");
+     bzero(&tm, sizeof(tm));
+
+     if (stat_time && (
+	    strptime(stat_time, "%Y-%m-%d %H:%M:%S", &tm) ||
+	    strptime(stat_time, "%Y-%m-%d %H:%M", &tm) ||
+	    strptime(stat_time, "%Y-%m-%d %H:%M", &tm) ||
+	    strptime(stat_time, "%Y-%m-%d %H", &tm) ||
+	    strptime(stat_time, "%Y-%m-%d", &tm) ||
+	    strptime(stat_time, "%Y-%m", &tm))) {
+
+    	    Stats.time = mktime(&tm);
+     } else if (stat_time && (Stats.time = Udm_dp2time_t(stat_time)) >= 0) {
+    	    Stats.time += time(NULL);
+	    localtime_r(&Stats.time, &tm);
+     } else {
+    	    Stats.time = time(NULL);
+	    localtime_r(&Stats.time, &tm);
+     }
+
+     strftime(sbuf, sizeof(sbuf), "%Y-%m-%d %H:%M:%S", &tm);
      res=UdmStatAction(Indexer,&Stats);
-     printf("\n          Database statistics\n\n");
+
+     printf("\n          Database statistics [%s]\n\n", sbuf);
      printf("%10s %10s %10s\n","Status","Expired","Total");
      printf("   -----------------------------\n");
      for(snum=0;snum<Stats.nstats;snum++){
@@ -345,7 +276,7 @@ int res;
 #undef THINFO_TEST
 #ifdef THINFO_TEST
 /* CallBack function for Thread information */
-/*void UdmThreadProc(int handle,char *state, char* str){
+/*void UdmShowInfo(int handle,char *state, char* str){
      printf("%d %s %s\n",handle,state,str);
 }*/
 #endif
@@ -391,11 +322,6 @@ static void UdmFeatures(UDM_VARLIST *V){
 #else
      UdmVarListReplaceStr(V,"HAVE_PTHREAD","no");
 #endif
-#ifdef USE_XML
-     UdmVarListReplaceStr(V,"USE_XML","yes");
-#else
-     UdmVarListReplaceStr(V,"USE_XML","no");
-#endif
 #ifdef USE_HTTPS
      UdmVarListReplaceStr(V,"USE_HTTPS","yes");
 #else
@@ -406,6 +332,11 @@ static void UdmFeatures(UDM_VARLIST *V){
 #else
      UdmVarListReplaceStr(V,"DMALLOC","no");
 #endif
+#ifdef EFENCE
+     UdmVarListReplaceStr(V,"EFENCE","yes");
+#else
+     UdmVarListReplaceStr(V,"EFENCE","no");
+#endif
 #ifdef CHASEN
      UdmVarListReplaceStr(V,"CHASEN","yes");
 #else
@@ -466,11 +397,6 @@ static void UdmFeatures(UDM_VARLIST *V){
 #else
      UdmVarListReplaceStr(V,"HAVE_PGSQL","no");
 #endif
-#ifdef HAVE_MSQL
-     UdmVarListReplaceStr(V,"HAVE_MSQL","yes");
-#else
-     UdmVarListReplaceStr(V,"HAVE_MSQL","no");
-#endif
 #ifdef HAVE_IODBC
      UdmVarListReplaceStr(V,"HAVE_IODBC","yes");
 #else
@@ -516,11 +442,6 @@ static void UdmFeatures(UDM_VARLIST *V){
 #else
      UdmVarListReplaceStr(V,"HAVE_CTLIB","no");
 #endif
-#ifdef HAVE_ORACLE7
-     UdmVarListReplaceStr(V,"HAVE_ORACLE7","yes");
-#else
-     UdmVarListReplaceStr(V,"HAVE_ORACLE7","no");
-#endif
 #ifdef HAVE_ORACLE8
      UdmVarListReplaceStr(V,"HAVE_ORACLE8","yes");
 #else
@@ -571,7 +492,7 @@ static int usage(int level){
      fprintf(stderr,
 "\n\
 indexer from %s-%s-%s\n\
-http://www.mnogosearch.org/ (C) 1998-2002, LavTech Corp.\n\
+http://www.mnogosearch.org/ (C)1998-2004, LavTech Corp.\n\
 \n\
 Usage: indexer [OPTIONS]  [configfile]\n\
 \n\
@@ -641,10 +562,13 @@ Misc. options:\n\
 
 "  -C              clear database and exit\n\
   -S              print statistics and exit\n\
+  -j t            set current time for statistic (use with -S),\n\
+                  YYYY-MM[-DD[ HH[:MM[:SS]]]]\n\
+	          or time offset, e.g. 1d12h (see Period in indexer.conf)\n\
   -I              print referers and exit\n\
   -R              calculate popularity rank\n\
-  -Y              check-up stored database at exit\n\
-  -Z              check-up cached database at exit\n\
+  -Ecreate        create SQL table structure and exit\n\
+  -Edrop          drop SQL table structure and exit\n\
 "
 #endif
 "  -h,-?           print help page and exit\n\
@@ -676,18 +600,107 @@ static int UdmIndexerEnvLoad(UDM_AGENT *Indexer, const char *fname,int lflags){
      return rc;
 }
 
+static enum udm_indcmd UdmIndCmd(const char *cmd) {
+  if (!cmd)return UDM_IND_INDEX;
+  if (!strncasecmp(cmd,"ind",3))return UDM_IND_INDEX;
+  if (!strncasecmp(cmd,"sta",3))return UDM_IND_STAT;
+  if (!strncasecmp(cmd,"cre",3))return UDM_IND_CREATE;
+  if (!strncasecmp(cmd,"dro",3))return UDM_IND_DROP;
+  if (!strncasecmp(cmd,"del",3))return UDM_IND_DELETE;
+  if (!strncasecmp(cmd,"ref",3))return UDM_IND_REFERERS;
+  if (!strncasecmp(cmd,"sql",3))return UDM_IND_SQLMON;
+  if (!strncasecmp(cmd,"che",3))return UDM_IND_CHECKCONF;
+  if (!strncasecmp(cmd,"blo",3))return UDM_IND_MULTI2BLOB;
+  return UDM_IND_INDEX;
+}
+
+/*
+  Parse command line
+*/
+static int UdmARGC;
+static char **UdmARGV;
+static enum udm_indcmd cmd = UDM_IND_INDEX;
+static int insert = 0, expire = 0, pop_rank = 0, mkind = 0, block = 0, help = 0;
+static char *url_filename=NULL;
+
+static void UdmParseCmdLine(void) {
+  int ch;
+
+     while ((ch = getopt(UdmARGC, UdmARGV, "QUCSIRMabheorldmqiw?E:F:t:u:s:n:v:L:A:D:p:N:f:c:g:y:j:")) != -1){
+          switch (ch) {
+          case 'F': {
+               UDM_VARLIST V,W;
+               size_t i;
+               
+               UdmVarListInit(&V);
+               UdmVarListInit(&W);
+               UdmFeatures(&V);
+               UdmVarListAddLst(&W,&V,NULL,optarg);
+               for(i=0;i<W.nvars;i++)
+                    printf("%s:%s\n",W.Var[i].name,W.Var[i].val);
+               exit(0);
+          }
+          case 'C': cmd= UDM_IND_DELETE;  add_servers=0;load_langmaps=0;load_spells=0;break;
+          case 'S': cmd= UDM_IND_STAT;    add_servers=0;load_langmaps=0;load_spells=0;break;
+          case 'I': cmd= UDM_IND_REFERERS;add_servers=0;load_langmaps=0;load_spells=0;break;
+          case 'Q': cmd= UDM_IND_SQLMON;  add_servers=0;load_langmaps=0;load_spells=0;break;
+          case 'E': cmd= UdmIndCmd(optarg);break;
+          case 'R': pop_rank++; break;
+          case 'M': mkind=1;break;
+          case 'q': add_server_urls = 0; break;
+          case 'l': log2stderr=0;break;
+          case 'a': expire=1;break;
+          case 'b': block++;break;
+          case 'e': flags|=UDM_FLAG_SORT_EXPIRED;break;
+          case 'o': flags|=UDM_FLAG_SORT_HOPS;break;
+          case 'r': flags|=UDM_FLAG_DONTSORT_SEED; break;
+          case 'm': flags|=UDM_FLAG_REINDEX;break;
+          case 'n': Conf.url_number=atoi(optarg);break;
+          case 'c': max_index_time=atoi(optarg);break;
+          case 'v': UdmSetLogLevel(NULL, atoi(optarg)); break;
+          case 'p': seconds=atoi(optarg);break;
+          case 't': UdmVarListAddStr(&Conf.Vars,"tag" , optarg);break;
+          case 'g': UdmVarListAddStr(&Conf.Vars,"cat" , optarg);break;
+          case 's': UdmVarListAddStr(&Conf.Vars, "status", optarg);break;
+          case 'y': UdmVarListAddStr(&Conf.Vars,"type", optarg);break;
+          case 'L': UdmVarListAddStr(&Conf.Vars,"lang", optarg);break;
+          case 'u': UdmVarListAddStr(&Conf.Vars,"u"   , optarg);
+               if(insert){
+                    UDM_HREF Href;
+                    UdmHrefInit(&Href);
+                    Href.url=optarg;
+                    Href.method=UDM_METHOD_GET;
+                    UdmHrefListAdd(&Main.Hrefs, &Href);
+               }
+               break;
+          case 'N': maxthreads=atoi(optarg);break;
+          case 'f': url_filename=optarg;break;
+          case 'i': insert=1;break;
+          case 'w': warnings=0;break;
+	  case 'j': UdmVarListAddStr(&Conf.Vars, "stat_time", optarg); break;
+          case '?':
+          case 'h':
+          default:
+               help++;
+          }
+     }
+
+}
+
+
+
+
 
 #ifdef  WIN32
 unsigned int __stdcall thread_main(void *arg) {
+	char *str_buf;
 #else
 static void * thread_main(void *arg){
 #endif
      UDM_AGENT * Indexer = (UDM_AGENT *)arg;
-     UDM_ENV *Env = Indexer->Conf;
      int res=UDM_OK;
      int done=0;
      int i_sleep=0;
-	 char * str_buf;
      
      while(!done){
      
@@ -699,6 +712,7 @@ static void * thread_main(void *arg){
                     break;
           }
 
+	  UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
           if(have_sighup){
                UDM_ENV   NewConf;
                int  rc;
@@ -712,6 +726,7 @@ static void * thread_main(void *arg){
 	       UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
                Indexer->Conf = &NewConf;
                rc = UdmIndexerEnvLoad(Indexer, cname, add_servers + load_langmaps + UDM_FLAG_SPELL);
+               Indexer->Conf = &Conf;
                
                if(rc!=UDM_OK){
 		    UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
@@ -721,7 +736,11 @@ static void * thread_main(void *arg){
                }else{
                     UdmEnvFree(&Conf);
                     Conf=NewConf;
-                    UdmOpenLog("indexer",&Conf, log2stderr);
+		    UdmParseCmdLine();
+#ifdef WIN32
+#else
+                    UdmOpenLog("indexer", &Conf, log2stderr);
+#endif
 		    UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
               }
 
@@ -730,11 +749,9 @@ static void * thread_main(void *arg){
 	  if (have_sigint || have_sigterm) {
 	    int z;
             UdmLog(Indexer, UDM_LOG_ERROR, "%s received. Terminating. Please wait...", (have_sigint) ? "SIGINT" : "SIGTERM");
-	    UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
 	    for (z = 0 ; z < total_threads; z++)
 	      if (ThreadIndexers[z]) UdmAgentSetAction(ThreadIndexers[z], UDM_TERMINATED);
 	    UdmAgentSetAction(&Main, UDM_TERMINATED);
-	    UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
 	    have_sigint = have_sigterm = 0;
 	  }
 
@@ -746,6 +763,7 @@ static void * thread_main(void *arg){
 	    UdmDecLogLevel(Indexer);
 	    have_sigusr2 = 0;
 	  }
+	  UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
 
           if(done)break;
           
@@ -794,15 +812,15 @@ static void * thread_main(void *arg){
 #endif
                case UDM_ERROR:
 #ifdef WIN32
-			str_buf = (char *)malloc(1024);
+			str_buf = (char *)UdmMalloc(1024);
 			udm_snprintf(str_buf, 1024, "Error: %s",  UdmEnvErrMsg(Indexer->Conf));
-		    UdmThreadProc(0,NULL, str_buf);
+		    UdmShowInfo(0,NULL, str_buf);
 			UDM_FREE(str_buf);
-		    UdmThreadProc(Indexer, "Error", UdmEnvErrMsg(Indexer->Conf));
+		    UdmShowInfo(Indexer, "Error", UdmEnvErrMsg(Indexer->Conf));
 #endif
 	       case UDM_TERMINATED:
 #ifdef WIN32
-		    UdmThreadProc(Indexer, "Aborted", "");
+		    UdmShowInfo(Indexer, "Aborted", "");
 #endif
 
                default:
@@ -857,6 +875,7 @@ static void * thread_main(void *arg){
 	  int z;
 
           UdmURLAction(Indexer, NULL, UDM_URL_ACTION_FLUSH); /* flush DocCache */
+	  UdmMultiCacheFlush(Indexer);
 
           time(&now);
           sec = now - Indexer->start_time - Indexer->nsleepsecs;
@@ -866,7 +885,7 @@ static void * thread_main(void *arg){
           }
           UdmLog(Indexer,UDM_LOG_ERROR,"Done (%d seconds, %u documents, %u bytes, %5.2f %cbytes/sec.)",
                sec, Indexer->ndocs, Indexer->nbytes, (M < 1.0) ? K : M, (M < 1.0) ? 'K' : 'M' );
-#ifndef WIN32
+#if !defined(WIN32) && defined(HAVE_PTHREAD)
 		  for (z = 0 ; z < total_threads; z++)
 	      if (ThreadIndexers[z])  pthread_kill(threads[z], SIGALRM); /* wake-up sleeping threads */
 #endif
@@ -937,7 +956,12 @@ static int UdmClear(UDM_AGENT *A, const char *url_fname)
      if(warnings) {
           size_t i;
           printf("You are going to delete content from database(s):\n");
-          for (i = 0; i < Conf.dbl.nitems; i++) printf("%s\n", Conf.dbl.db[i].DBADDR);
+          for (i = 0; i < Conf.dbl.nitems; i++)
+          {
+            const char *dbaddr;
+            dbaddr= UdmVarListFindStr(&Conf.dbl.db[i].Vars,"DBAddr","<noaddr>");
+            printf("%s\n", dbaddr);
+          }
           clear_confirmed=UdmConfirm("Are you sure?(YES/no)");
      }
      
@@ -961,20 +985,21 @@ static int UdmClear(UDM_AGENT *A, const char *url_fname)
 }
 
 
-static int UdmIndex(UDM_AGENT *A)
-{
-	int int_res;	// Temporary variable used to check return result
+static int UdmIndex(UDM_AGENT *A) {
+#ifdef WIN32
+	int int_res;	/* Temporary variable used to check return result */
+#endif
 
-	ThreadIndexers = (UDM_AGENT**)malloc(maxthreads * sizeof(UDM_AGENT*));
+	ThreadIndexers = (UDM_AGENT**)UdmMalloc(maxthreads * sizeof(UDM_AGENT*));
                
 #ifdef HAVE_PTHREAD
      {
           int i;
 
 #ifdef WIN32
-          threads = (HANDLE*)malloc(maxthreads * sizeof(HANDLE));
+          threads = (HANDLE*)UdmMalloc(maxthreads * sizeof(HANDLE));
 #else
-          threads = (pthread_t*)malloc(maxthreads * sizeof(pthread_t));
+          threads = (pthread_t*)UdmMalloc(maxthreads * sizeof(pthread_t));
 #endif
                
           for(i=0;i<maxthreads;i++){
@@ -995,7 +1020,7 @@ static int UdmIndex(UDM_AGENT *A)
 #else
                {
                     pthread_attr_t attr;
-                    size_t stksize=1024*256;
+                    size_t stksize = 1024 * 512;
                     
                     pthread_attr_init(&attr);
                     pthread_attr_setstacksize(&attr, stksize);
@@ -1007,7 +1032,9 @@ static int UdmIndex(UDM_AGENT *A)
 	       total_threads = i + 1;
 	       UDM_RELEASELOCK(A,UDM_LOCK_THREAD);
           }
-          
+#ifndef WIN32        
+	  for (i = 0; i < maxthreads; i++) pthread_join(threads[i], NULL);
+#else
           while(1){
                int num;
                UDM_GETLOCK(A,UDM_LOCK_THREAD);
@@ -1016,9 +1043,11 @@ static int UdmIndex(UDM_AGENT *A)
                if(!num)break;
                UDMSLEEP(1);
           }
+#endif
           UDM_FREE(threads);
      }
 #else
+     Main.handle = 1;
      thread_main(&Main);
 #endif
 
@@ -1027,31 +1056,12 @@ static int UdmIndex(UDM_AGENT *A)
      return UDM_OK;
 }
 
-static enum udm_indcmd UdmIndCmd(const char *cmd)
-{
-  if (!cmd)return UDM_IND_INDEX;
-  if (!strncasecmp(cmd,"ind",3))return UDM_IND_INDEX;
-  if (!strncasecmp(cmd,"sta",3))return UDM_IND_STAT;
-  if (!strncasecmp(cmd,"cre",3))return UDM_IND_CREATE;
-  if (!strncasecmp(cmd,"dro",3))return UDM_IND_DROP;
-  if (!strncasecmp(cmd,"del",3))return UDM_IND_DELETE;
-  if (!strncasecmp(cmd,"ref",3))return UDM_IND_REFERERS;
-  if (!strncasecmp(cmd,"sql",3))return UDM_IND_SQLMON;
-  if (!strncasecmp(cmd,"che",3))return UDM_IND_CHECKCONF;
-  return UDM_IND_INDEX;
-}
-
 #ifndef WIN32
 
 int main(int argc, char **argv) {
      char      *language=NULL,*affix=NULL,*dictionary=NULL, *env;
-     int       insert = 0, expire = 0, pop_rank = 0, check_cached = 0, check_stored = 0;
-     int       mkind=0, block=0;
-     int       ch, pid_fd;
-     int       help=0;
-     char      *url_fname=NULL;
+     int       pid_fd;
      char      pidbuf[1024];
-     enum udm_indcmd cmd= UDM_IND_INDEX;
 #ifdef CHASEN
      char            *chasen_argv[] = { "chasen", "-b", "-f", "-F", "%m ", NULL };
      chasen_getopt_argv(chasen_argv, NULL);
@@ -1066,69 +1076,14 @@ int main(int argc, char **argv) {
      UdmSetLockProc(&Conf,UdmLockProc);
      UdmSetRefProc(&Conf,UdmRefProc);
 #ifdef THINFO_TEST
-     UdmSetThreadProc(&Conf,UdmThreadProc);
+     UdmSetThreadProc(&Conf,UdmShowInfo);
 #endif
      UdmAgentInit(&Main,&Conf,0);
      
-     while ((ch = getopt(argc, argv, "QUCSIRMYZabheorlmdqiw?E:F:t:u:s:n:v:L:A:D:p:N:f:c:g:y:")) != -1){
-          switch (ch) {
-          case 'F': {
-               UDM_VARLIST V,W;
-               size_t i;
-               
-               UdmVarListInit(&V);
-               UdmVarListInit(&W);
-               UdmFeatures(&V);
-               UdmVarListAddLst(&W,&V,NULL,optarg);
-               for(i=0;i<W.nvars;i++)
-                    printf("%s:%s\n",W.Var[i].name,W.Var[i].val);
-               exit(0);
-          }
-          case 'C': cmd= UDM_IND_DELETE;  add_servers=0;load_langmaps=0;load_spells=0;break;
-          case 'S': cmd= UDM_IND_STAT;    add_servers=0;load_langmaps=0;load_spells=0;break;
-          case 'I': cmd= UDM_IND_REFERERS;add_servers=0;load_langmaps=0;load_spells=0;break;
-          case 'Q': cmd= UDM_IND_SQLMON;  add_servers=0;load_langmaps=0;load_spells=0;break;
-          case 'E': cmd= UdmIndCmd(optarg);break;
-          case 'R': pop_rank++; break;
-          case 'Y': check_stored++; break;
-          case 'Z': check_cached++; break;
-          case 'M': mkind=1;break;
-          case 'q': add_server_urls = 0; break;
-          case 'l': log2stderr=0;break;
-          case 'a': expire=1;break;
-          case 'b': block++;break;
-          case 'e': flags|=UDM_FLAG_SORT_EXPIRED;break;
-          case 'o': flags|=UDM_FLAG_SORT_HOPS;break;
-          case 'r': flags|=UDM_FLAG_DONTSORT_SEED; break;
-          case 'm': flags|=UDM_FLAG_REINDEX;break;
-          case 'n': Conf.url_number=atoi(optarg);break;
-          case 'c': max_index_time=atoi(optarg);break;
-          case 'v': UdmSetLogLevel(NULL, atoi(optarg)); break;
-          case 'p': seconds=atoi(optarg);break;
-          case 't': UdmVarListAddStr(&Conf.Vars,"tag" , optarg);break;
-          case 'g': UdmVarListAddStr(&Conf.Vars,"cat" , optarg);break;
-          case 's': UdmVarListAddStr(&Conf.Vars, "status", optarg);break;
-          case 'y': UdmVarListAddStr(&Conf.Vars,"type", optarg);break;
-          case 'L': UdmVarListAddStr(&Conf.Vars,"lang", optarg);break;
-          case 'u': UdmVarListAddStr(&Conf.Vars,"u"   , optarg);
-               if(insert){
-                    UDM_HREF Href;
-                    UdmHrefInit(&Href);
-                    Href.url=optarg;
-                    Href.method=UDM_METHOD_GET;
-                    UdmHrefListAdd(&Main.Hrefs, &Href);
-               }
-               break;
-          case 'N': maxthreads=atoi(optarg);break;
-          case 'f': url_fname=optarg;break;
-          case 'i': insert=1;break;
-          case 'w': warnings=0;break;
-          case '?':
-          case 'h':
-          default:
-               help++;
-          }
-     }
+     UdmARGC = argc;
+     UdmARGV = argv;
+
+     UdmParseCmdLine();
      
      if (cmd != UDM_IND_INDEX)
      {
@@ -1138,7 +1093,9 @@ int main(int argc, char **argv) {
      }
      
      flags|=add_servers;
-     
+     flags |= add_server_urls;
+     Main.flags = flags;
+
      argc -= optind;argv += optind;
 
      if((argc>1) || (help)){
@@ -1178,41 +1135,45 @@ int main(int argc, char **argv) {
      UdmSigHandlersInit(&Main);
 #endif
      
-     Main.flags=flags;
-
-     /* UdmOpenCache was here */
-     if(UDM_OK != UdmOpenCache(&Main, 0)){
-          fprintf(stderr,"Cache mode initializing error: '%s'\n",UdmEnvErrMsg(&Conf));
-          UdmEnvFree(&Conf);
-          exit(1);
-     }
-     
      if (cmd==UDM_IND_SQLMON){
-          SQLMonitor(&Main, &Conf, stdin, stdout);
+          UDM_SQLMON_PARAM prm;
+          bzero((void*)&prm,sizeof(prm));
+          prm.infile= stdin;
+          prm.outfile= stdout;
+          prm.flags= UDM_SQLMON_DISPLAY_FIELDS;
+          prm.gets= sqlmongets;
+          prm.display= UdmDisplaySQLQuery;
+          prm.prompt= sqlmonprompt;
+          UdmSQLMonitor(&Main, &Conf, &prm);
           exit(0);
      }
+
+     if (cmd == UDM_IND_MULTI2BLOB) {
+         UdmMulti2Blob(&Main);
+         exit(0);
+     }
      
-     if(url_fname && strcmp(url_fname,"-")) {
+     if(url_filename && strcmp(url_filename,"-")) {
           /* Make sure URL file is readable if not STDIN */
           FILE *url_file;
-          if(!(url_file=fopen(url_fname,"r"))){
-               UdmLog(&Main,UDM_LOG_ERROR,"Error: can't open url file '%s': %s",url_fname, strerror(errno));
+          if(!(url_file=fopen(url_filename,"r"))){
+               UdmLog(&Main,UDM_LOG_ERROR,"Error: can't open url file '%s': %s",url_filename, strerror(errno));
                goto ex;
           }
           fclose(url_file);
      }
      
-     if(insert && url_fname) {
+     if(insert && url_filename) {
           
-          if(strcmp(url_fname,"-")){
+          if(strcmp(url_filename,"-")){
                /* Make sure all URLs to be inserted are OK */
-               if(UDM_OK!=UdmURLFile(&Main, url_fname,UDM_URL_FILE_PARSE)){
-                    UdmLog(&Main,UDM_LOG_ERROR,"Error: Invalid URL in '%s'",url_fname);
+               if(UDM_OK!=UdmURLFile(&Main, url_filename,UDM_URL_FILE_PARSE)){
+                    UdmLog(&Main,UDM_LOG_ERROR,"Error: Invalid URL in '%s'",url_filename);
                     goto ex;
                }
           }
           
-          if(UDM_OK!=UdmURLFile(&Main,url_fname,UDM_URL_FILE_INSERT)){
+          if(UDM_OK!=UdmURLFile(&Main,url_filename,UDM_URL_FILE_INSERT)){
                UdmLog(&Main,UDM_LOG_ERROR,"Error: '%s'",UdmEnvErrMsg(Main.Conf));
                goto ex;
           }
@@ -1221,8 +1182,8 @@ int main(int argc, char **argv) {
      if(expire){
           int  res;
           
-          if(url_fname){
-               res=UdmURLFile(&Main,url_fname,UDM_URL_FILE_REINDEX);
+          if(url_filename){
+               res=UdmURLFile(&Main,url_filename,UDM_URL_FILE_REINDEX);
           }else{
                res = UdmURLAction(&Main, NULL, UDM_URL_ACTION_EXPIRE);
           }
@@ -1244,7 +1205,7 @@ int main(int argc, char **argv) {
 
      switch(cmd){
           case UDM_IND_DELETE:
-               if(UDM_OK!=UdmClear(&Main,url_fname)){
+               if(UDM_OK!=UdmClear(&Main,url_filename)){
                     UdmLog(&Main,UDM_LOG_ERROR,"Error: '%s'",UdmEnvErrMsg(Main.Conf));
                }
                break;
@@ -1282,12 +1243,13 @@ int main(int argc, char **argv) {
                     }
                     sprintf(pidbuf,"%d\n",(int)getpid());
                     write(pid_fd,&pidbuf,strlen(pidbuf));
+#ifdef HAVE_ATEXIT
                     atexit(&exitproc);
+#endif
                }
                UdmLog(&Main,UDM_LOG_ERROR, "indexer from %s-%s-%s started with '%s'", PACKAGE, VERSION, UDM_DBTYPE, cname);
                UdmStoreHrefs(&Main);    /**< store hrefs from config and command line */
                UdmIndex(&Main);
-               UdmURLAction(&Main, NULL, UDM_URL_ACTION_WRITEDATA);
           }
      }
      
@@ -1295,22 +1257,17 @@ int main(int argc, char **argv) {
      if (pop_rank) {
        UdmSrvAction(&Main, NULL, UDM_SRV_ACTION_POPRANK);
      }
-     if (check_stored) {
-       UdmStoreCheckUp(&Main);
-     }
-     if (check_cached) {
-       UdmCachedCheck(&Main);
-     }
 #endif
      
 ex:
-     UdmCloseCache(&Main, 0);
      total_threads=0;
      UdmAgentFree(&Main);
      UdmEnvFree(&Conf);
      UdmDestroyMutexes();
      UdmWSACleanup();
-     
+#ifndef HAVE_ATEXIT
+     exitproc();
+#endif
      return(0);
 }
 
