@@ -49,6 +49,35 @@
 /****************************************************************/
 
 
+static int UdmReallocSection(UDM_AGENT *Indexer, UDM_VAR *Sec)
+{
+	if(!Sec->val){
+		Sec->val=(char*)UdmMalloc(Sec->maxlen+1);
+	}else{
+		/* Add separator */
+		const char *uspace;
+		char *vn = UdmStrStore(NULL, "separator.");
+		size_t uspacel;
+		size_t space_left;
+
+		vn = UdmStrStore(vn, Sec->name);
+
+		UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
+		uspace = UdmVarListFindStr(&Indexer->Conf->Vars, vn, " ");
+		UdmFree(vn);
+		uspacel = uspace ? strlen(uspace) : 0;
+		space_left = Sec->maxlen - Sec->curlen;
+		if (space_left > uspacel) {
+			Sec->curlen += snprintf(Sec->val + Sec->curlen, space_left, "%s", uspace);
+		} else {
+			Sec->curlen = Sec->maxlen;
+		}
+
+		UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
+	}
+	return UDM_OK;
+}
+
 int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 	size_t		i;
 	const char	*doccset;
@@ -65,14 +94,27 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 	int             *uword;    /* Word in UNICODE      */
 	char            *lcsword;  /* Word in LocalCharset */
 	size_t          max_word_len, min_word_len, uwordlen = UDM_MAXWORDSIZE;
+#ifdef USE_PARANOIA
+	void *paran = UdmViolationEnter();
+#endif
 	
-	if ((uword = (int*)malloc((uwordlen + 1) * sizeof(int))) == NULL) return UDM_ERROR;
-	if ((lcsword = (char*)malloc(12 * uwordlen + 1)) == NULL) { UDM_FREE(uword); return UDM_ERROR; }
+	if ((uword = (int*)UdmMalloc((uwordlen + 1) * sizeof(int))) == NULL) {
+#ifdef USE_PARANOIA
+	  UdmViolationExit(paran);
+#endif
+	  return UDM_ERROR;
+	}
+	if ((lcsword = (char*)UdmMalloc(12 * uwordlen + 1)) == NULL) { UDM_FREE(uword); 
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
+	return UDM_ERROR; }
 
 	Sec=UdmVarListFind(&Doc->Sections,"crosswords");
 	crossec=Sec?Sec->section:0;
 	
-	doccset=UdmVarListFindStr(&Doc->Sections,"Charset",NULL);
+	doccset=UdmVarListFindStr(&Doc->Sections,"Parser.Charset",NULL);
+	if(!doccset)doccset=UdmVarListFindStr(&Doc->Sections,"Charset",NULL);
 	if(!doccset||!*doccset)doccset=UdmVarListFindStr(&Doc->Sections,"RemoteCharset","iso-8859-1");
 	doccs=UdmGetCharSet(doccset);
 	if(!doccs)doccs=UdmGetCharSet("iso-8859-1");
@@ -93,20 +135,24 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 	
 	for(i=0;i<tlist->nitems;i++){
 		size_t		srclen;
+		size_t		srclen0;
 		size_t		dstlen;
 		size_t		reslen;
 		char		*src,*dst;
-		int		uspace=' ';
 		int		*lt, *tok, *ustr = NULL, *UStr = NULL;
 		UDM_TEXTITEM	*Item=&tlist->Item[i];
+		char		secname[128];
 		
-		
-		srclen=strlen(Item->str)+1;	/* with '\0' */
+		srclen0=strlen(Item->str);
+		srclen=srclen0+1;			/* with '\0' */
 		dstlen=(3*srclen+1)*sizeof(int);	/* with '\0' */
 		
-		if ((ustr = (int*)malloc(dstlen)) == NULL) {
+		if ((ustr = (int*)UdmMalloc(dstlen)) == NULL) {
 		  UdmLog(Indexer, UDM_LOG_ERROR, "%s:%d Can't alloc %u bytes", __FILE__, __LINE__, dstlen);
 		  UDM_FREE(uword); UDM_FREE(lcsword);
+#ifdef USE_PARANOIA
+		  UdmViolationExit(paran);
+#endif
 		  return UDM_ERROR;
 		}
 		
@@ -117,6 +163,9 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 		if ((UStr = UdmUniDup(ustr)) == NULL) {
 		  UdmLog(Indexer, UDM_LOG_ERROR, "%s:%d Can't UdmUniDup", __FILE__, __LINE__);
 		  UDM_FREE(uword); UDM_FREE(lcsword); UDM_FREE(ustr);
+#ifdef USE_PARANOIA
+		  UdmViolationExit(paran);
+#endif
 		  return UDM_ERROR;
 		}
 		reslen = UdmUniLen(ustr);
@@ -145,11 +194,19 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 				
 				        if (tlen > uwordlen) {
 					  uwordlen = tlen;
-					  if ((uword = (int*)realloc(uword, (uwordlen + 1) * sizeof(int))) == NULL) { 
-					    UDM_FREE(lcsword); UDM_FREE(ustr); UDM_FREE(UStr); return UDM_ERROR;
+					  if ((uword = (int*)UdmRealloc(uword, (uwordlen + 1) * sizeof(int))) == NULL) { 
+					    UDM_FREE(lcsword); UDM_FREE(ustr); UDM_FREE(UStr); 
+#ifdef USE_PARANOIA
+					    UdmViolationExit(paran);
+#endif
+					    return UDM_ERROR;
 					  }
-					  if ((lcsword = (char*)realloc(lcsword, 12 *uwordlen + 1)) == NULL) { 
-					    UDM_FREE(uword); UDM_FREE(ustr); UDM_FREE(UStr); return UDM_ERROR;
+					  if ((lcsword = (char*)UdmRealloc(lcsword, 12 *uwordlen + 1)) == NULL) { 
+					    UDM_FREE(uword); UDM_FREE(ustr); UDM_FREE(UStr); 
+#ifdef USE_PARANOIA
+					    UdmViolationExit(paran);
+#endif
+					    return UDM_ERROR;
 					  }
 					}
 
@@ -165,7 +222,7 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 					if(Item->href && crossec){
 						UDM_CROSSWORD cw;
 						cw.url=Item->href;
-						cw.weight=Sec->section;
+						cw.weight = crossec;
 						cw.pos=Doc->CrossWords.wordpos;
 						cw.word=lcsword;
 						UdmCrossListAdd(Doc,&cw);
@@ -175,23 +232,12 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 		}
 		
 		if((Sec=UdmVarListFind(&Doc->Sections,Item->section_name))){
-			
-			/* +4 to avoid attempts to fill the only one  */
-			/* last byte with multibyte sequence          */
-			/* as well as to add a space between portions */
-			
+
 			if(Sec->curlen < Sec->maxlen){
 				int cnvres;
-				
-				if(!Sec->val){
-					Sec->val=(char*)malloc(Sec->maxlen+1);
-				}else{
-					/* Add space */
-					UdmConv(&uni_lc,Sec->val+Sec->curlen,1,(char*)(&uspace),sizeof(uspace));
-					Sec->curlen+=uni_lc.obytes;
-					Sec->val[Sec->curlen]='\0';
-				}
-				
+
+				UdmReallocSection(Indexer,Sec);
+
 				src = (char*)UStr;
 				srclen = UdmUniLen(UStr) * sizeof(int);
 				dstlen=Sec->maxlen-Sec->curlen;
@@ -205,13 +251,37 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 			}
 		}
 		
+		udm_snprintf(secname,sizeof(secname)-1,"Raw.%s",Item->section_name);
+		if ((Sec=UdmVarListFind(&Doc->Sections,secname)))
+		{
+			if(Sec->curlen < Sec->maxlen){
+				size_t nbytes;
+				
+				UdmReallocSection(Indexer,Sec);
+				
+				dstlen= Sec->maxlen-Sec->curlen;
+				nbytes= dstlen < srclen0 ? dstlen : srclen0;
+				memcpy(Sec->val+Sec->curlen,Item->str,nbytes);
+				Sec->curlen+=nbytes;
+				Sec->val[Sec->curlen]='\0';
+				
+				if (dstlen<srclen0){
+					Sec->curlen=Sec->maxlen;
+				}
+			}
+		}
+		
 		UDM_FREE(ustr);
 		UDM_FREE(UStr);
 		if(res!=UDM_OK)break;
+		
 	}
 	UdmVarListReplaceInt(&Doc->Sections,"crc32",crc32);
 	
 	UDM_FREE(uword); UDM_FREE(lcsword);
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
 	return res;
 }
 
@@ -247,7 +317,7 @@ int UdmParseURLText(UDM_AGENT *A,UDM_DOCUMENT *Doc){
 	}
 	if((Sec=UdmVarListFind(&Doc->Sections,"url.file"))) {
 	        char *str, sc[]="url.file";
-		str = (char*)malloc(strlen(UDM_NULL2EMPTY(Doc->CurURL.filename)) + 1);
+		str = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(Doc->CurURL.filename)) + 1);
 		if (str != NULL) {
 		  UdmUnescapeCGIQuery(str, UDM_NULL2EMPTY(Doc->CurURL.filename));
 		  Item.str = str;
@@ -353,6 +423,11 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t){
 					return(t->s);
 				}
 
+				if(*t->b=='<'){ /* Probably broken tag occure */
+					*lt = t->b;
+					return(t->s);
+				}
+
 				/* Skip non-spaces, i.e. name */
 				for(t->e = t->b; (*t->e) && !strchr(" =>\t\r\n", *t->e); (*t->next_e)(t));
 				
@@ -364,8 +439,16 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t){
 				t->toks[nt].name = t->b;
 				t->toks[nt].nlen = t->e - t->b;
 
-				if(!strncasecmp(t->b,"script",6)) t->script = 1;
-				if(!strncasecmp(t->b,"/script",7)) t->script = 0;
+				if (nt == 0) {
+				  if(!strncasecmp(t->b,"script",6)) t->script = 1;
+				  if(!strncasecmp(t->b,"/script",7)) t->script = 0;
+				  if(!strncasecmp(t->b, "noindex", 7)) t->comment = 1;
+				  if(!strncasecmp(t->b, "/noindex", 8)) t->comment = 0;
+				  if(!strncasecmp(t->b, "style", 5)) t->style = 1;
+				  if(!strncasecmp(t->b, "/style", 6)) t->style = 0;
+				  if(!strncasecmp(t->b, "body", 4)) t->body = 1;
+				  if(!strncasecmp(t->b, "/body", 5)) t->body = 0;
+				}
 
 				if(*t->e=='>'){
 					*lt = t->e + 1;
@@ -482,44 +565,63 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc){
 	char *secname;
 	size_t i, seclen = 128;
 
-	if(!tag->ntoks)return(0);
-	if(!tag->toks[0].name)return(0);
-	if(tag->toks[0].nlen>sizeof(name)-1)return(0);
+#ifdef USE_PARANOIA
+	void *paran = UdmViolationEnter();
+#endif
 
-	secname = (char*)malloc(seclen);
+	if(!tag->ntoks) {
+#ifdef USE_PARANOIA
+	  UdmViolationExit(paran);
+#endif
+	  return(0);
+	}
+	if(!tag->toks[0].name) {
+#ifdef USE_PARANOIA
+	  UdmViolationExit(paran);
+#endif
+	  return(0);
+	}
+	if(tag->toks[0].nlen>sizeof(name)-1) {
+#ifdef USE_PARANOIA
+	  UdmViolationExit(paran);
+#endif
+	  return(0);
+	}
+
+	secname = (char*)UdmMalloc(seclen);
 	strncpy(name,tag->toks[0].name,tag->toks[0].nlen);
 	name[tag->toks[0].nlen]='\0';
 	
 	for(i=0;i<tag->ntoks;i++){
 		if(ISTAG(i,"name")){
-			metaname=strndup(tag->toks[i].val,tag->toks[i].vlen);
+			metaname = UdmStrndup(tag->toks[i].val,tag->toks[i].vlen);
 		}else
 		if(ISTAG(i,"http-equiv")){
-			metaname=strndup(tag->toks[i].val,tag->toks[i].vlen);
+			metaname = UdmStrndup(tag->toks[i].val,tag->toks[i].vlen);
 		}else
 		if(ISTAG(i,"content")){
-			metacont=strndup(tag->toks[i].val,tag->toks[i].vlen);
+			metacont = UdmStrndup(tag->toks[i].val,tag->toks[i].vlen);
 		}else
 		if(ISTAG(i,"href")){
 			/* A, LINK, AREA*/
-			char *y = strndup(tag->toks[i].val,tag->toks[i].vlen);
-			href = (char*)strdup(UdmTrim(y, " \t\r\n"));
+			char *y = UdmStrndup(tag->toks[i].val,tag->toks[i].vlen);
+			href = (char*)UdmStrdup(UdmTrim(y, " \t\r\n"));
 			UDM_FREE(y);
 		}else
 		if(ISTAG(i,"src")){
 			/* IMG, FRAME */
-			char *y = strndup(tag->toks[i].val,tag->toks[i].vlen);
-			href = (char*)strdup(UdmTrim(y, " \t\r\n"));
+			char *y = UdmStrndup(tag->toks[i].val,tag->toks[i].vlen);
+			href = (char*)UdmStrdup(UdmTrim(y, " \t\r\n"));
 			UDM_FREE(y);
 		}else
 		if (ISTAG(i, "lang")) {
-			char *y = strndup(tag->toks[i].val,tag->toks[i].vlen);
-			lang = (char*)strdup(UdmTrim(y, " \t\r\n"));
+			char *y = UdmStrndup(tag->toks[i].val,tag->toks[i].vlen);
+			lang = (char*)UdmStrdup(UdmTrim(y, " \t\r\n"));
 			for(n = lang; *n; *n = tolower(*n),n++);
 			UDM_FREE(y);
 		} else {
 		        if (tag->toks[i].nlen + 12 > seclen) {
-			  secname = (char*)realloc(secname, seclen = (tag->toks[i].nlen + 12));
+			  secname = (char*)UdmRealloc(secname, seclen = (tag->toks[i].nlen + 12));
 			}
 			
 		        strcpy(secname, "attribute.");
@@ -527,7 +629,7 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc){
 			secname[seclen - 1]='\0';
 
 			if ((Sec = UdmVarListFind(&Doc->Sections, secname)) && Doc->Spider.index) {
-			  char *y = strndup(tag->toks[i].val,tag->toks[i].vlen);
+			  char *y = UdmStrndup(tag->toks[i].val,tag->toks[i].vlen);
 			  Item.str = y;
 			  Item.section = Sec->section;
 			  Item.section_name = secname;
@@ -587,7 +689,7 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc){
 		}else
 		if(!strcasecmp(metaname, "Content-Language") || !strcasecmp(metaname, "DC.Language")) {
 			char *l;
-			l = (char*)strdup(metacont);
+			l = (char*)UdmStrdup(metacont);
 			for(n = l; *n; *n = tolower(*n),n++);
 			UdmVarListReplaceStr(&Doc->Sections, "Meta-Language", l);
 			UDM_FREE(l);
@@ -600,7 +702,7 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc){
 			
 			if((p = strchr(metacont, '='))){
 				if((p >= metacont + 3) && (!strncasecmp(p-3,"URL=",4))){
-					href = (char*)strdup(p + 1);
+					href = (char*)UdmStrdup(p + 1);
 				}else{
 					UDM_FREE(href);
 				}
@@ -671,7 +773,7 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc){
 		
 		/* For crosswords */
 		UDM_FREE(tag->lasthref);
-		tag->lasthref = (char*)strdup(href);
+		tag->lasthref = (char*)UdmStrdup(href);
 	}
 	UDM_FREE(metaname);
 	UDM_FREE(metacont);
@@ -679,6 +781,9 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc){
 	UDM_FREE(lang);
 	UDM_FREE(secname);
 	
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
 	return 0;
 }
 
@@ -707,6 +812,10 @@ int UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 	int		title_sec = TSec ? TSec->section : 0;
 	char		scb[]="body";
 	char		sct[]="title";
+
+#ifdef USE_PARANOIA
+	void *paran = UdmViolationEnter();
+#endif
 	
 	bzero((void*)&Item, sizeof(Item));
 	UdmHTMLTOKInit(&tag);
@@ -732,7 +841,7 @@ int UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 				for( tmpend=last-1; htok<tmpend && strchr(" \r\n\t",tmpend[0]) ; tmpend--);
 				if(tmpbeg>=tmpend)break;
 				
-				tmp=strndup(tmpbeg,(size_t)(tmpend-tmpbeg+1));
+				tmp = UdmStrndup(tmpbeg,(size_t)(tmpend-tmpbeg+1));
 				
 				if (BSec && !tag.comment && tag.body && !tag.script && !tag.style && tag.index) {
 					Item.href=tag.lasthref;
@@ -801,5 +910,8 @@ int UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 		
 	}
 	UDM_FREE(tag.lasthref);	
+#ifdef USE_PARANOIA
+	UdmViolationExit(paran);
+#endif
 	return UDM_OK;
 }
