@@ -47,12 +47,13 @@
 #include "udm_log.h"
 #include "udm_vars.h"
 #include "udm_wild.h"
+#include "udm_signals.h"
 
 __C_LINK int __UDMCALL UdmParserAdd(UDM_PARSERLIST *List,UDM_PARSER *P){
-	List->Parser=(UDM_PARSER*)(realloc(List->Parser,(List->nparsers+1)*sizeof(UDM_PARSER)));
-	List->Parser[List->nparsers].from_mime = (char*)strdup(P->from_mime);
-	List->Parser[List->nparsers].to_mime = (char*)strdup(P->to_mime);
-	List->Parser[List->nparsers].cmd = (char*)strdup(P->cmd);
+	List->Parser=(UDM_PARSER*)(UdmRealloc(List->Parser,(List->nparsers+1)*sizeof(UDM_PARSER)));
+	List->Parser[List->nparsers].from_mime = (char*)UdmStrdup(P->from_mime);
+	List->Parser[List->nparsers].to_mime = (char*)UdmStrdup(P->to_mime);
+	List->Parser[List->nparsers].cmd = (char*)UdmStrdup(P->cmd);
 	List->nparsers++;
 	return 0;
 }
@@ -74,7 +75,7 @@ __C_LINK void __UDMCALL UdmParserListFree(UDM_PARSERLIST *List){
 /* Parser1: from STDIN to STDOUT */
 
 #ifdef WIN32
-static char *parse1(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size_t maxlen, const char *url){
+static char *parse1(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size_t maxlen) {
 	UdmLog(Agent,UDM_LOG_ERROR,"This type of parser isn't supported");
 	return NULL;
 }
@@ -96,7 +97,7 @@ static void sighandler(int sign){
 	init_signals();
 }
 
-static char *parse1(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size_t maxlen, const char *url){
+static char *parse1(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size_t maxlen) {
 	char *result = NULL;
 	int wr[2];
 	int rd[2];    
@@ -180,7 +181,7 @@ static char *parse1(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size
 #endif
 
 /* Parser2: from FILE to STDOUT */
-static char *parse2(UDM_AGENT * Agent, char *buf, char *cmd, size_t maxlen, const char *url){
+static char *parse2(UDM_AGENT * Agent, char *buf, char *cmd, size_t maxlen) {
 	FILE *f;
 	char *result = NULL;
 
@@ -215,7 +216,7 @@ static char *parse2(UDM_AGENT * Agent, char *buf, char *cmd, size_t maxlen, cons
 }
 
 /* Parser3: from FILE to FILE */
-static char *parse3(UDM_AGENT * Agent, char *buf, char *cmd, size_t maxlen, const char *url, char *to_file){
+static char *parse3(UDM_AGENT * Agent, char *buf, char *cmd, size_t maxlen, char *to_file) {
 	char *result = NULL;
 	int fd;
 
@@ -235,7 +236,7 @@ static char *parse3(UDM_AGENT * Agent, char *buf, char *cmd, size_t maxlen, cons
 }
 
 /* Parser4: from STDIN to FILE */
-static char *parse4(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size_t maxlen, const char *url, char *to_file){
+static char *parse4(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size_t maxlen, char *to_file){
 	FILE *f;
 	char *result = NULL;
 
@@ -324,10 +325,10 @@ static char *parse_file (UDM_AGENT * Agent, UDM_PARSER * parser, char *buf, size
 	UdmLog(Agent,UDM_LOG_EXTRA,"Starting external parser: '%s'",cmd);
 	UdmSetEnv("UDM_URL",url);
 	switch(parser_type){
-		case 1: result=parse1(Agent,buf,length,cmd,maxlen,url); break;
-		case 2: result=parse2(Agent,buf,cmd,maxlen,url); break;
-		case 3: result=parse3(Agent,buf,cmd,maxlen,url,fnames[1]); break;
-		case 4: result=parse4(Agent,buf,length,cmd,maxlen,url,fnames[1]); break;
+		case 1: result = parse1(Agent, buf, length, cmd, maxlen); break;
+		case 2: result = parse2(Agent, buf, cmd, maxlen); break;
+		case 3: result = parse3(Agent, buf, cmd, maxlen, fnames[1]); break;
+		case 4: result = parse4(Agent, buf, length, cmd, maxlen, fnames[1]); break;
 	}
 	UdmUnsetEnv("UDM_URL");
 
@@ -348,9 +349,12 @@ __C_LINK UDM_PARSER * __UDMCALL UdmParserFind(UDM_PARSERLIST *List,const char *m
 }
 
 char *UdmParserExec(UDM_AGENT * Agent,UDM_PARSER *P,UDM_DOCUMENT *Doc){
+  char *result;
 	size_t maxlen=(size_t)(Doc->Buf.maxsize-(Doc->Buf.content-Doc->Buf.buf));
 	size_t srclen=(size_t)(Doc->Buf.size-(Doc->Buf.content-Doc->Buf.buf));
-	return parse_file(Agent,P,Doc->Buf.content,srclen,maxlen,UdmVarListFindStr(&Doc->Sections,"URL",""));
+	result = parse_file(Agent,P,Doc->Buf.content,srclen,maxlen,UdmVarListFindStr(&Doc->Sections,"URL",""));
+	Doc->Buf.size = strlen(Doc->Buf.content) + (Doc->Buf.content-Doc->Buf.buf);
+	return result;
 }
 
 #endif
