@@ -401,7 +401,7 @@ static int UdmFTPGet(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc)
 		char * pass=NULL;
 		
 		if (Doc->CurURL.auth != NULL) {
-			authstr = (char*)strdup(Doc->CurURL.auth);
+			authstr = (char*)UdmStrdup(Doc->CurURL.auth);
 			user=authstr;
 			if((pass=strchr(authstr,':'))){
 				*pass='\0';
@@ -431,8 +431,7 @@ static int UdmFTPGet(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc)
 		if (Doc->CurURL.filename == NULL) {
 			code = Udm_ftp_cwd(&Doc->connp, UDM_NULL2EMPTY(Doc->CurURL.path));
 			if (code != -1){
-				code = Udm_ftp_list(&Doc->connp, Doc->connp.connp, UDM_NULL2EMPTY(Doc->CurURL.path),
-						    NULL, Doc->Buf.maxsize);
+				code = Udm_ftp_list(Doc);
 				if (code == -1){
 					if (Doc->connp.err >0){
 						sprintf(Doc->Buf.buf, "HTTP/1.1 403 OK\r\n\r\n");
@@ -456,7 +455,7 @@ static int UdmFTPGet(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc)
 			}
 		}else{
 			len = strlen(UDM_NULL2EMPTY(Doc->CurURL.path)) + strlen(UDM_NULL2EMPTY(Doc->CurURL.filename));
-			full_path = malloc(len+1);
+			full_path = UdmMalloc(len+1);
 			udm_snprintf(full_path, len + 1, "%s%s", UDM_NULL2EMPTY(Doc->CurURL.path), UDM_NULL2EMPTY(Doc->CurURL.filename));
 			full_path[len]='\0';
 			last_mod_tm = Udm_ftp_mdtm(&Doc->connp,full_path);
@@ -1184,6 +1183,7 @@ static int UdmFILEGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 	char proto[32]="";
 	char *lt,*s;
 	time_t ims=0;
+	int  cvs_ignore= Indexer->Conf->CVS_ignore;
 
 	Doc->Buf.size=0;
 	sscanf(Doc->Buf.buf,"%s%s%s",command,filename,proto);
@@ -1207,6 +1207,8 @@ static int UdmFILEGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 		s = udm_strtok_r(NULL,"\r\n",&lt);
 	}
 
+	strcpy(mystatname,openname);
+
 #ifdef WIN32
 	/* Remove traling slash sign stat("c:\a\") doesn't work  */
 	/* under win. We'll pass stat("c:\a") instead            */
@@ -1215,17 +1217,16 @@ static int UdmFILEGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 	{
 		int len;
 		
-		s = openname;
-		len=strlen(openname);
+		s = mystatname;
+		len=strlen(s);
 		while(*s){
 			if(*s=='/')*s='\\';
 			s++;
 		}
-		if ((len>3)&&(openname[len-1]=='\\'))
-			openname[len-1]='\0';
+		if ((len>3)&&(mystatname[len-1]=='\\'))
+			mystatname[len-1]='\0';
 	}
 #endif
-	strcpy(mystatname,openname);
 	
 	if(stat(mystatname,&sb)){
 		switch(errno){
@@ -1261,14 +1262,25 @@ static int UdmFILEGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 				This does not work on Solaris
 				is_dir=(rec->d_type==DT_DIR);
 				*/
+				if (cvs_ignore && !strcmp(rec->d_name,"CVS"))
+					continue;
+				
 				sprintf(newfilename, "%s%s", openname, rec->d_name);
+
+#ifndef WIN32
+				if (! lstat(newfilename, &sb1) &&
+					S_ISLNK(sb1.st_mode) &&
+					! UdmVarListFindInt(&Doc->Sections, "FollowSymLinks", 0)) {
+					continue;
+				}
+#endif
 				if(stat(newfilename,&sb1)){
 					UdmLog(Indexer, UDM_LOG_EXTRA, "Can't stat '%s'", newfilename);
 					is_dir = 0;
 				} else { 
 					is_dir = ((sb1.st_mode&S_IFDIR) > 0);
 				}
-				
+
 				e=escaped_name;
 				for(src=rec->d_name;*src;src++){
 					if(strchr(" %&<>+[](){}/?#'\"\\;,",*src)){
@@ -1283,7 +1295,7 @@ static int UdmFILEGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 				sprintf(stre,"<A HREF=\"%s%s\">\n",escaped_name,is_dir?"/":"");
 				stre=UDM_STREND(stre);
 				if ((size_t)(stre - Doc->Buf.buf) > (Doc->Buf.maxsize - 2048)) {
-				  e = (char*) realloc(Doc->Buf.buf, (Doc->Buf.maxsize + 65536) * sizeof(char));
+				  e = (char*) UdmRealloc(Doc->Buf.buf, (Doc->Buf.maxsize + 65536) * sizeof(char));
 				  if (e == NULL) break;
 				  Doc->Buf.buf = e;
 				  Doc->Buf.maxsize += 65536;
@@ -1350,16 +1362,16 @@ static int UdmBuildHTTPRequest(UDM_DOCUMENT *Doc){
 	const char	*method=(Doc->method==UDM_METHOD_HEAD)?"HEAD":"GET";
 	const char	*proxy=UdmVarListFindStr(&Doc->RequestHeaders,"Proxy",NULL);
 	size_t		i;
-	char            *url = (char*)malloc((i = strlen(UDM_NULL2EMPTY(Doc->CurURL.path)) + 
+	char            *url = (char*)UdmMalloc((i = strlen(UDM_NULL2EMPTY(Doc->CurURL.path)) + 
 					      strlen(UDM_NULL2EMPTY(Doc->CurURL.filename)) + 1));
-	char            *eurl = (char*)malloc(3 * i);
+	char            *eurl = (char*)UdmMalloc(3 * i);
 
 	if (url == NULL || eurl == NULL) return UDM_ERROR;
 
 	sprintf(url, "%s%s", UDM_NULL2EMPTY(Doc->CurURL.path), UDM_NULL2EMPTY(Doc->CurURL.filename)); 
 	UdmEscapeURI(eurl, url);
 	
-	if(!Doc->Buf.buf)Doc->Buf.buf=(char*)malloc(Doc->Buf.maxsize + 1);
+	if(!Doc->Buf.buf)Doc->Buf.buf=(char*)UdmMalloc(Doc->Buf.maxsize + 1);
 	
 	if(proxy && strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "file")) {
 		sprintf(Doc->Buf.buf,"%s %s://%s%s HTTP/1.0\r\n", method, UDM_NULL2EMPTY(Doc->CurURL.schema), 
@@ -1416,11 +1428,13 @@ int UdmGetURL(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
 	}else
 	if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "cgi")) {
 		res = UdmExecGet(Indexer,Doc);
-	}else
+	}
+#ifdef HAVE_SQL
+	else
 	if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "htdb")) {
 		res = UdmHTDBGet(Indexer, Doc);
 	}
-	
+#endif
 #ifdef USE_FILE
 	else
 	if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "file")) {
