@@ -37,12 +37,17 @@
 #include "dmalloc.h"
 #endif
 
+#ifdef WIN32
+#include "udm_winutils.h"
+#endif
+
+#define POWERED_LOGO "<BR><a href=\"http://www.mnogosearch.org/\"><IMG BORDER=0 SRC=\"http://www.mnogosearch.org/img/mwin.gif\"></A>&nbsp<FONT SIZE=1><A HREF=\"http://www.mnogosearch.org/\">Powered by mnoGoSearch - free web search engine software</A></FONT><BR>\n"
 
 /******************** Misc functions *********************************/
 
 static void ParseQStringUnescaped(UDM_VARLIST *vars, const char *qstring){
 	char *tok, *lt; 
-	char *qs = (char*)strdup(qstring);
+	char *qs = (char*)UdmStrdup(qstring);
 
 	if (qs != NULL) {
 	  UdmSGMLUnescape(qs);
@@ -53,7 +58,7 @@ static void ParseQStringUnescaped(UDM_VARLIST *vars, const char *qstring){
 		UdmVarListAddStr(vars,tok,arg?arg:"");
 		tok = udm_strtok_r(NULL, "&", &lt);
 	  }
-	  free(qs);
+	  UdmFree(qs);
 	}
 }
 
@@ -64,7 +69,7 @@ static char * BuildPageURL(UDM_VARLIST * vars, char **dst) {
 	for(i = 0; i < vars->nvars; i++) {
 	  dstlen += 7 + strlen(vars->Var[i].name) + strlen(vars->Var[i].val);
 	}
-	*dst = (char*)realloc(*dst, dstlen);
+	*dst = (char*)UdmRealloc(*dst, dstlen);
 	if (*dst == NULL) return NULL;
 	end = *dst;
 
@@ -91,10 +96,10 @@ static int WordInfo(UDM_ENV *Env, UDM_RESULT *Res){
 	size_t	corder = (size_t)-1, ccount = 0;
 	
 	for(len = i = 0; i < Res->WWList.nwords; i++) 
-		len += Res->WWList.Word[i].len;
+		len += Res->WWList.Word[i].len + 64;
 	
-	wsize=(1+len*15)*sizeof(char);
-	wordinfo = (char*) malloc(wsize);
+	wsize=(1+len)*sizeof(char);
+	wordinfo = (char*) UdmMalloc(wsize);
 	*wordinfo = '\0';
 	
 	for(i = 0; i < Res->WWList.nwords; i++){
@@ -121,9 +126,9 @@ static int WordInfo(UDM_ENV *Env, UDM_RESULT *Res){
 	  if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_STOP) {
 		  sprintf(UDM_STREND(wordinfo),"%s%s : stopword", (*wordinfo) ? ", " : "",  Res->WWList.Word[i].word);
 	  } else if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_QUERY) {
-		  sprintf(UDM_STREND(wordinfo),"%s%s : %d", (*wordinfo) ? ", " : "", Res->WWList.Word[i].word, Res->WWList.Word[i].count);
-	  } else continue;
-	  sprintf(UDM_STREND(wordinfo)," / %d", ccount);
+		  sprintf(UDM_STREND(wordinfo),"%s%s : %d / %d", 
+			  (*wordinfo) ? ", " : "", Res->WWList.Word[i].word, Res->WWList.Word[i].count, ccount);
+	  }
 	}
 	
 	UdmVarListAddStr(&Env->Vars, "W", wordinfo);
@@ -134,10 +139,120 @@ static int WordInfo(UDM_ENV *Env, UDM_RESULT *Res){
 
 /*****************************************************************/
 
+static char *UdmConfDir(void)
+{
+#ifdef WIN32
+	return WinUdmConfDir();
+#else
+	char* result;
+	result = (char *)UdmStrdup(UDM_CONF_DIR);
+	return result;
+#endif
+}
+
+
+/************************************************************/
+static int StoreDocCGI (UDM_AGENT *Agent, UDM_ENV *Env, UDM_VARLIST *tmpl) {
+	const char	*htok;
+	char            *last = NULL;
+	UDM_DOCUMENT	*Doc;
+	UDM_RESULT	*Res;
+	UDM_HTMLTOK	tag;
+	char		*HDoc=NULL;
+	char		*HEnd=NULL;
+	char            ch;
+	const char            *content_type, *charset;
+	UDM_CHARSET	*cs;
+#ifdef USE_PARSER
+	UDM_PARSER	*Parser;
+#endif
+
+	
+	Doc=UdmDocInit(NULL);
+	Res=UdmResultInit(NULL);
+	
+	UdmPrepare(Agent,Res);
+	UdmVarListReplaceStr(&Doc->Sections, "URL", UdmVarListFindStr(&Env->Vars, "URL", "0"));
+	UdmURLAction(Agent, Doc, UDM_URL_ACTION_GET_CACHED_COPY);
+	UdmVarListReplaceLst(&Env->Vars, &Doc->Sections, NULL, "*");
+
+	charset = UdmVarListFindStr(&Env->Vars,"Charset","iso-8859-1");
+	cs = UdmGetCharSet(charset);
+
+	printf("Content-type: text/html; charset=%s\r\n\r\n", charset);
+
+	/* Now start displaying template*/
+	UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, tmpl, "storedoc_top");
+	
+	/* UnStore Doc, Highlight and Display */
+	
+	content_type = UdmVarListFindStr(&Env->Vars, "Content-Type", "text/html");
+
+#ifdef USE_PARSER
+	Parser = UdmParserFind(&Env->Parsers, content_type);
+	if (Parser) {
+		content_type = Parser->to_mime ? Parser->to_mime : "text/html";
+	}
+#endif
+	if(!Doc->Buf.content)goto fin;
+	
+	HEnd=HDoc = (char*)UdmMalloc(UDM_MAXDOCSIZE + 32);
+	*HEnd='\0';
+	
+	if (strncasecmp(content_type, "text/plain", 10) == 0) {
+	  sprintf(HEnd, "<pre>\n");
+	  HEnd += strlen(HEnd);
+	}
+
+	UdmHTMLTOKInit(&tag);
+	for(htok = UdmHTMLToken(Doc->Buf.content, (const char **)&last, &tag) ; htok ;) {
+		switch(tag.type) {
+		case UDM_HTML_COM:
+		case UDM_HTML_TAG:
+			memcpy(HEnd,htok,(size_t)(last-htok));
+			HEnd+=last-htok;
+			HEnd[0]='\0';
+				UdmHTMLParseTag(&tag,Doc);
+			break;
+		case UDM_HTML_TXT:
+		        ch = *last; *last = '\0';
+			if (tag.title || tag.script) {
+			  sprintf(HEnd, "%s", UdmHlConvert(NULL, htok, cs, cs));
+			} else {
+			  sprintf(HEnd, "%s", UdmHlConvert(&Res->WWList, htok, cs, cs));
+			}
+			HEnd=UDM_STREND(HEnd);
+			*last = ch;
+			break;
+		}
+		htok = UdmHTMLToken(NULL, (const char **)&last, &tag);
+	}
+
+	if (strncasecmp(content_type, "text/plain", 10) == 0) {
+	  sprintf(HEnd, "</pre>\n");
+	  HEnd += strlen(HEnd);
+	}
+
+	UdmVarListAddStr(&Env->Vars, "document", HDoc);
+	UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, tmpl, "storedoc");
+	
+fin:
+	UdmResultFree(Res);
+	UdmDocFree(Doc);
+	
+	UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, tmpl, "storedoc_bottom");
+	UdmVarListFree(tmpl);
+	UdmAgentFree(Agent);
+	UdmEnvFree(Env);
+	UDM_FREE(HDoc);
+	return(0);
+}
+/************************************************************/
+
+
 int main(int argc, char ** argv){
 	const char	*env, *bcharset, *lcharset;
 	char		template_name[1024]="";
-	char            *template_filename = NULL;
 	char		*query_string = NULL;
 	char		self[1024]="";
 	char		*nav = NULL;
@@ -149,7 +264,7 @@ int main(int argc, char ** argv){
 	size_t          catcolumns = 0;
 	ssize_t		page1,page2,npages,ppp=10;
 	int		page_size,page_number;
-	size_t		i, swlen = 0, nav_len, ExcerptSize, storedlen;
+	size_t		i, swlen = 0, nav_len, ExcerptSize, ExcerptPadding, storedlen;
 	UDM_ENV		*Env;
 	UDM_AGENT	*Agent;
 	UDM_RESULT	*Res;
@@ -160,8 +275,16 @@ int main(int argc, char ** argv){
 	/* Some servers do not pass QUERY_STRING */
 	/* if the query was empty, so check	 */
 	/* REQUEST_METHOD too     to be safe     */
-	
+
 	httpd=(getenv("QUERY_STRING")||getenv("REQUEST_METHOD"));
+
+#if ( (defined WIN32) && (defined _DEBUG) )
+	if(httpd)
+	{
+		WIN32_IIS_DEBUG_MSG
+		DebugBreak();
+	}
+#endif
 	
 	UdmInit();
 	Env=UdmEnvInit(NULL);
@@ -170,126 +293,121 @@ int main(int argc, char ** argv){
 	Agent = UdmAgentInit(NULL, Env, 0);
 	UdmVarListAddEnviron(&Env->Vars,"ENV");
 	
-	/* Detect self and template name */
-	if((env=getenv("UDMSEARCH_TEMPLATE")))
-		strncpy(template_name,env,sizeof(template_name)-1);
 	
-	if((env=getenv("UDMSEARCH_SELF")))
+	/* Detect self */
+	if((env=getenv("UDMSEARCH_SELF")) || (env=getenv("REQUEST_URI")))
+	{
+		/* Use Apache's REQUEST_URI, which is not in CGI standard */
+		char *end;
 		strncpy(self,env,sizeof(self)-1);
-	
-	if((env=getenv("QUERY_STRING"))){
-	        query_string = (char*)realloc(query_string, strlen(env) + 2);
-		strncpy(query_string, env, strlen(env) + 1);
-
-		/* Hack for Russian Apache from apache.lexa.ru  */
-		/* QUERY_STRING is already converted to server  */
-		/* character set. We must print original query  */
-		/* string instead however. Under usual apache   */ 
-		/* we'll use QUERY_STRING. Note that query_vars */
-		/* list will contain not unescaped values, so   */
-		/* we don't have to escape them when displaying */
-		env = getenv("CHARSET_SAVED_QUERY_STRING");
-		ParseQStringUnescaped(&query_vars,env?env:query_string);
-	
-		/* Unescape and save variables from QUERY_STRING */
-		/* Env->Vars will have unescaped values however  */
-		UdmParseQueryString(Agent,&Env->Vars,query_string);
-	
-		template_filename = (char*)strdup(UdmVarListFindStr(&Env->Vars, "tmplt", ""));
-	
-		if((env=getenv("REDIRECT_STATUS"))){
-
-			/* Check Apache internal redirect  */
-			/* via   "AddHandler" and "Action" */
-			if(!self[0]){
-				strncpy(self,(env=getenv("REDIRECT_URL"))?env:"search.cgi",sizeof(self)-1);
-			}
-			if(!template_name[0]){
-				strncpy(template_name,(env=getenv("PATH_TRANSLATED"))?env:"",sizeof(template_name)-1);
-			}
-			if (strlen(template_filename) == 0) { template_filename = (char*)strdup("search.htm"); }
-		}else{
-			/* CGI executed without Apache internal redirect */
-
-			/* Detect $Self variable with OS independant SLASHES */
-			if(!self[0]){
-				strncpy(self,(env=getenv("SCRIPT_NAME"))?env:"search.cgi",sizeof(self)-1);
-			}
-
-			if(!template_name[0]){
-				char *s,*e;
-				
-				/*This is with OS specific SLASHES */
-				env=((env=getenv("SCRIPT_FILENAME"))?env:"search.cgi");
-
-				if(strcmp(UDM_CONF_DIR,".")){
-					/* Take from the config directory */
-					udm_snprintf(template_name, sizeof(template_name)-1, "%s/%s", 
-						     UDM_CONF_DIR,(s=strrchr(env,UDMSLASH))?(s+1):(self));
-				}else{
-					/* Take from the current directory */
-					strncpy(template_name,env,sizeof(template_name)-1);
-				}
-
-				/* Find right slash if it presents */
-				s=((s=strrchr(template_name,UDMSLASH))?s:template_name);
-
-				if (strlen(template_filename) == 0) {
-
-				  /* Find .cgi substring */
-				  if ((e = strstr(s, ".cgi")) != NULL) {
-					/* Replace ".cgi" with ".htm" */
-					e[1]='h';e[2]='t';e[3]='m';
-				  } else {
-				        strcat(s, ".htm");
-				  }
-				  e = strrchr(s, '/');
-				  template_filename = (char*)strdup(e + 1);
-				} else {
-				  strncpy(s + 1, template_filename, sizeof(template_name) - (s - template_name) - 2);
-				}
-			}
-		}
-	}else{
-		/* Executed from command line     */
-		/* or under server which does not */
-		/* pass an empty QUERY_STRING var */
-		if(argv[1]) {
-		  query_string = (char*)realloc(query_string, strlen(argv[1]) + 10);
-		  sprintf(query_string, "q=%s", argv[1]);
-		} else {
-		  query_string = (char*)realloc(query_string, 1024);
-		  sprintf(query_string, "q=");
-		}
-
-		/* Hack for Russian Apache from apache.lexa.ru  */
-		/* QUERY_STRING is already converted to server  */
-		/* character set. We must print original query  */
-		/* string instead however. Under usual apache   */ 
-		/* we'll use QUERY_STRING. Note that query_vars */
-		/* list will contain not unescaped values, so   */
-		/* we don't have to escape them when displaying */
-		env = getenv("CHARSET_SAVED_QUERY_STRING");
-		ParseQStringUnescaped(&query_vars,env?env:query_string);
-	
-		/* Unescape and save variables from QUERY_STRING */
-		/* Env->Vars will have unescaped values however  */
-		UdmParseQueryString(Agent,&Env->Vars,query_string);
-
-		template_filename = (char*)strdup(UdmVarListFindStr(&Env->Vars, "tmplt", ""));
-		if (strlen(template_filename) == 0) { UDM_FREE(template_filename); template_filename = (char*)strdup("search.htm"); }
-	
-		if(!template_name[0]) udm_snprintf(template_name,sizeof(template_name),"%s/%s", UDM_CONF_DIR, template_filename);
+		if ((end= strchr(self,'?')))
+			end[0]='\0';
+	}
+	else if ((env=getenv("SCRIPT_NAME")))
+	{
+		/* Use standard CGI variables if not under Apache */
+		const char *path_info= getenv("PATH_INFO");
+		udm_snprintf(self,sizeof(self),"%s%s",env,path_info?path_info:"");
+	}
+	else
+	{
+		strncpy(self,"search.cgi",sizeof(self)-1);
 	}
 	
-	UdmVarListReplaceStr(&Agent->Conf->Vars, "tmplt", template_filename);
-	UDM_FREE(template_filename);
+	
+	/* Detect template name */
+	if((env=getenv("UDMSEARCH_TEMPLATE")))
+	{
+		char ename[1024];
+		strncpy(ename,env,sizeof(ename)-1);
+		ename[sizeof(ename)]='\0';
+		UdmUnescapeCGIQuery(template_name,ename);
+	}
+	else if(getenv("REDIRECT_STATUS") && (env=getenv("PATH_TRANSLATED")))
+	{
+		/* Check Apache internal redirect  */
+		/* via   "AddHandler" and "Action" */
+		strncpy(template_name,env,sizeof(template_name)-1);
+	}
+	else if ((env=getenv("PATH_INFO")) && env[0])
+	{
+		/* http://localhost/cgi-bin/search.cgi/path/to/search.htm */
+		char name[128], *conf_dir;
+		
+		strncpy(name,env,sizeof(name)-1);
+		name[sizeof(name)-1]='\0';
+		if (UDMSLASH !='/')
+		{
+			char *s;
+			for (s=name; s[0]; s++)
+				if (s[0]=='\\')s[0]='/';
+		}
+		
+		conf_dir= (env=getenv("UDM_ETC_DIR")) ? (char *)UdmStrdup(env) : UdmConfDir();
+		/* Take from the config directory */
+		udm_snprintf(template_name, sizeof(template_name)-5, "%s%s%s", 
+			     conf_dir,UDMSLASHSTR,name);
+		UDM_FREE(conf_dir);
+	}
+	else
+	{
+		/* CGI executed without Apache internal redirect */
+		/* or started from command line                  */
+		char *end, *conf_dir;
+		size_t length;
+		
+		conf_dir= (env=getenv("UDM_ETC_DIR")) ? (char *)UdmStrdup(env) : UdmConfDir();
+		
+		/* Take from the config directory */
+		length= udm_snprintf(template_name, sizeof(template_name)-5, "%s%s%s", 
+			     conf_dir,UDMSLASHSTR,
+			     (end=strrchr(self,'/'))?(end+1):(self));
+		end= template_name+length;
+		if ((length>3) && (!strcmp(end-4,".exe")||!strcmp(end-4,".cgi")))
+			end-=4;
+		strcpy(end,".htm");
+		UDM_FREE(conf_dir);
+	}
+	
+	
+	if((env=getenv("QUERY_STRING")))
+	{
+		query_string = (char*)UdmRealloc(query_string, strlen(env) + 2);
+		strncpy(query_string, env, strlen(env) + 1);
+	}
+	else if(argv[1])
+	{
+		query_string = (char*)UdmRealloc(query_string, strlen(argv[1]) + 10);
+		sprintf(query_string, "q=%s", argv[1]);
+	}
+	else
+	{
+		query_string = (char*)UdmRealloc(query_string, 1024);
+		sprintf(query_string, "q=");
+	}
+	
+	/* Hack for Russian Apache from apache.lexa.ru  */
+	/* QUERY_STRING is already converted to server  */
+	/* character set. We must print original query  */
+	/* string instead however. Under usual apache   */ 
+	/* we'll use QUERY_STRING. Note that query_vars */
+	/* list will contain not unescaped values, so   */
+	/* we don't have to escape them when displaying */
+	env = getenv("CHARSET_SAVED_QUERY_STRING");
+	ParseQStringUnescaped(&query_vars,env?env:query_string);
+	
+	/* Unescape and save variables from QUERY_STRING */
+	/* Env->Vars will have unescaped values however  */
+	UdmParseQueryString(Agent,&Env->Vars,query_string);
 	
 	if((res=UdmTemplateLoad(Env,&Env->Vars,template_name,&tmpl))){
 		if(httpd)printf("Content-Type: text/plain\r\n\r\n");
 		printf("%s\n",Env->errstr);
 		UdmVarListFree(&tmpl);
+		UdmVarListFree(&query_vars);
 		UdmEnvFree(Env);
+		UDM_FREE(query_string);
+		UdmAgentFree(Agent);
 		return(0);
 	}
 	
@@ -300,6 +418,7 @@ int main(int argc, char ** argv){
 	UdmLog(Agent,UDM_LOG_ERROR,"search.cgi started with '%s'",template_name);
 	
 	/* This is for query tracking */
+	UdmVarListReplaceStr(&Env->Vars, "tmplt", template_name);
 	UdmVarListAddStr(&Env->Vars,"QUERY_STRING",query_string);
 	UdmVarListAddStr(&Env->Vars,"self",self);
 	env = getenv("REMOTE_ADDR");
@@ -311,32 +430,27 @@ int main(int argc, char ** argv){
 	Env->lcs=UdmGetCharSet(lcharset);
 	ppp=UdmVarListFindInt(&Env->Vars,"PagesPerScreen",10);
 	
-	if(httpd){
-		if(!Env->bcs){
-			printf("Content-Type: text/plain\r\n\r\n");
-			printf("Unknown BrowserCharset '%s' in template '%s'\n",bcharset,template_name);
-			exit(0);
-		}else if(!Env->lcs){
-			printf("Content-Type: text/plain\r\n\r\n");
-			printf("Unknown LocalCharset '%s' in template '%s'\n",lcharset,template_name);
-			exit(0);
-		}else{
-			printf("Content-type: text/html; charset=%s\r\n\r\n",bcharset);
-		}
-	}else{
-		if(!Env->bcs){
-			printf("Unknown BrowserCharset '%s' in template '%s'\n",bcharset,template_name);
-			exit(0);
-		}
-		if(!Env->lcs){
-			printf("Unknown LocalCharset '%s' in template '%s'\n",lcharset,template_name);
-			exit(0);
-		}
+	if (! Env->bcs) {
+		if (httpd) printf("Content-Type: text/plain\r\n\r\n");
+		printf("Unknown BrowserCharset '%s' in template '%s'\n",bcharset,template_name);
+		exit(0);
 	}
+
+	if(! Env->lcs) {
+		if (httpd) printf("Content-Type: text/plain\r\n\r\n");
+		printf("Unknown LocalCharset '%s' in template '%s'\n",lcharset,template_name);
+		exit(0);
+	}
+
+	if (UdmVarListFindInt(&Env->Vars, "cc", 0)) {
+		return(StoreDocCGI(Agent, Env, &tmpl));
+	}
+
+	if (httpd) printf("Content-type: text/html; charset=%s\r\n\r\n",bcharset);
 	
 	/* These parameters taken from "variable section of template"*/
 	
-	res=UdmVarListFindInt(&Env->Vars,"np",0)*UdmVarListFindInt(&Env->Vars,"ps",20);
+	res=UdmVarListFindInt(&Env->Vars,"np",0)*UdmVarListFindInt(&Env->Vars,"ps",10);
 	UdmVarListAddInt(&Env->Vars,"pn",res);
 	
 	catcolumns = (size_t)atoi(UdmVarListFindStr(&Env->Vars,"CatColumns",""));
@@ -351,7 +465,7 @@ int main(int argc, char ** argv){
 			char *catlist = NULL;
 			
 			for(i = 0; i < C.ncategories; i++) l += 128 + strlen(C.Category[i].path) + strlen(C.Category[i].name);
-			if (l > 0) catlist = (char*)malloc(l);
+			if (l > 0) catlist = (char*)UdmMalloc(l);
 			if (catlist != NULL) {
 			  sprintf(catlist, "<table>\n");
 			  for(i = 0; i < C.ncategories; i++){
@@ -387,7 +501,7 @@ int main(int argc, char ** argv){
 			size_t l = 0;
 			
 			for(i = 0; i < C.ncategories; i++) l += 32 + strlen(C.Category[i].path) + strlen(C.Category[i].name);
-			catpath = (char*)malloc(l);
+			catpath = (char*)UdmMalloc(l);
 			if (catpath != NULL) {
 			  catpath[0] = '\0';
 			  for(i = 0; i < C.ncategories; i++){
@@ -422,7 +536,10 @@ int main(int argc, char ** argv){
 	UdmVarListAddStr(&Env->Vars,"SearchTime",search_time);
 	WordInfo(Env,Res);
 	UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, &tmpl, "top");
-	
+#ifdef TRIAL_VER
+	fprintf(stdout,"%s",POWERED_LOGO);
+#endif
+
 	if(!Res->WWList.nwords){
 		UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, &tmpl, "noquery");
 		goto freeres;
@@ -433,7 +550,7 @@ int main(int argc, char ** argv){
 		goto freeres;
 	}
 	
-	page_size   = UdmVarListFindInt(&Agent->Conf->Vars,"ps",20);
+	page_size   = UdmVarListFindInt(&Agent->Conf->Vars,"ps",10);
 	page_number = UdmVarListFindInt(&Agent->Conf->Vars,"np",0);
 	
 	for (i = 0; i < Res->WWList.nwords; i++) {
@@ -448,9 +565,9 @@ int main(int argc, char ** argv){
 			}
 		}
 	}
-	storedstr = realloc(storedstr, storedlen = 1024 + 10 * swlen );
+	storedstr = UdmRealloc(storedstr, storedlen = 1024 + 10 * swlen );
 	
-	npages=(Res->total_found/(page_size?page_size:20)) + ((Res->total_found % (page_size?page_size:20) != 0 ) ?  1 : 0);
+	npages=(Res->total_found/(page_size?page_size:10)) + ((Res->total_found % (page_size?page_size:10) != 0 ) ?  1 : 0);
 	page1=page_number-ppp/2;
 	page2=page_number+ppp/2;
 	if(page1<0){
@@ -461,9 +578,9 @@ int main(int argc, char ** argv){
 		page1-=(page2-npages);
 		page2=npages;
 	}
-	if(page1<0)page1=page1=0;
+	if(page1<0)page1=0;
 	if(page2>npages)page2=npages;
-	nav = (char *)realloc(nav, nav_len = (size_t)(page2 - page1 + 2) * (1024 + 1024)); 
+	nav = (char *)UdmRealloc(nav, nav_len = (size_t)(page2 - page1 + 2) * (1024 + 1024)); 
 	                                                    /* !!! 1024 - limit for navbar0/navbar1 template size */ 
 	nav[0] = '\0';
 	
@@ -510,6 +627,7 @@ int main(int argc, char ** argv){
 	UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars,&tmpl, "restop");
 	
 	ExcerptSize = (size_t)UdmVarListFindInt(&Env->Vars, "ExcerptSize", 256);
+	ExcerptPadding = (size_t)UdmVarListFindInt(&Env->Vars, "ExcerptPadding", 40);
 	
 	for(i=0;i<Res->num_rows;i++){
 		UDM_DOCUMENT	*Doc=&Res->Doc[i];
@@ -528,19 +646,19 @@ int main(int argc, char ** argv){
 		/* Create "Alias" variable */
 		alcopy=UdmRemoveHiLightDup(UdmVarListFindStr(&Doc->Sections,"URL",""));
 		if((Alias=UdmMatchListFind(&Agent->Conf->Aliases,alcopy,0,NULL))){
-		        aliastr = (char*)malloc(strlen(Alias->arg) + strlen(alcopy) + 1);
+		        aliastr = (char*)UdmMalloc(strlen(Alias->arg) + strlen(alcopy) + 1);
 			sprintf(aliastr, "%s%s", Alias->arg, alcopy + strlen(Alias->pattern));
 		}else{
-			aliastr = (char*)strdup(alcopy);
+			aliastr = (char*)UdmStrdup(alcopy);
 		}
 		UdmVarListReplaceStr(&Doc->Sections,"Alias", aliastr);
-		free(aliastr);
-		free(alcopy);
+		UdmFree(aliastr);
+		UdmFree(alcopy);
 		
 		/* Skip clones */
 		if(dc_origin_id)continue;
 		
-		clist = (char*)malloc(2048); /* FIXME */
+		clist = (char*)UdmMalloc(2048); /* FIXME */
 		clist[0] = '\0';
 		clistsize = 0;
 		for(cl=0;cl<Res->num_rows;cl++){
@@ -549,12 +667,12 @@ int main(int argc, char ** argv){
 			
 			if ((dc_url_id == cl_origin_id) && cl_origin_id){
 				UDM_VARLIST	CloneVars;
-				
+
 				UdmVarListInit(&CloneVars);
 				UdmVarListAddLst(&CloneVars,&Env->Vars,NULL,"*");
 				UdmVarListReplaceLst(&CloneVars,&Clone->Sections,NULL,"*");
-				clist = (char*)realloc(clist, (clistsize = strlen(clist)) + 2048);
-				UdmTemplatePrint(Agent, NULL, clist + clistsize, 2048, &Env->Vars, &tmpl, "clone");
+				clist = (char*)UdmRealloc(clist, (clistsize = strlen(clist)) + 2048);
+				UdmTemplatePrint(Agent, NULL, clist + clistsize, 2048, &CloneVars, &tmpl, "clone");
 				UdmVarListFree(&CloneVars);
 			}
 		}
@@ -581,7 +699,7 @@ int main(int argc, char ** argv){
 			size_t c, l = 0;
 
 			for(c = 0; c < C.ncategories; c++) l += 32 + strlen(C.Category[c].path) + strlen(C.Category[c].name);
-			catpath = (char*)malloc(l);
+			catpath = (char*)UdmMalloc(l);
 			if (catpath != NULL) {
 			  *catpath = '\0';
 			  for(c = 0; c < C.ncategories; c++){
@@ -599,43 +717,23 @@ int main(int argc, char ** argv){
 		UdmVarListReplaceInt(&Doc->Sections, "URL_ID", UdmStrHash32(al));
 
 		u =  UdmVarListFindStr(&Env->Vars, "URL", "");
-		eu = (char*)malloc(strlen(u)*10 + 10);
+		eu = (char*)UdmMalloc(strlen(u)*10 + 10);
 		UdmEscapeURL(eu, u);
 		dm = UdmVarListFindStr(&Env->Vars, "Last-Modified", "");
-		edm = (char*)malloc(strlen(dm)*10 + 10);
+		edm = (char*)UdmMalloc(strlen(dm)*10 + 10);
 		UdmEscapeURL(edm, dm);
 
-		udm_snprintf(storedstr, storedlen, "%s?rec_id=%s&amp;DM=%s&amp;DS=%d&amp;L=%s&amp;CS=%s&amp;DU=%s&amp;CT=%s&amp;q=%s",
-			 UdmVarListFindStr(&Env->Vars, "StoredocURL", "/cgi-bin/storedoc.cgi"),
-			 UdmVarListFindStr(&Doc->Sections, "URL_ID", 0),
-			 edm,
-			 UdmVarListFindInt(&Env->Vars, "Content-Length", 0),
-			 UdmVarListFindStr(&Env->Vars,"Content-Language",""),
-			 UdmVarListFindStr(&Env->Vars,"Charset",""),
+		udm_snprintf(storedstr, storedlen, "?cc=1&amp;URL=%s&amp;q=%s",
 			 eu,
-			 UdmVarListFindStr(&Doc->Sections,"Content-Type",""),
 			 searchwords
 			 );
-		free(eu);
-		free(edm);
+		UdmFree(eu);
+		UdmFree(edm);
 
-		/* temporary skip large files */
-		/* Excerpt = (Doc->size < 200000) ? UdmExcerptDoc(Agent, Res, Doc, 256) : NULL;*/
-		
-		Excerpt = UdmExcerptDoc(Agent, Res, Doc, ExcerptSize);
-		
-		if ((Excerpt != NULL) && (strlen(Excerpt) > 6)) {
-			  char *HlExcerpt = UdmHlConvert(&Res->WWList, Excerpt, Env->lcs, Env->bcs);
-			  UdmVarListReplaceStr(&Env->Vars,"body",HlExcerpt);
-			  UDM_FREE(HlExcerpt);
-			}
-		if (Excerpt != NULL && (UdmVarListFindStr(&Doc->Sections, "Z", NULL) == NULL)) {
-			UdmVarListReplaceInt(&Env->Vars,"ST",1);
-			UdmVarListReplaceStr(&Env->Vars, "stored_href", storedstr);
-		} else {
-			UdmVarListReplaceInt(&Env->Vars,"ST",0);
-			UdmVarListReplaceStr(&Env->Vars, "stored_href", "");
-		}
+		UdmVarListReplaceStr(&Env->Vars, "stored_href", UdmVarListFindStr(&Doc->Sections, "CachedCopy", NULL) ? storedstr : "");
+
+		Excerpt = UdmExcerptDoc(Agent, Res, Doc, ExcerptSize, ExcerptPadding);
+		if (Excerpt) UdmVarListReplaceStr(&Env->Vars, "body", Excerpt);
 		UDM_FREE(Excerpt);
 		
 		if ( (sc = UdmVarListFindInt(&Env->Vars, "site", 0)) == 0) {
@@ -645,10 +743,6 @@ int main(int argc, char ** argv){
 		  UdmVarListReplaceStr(&Env->Vars, "sitelimit_href", url);
 		}
 
-		if (Res->PerSite) {
-		  UdmVarListReplaceUnsigned(&Env->Vars, "PerSite", Res->PerSite[i]);
-		}
-
 		UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, &tmpl, "res");
 		
 		/* Revoke all found user-defined sections */
@@ -657,7 +751,7 @@ int main(int argc, char ** argv){
 			UdmVarListDel(&Env->Vars, S->name);
 		}
 		
-		free(clist);
+		UdmFree(clist);
 	}
 	UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, &tmpl, "resbot");
 	UDM_FREE(searchwords);
