@@ -28,10 +28,11 @@
 #include <string.h>
 #include <sys/types.h>
 
+#include "udm_common.h"
 #include "udm_search_tl.h"
 #include "udm_utils.h"
 
-extern int tz_offset;
+extern long tz_offset;
 
 /******************
  * Data definitions 
