@@ -173,7 +173,7 @@ static int do_client(UDM_ENV * Conf, int client){
 		}
 		switch(hdr.cmd){
 			case UDM_SEARCHD_CMD_DOCINFO:
-				dinfo = (char*)realloc(dinfo, hdr.len + 1);
+				dinfo = (char*)UdmRealloc(dinfo, hdr.len + 1);
 				nrecv = UdmRecvall(client, dinfo, hdr.len);
 				if((size_t)nrecv != hdr.len){
 					UdmLog(Agent,verb,"Received incomplete data nbytes=%d nrecv=%d",hdr.len,(int)nrecv);
@@ -186,7 +186,7 @@ static int do_client(UDM_ENV * Conf, int client){
 				tok = udm_strtok_r(dinfo, "\r\n", &lt);
 				
 				while(tok){
-					Res->Doc = (UDM_DOCUMENT*)realloc(Res->Doc, sizeof(UDM_DOCUMENT) * (ndocs + 1));
+					Res->Doc = (UDM_DOCUMENT*)UdmRealloc(Res->Doc, sizeof(UDM_DOCUMENT) * (ndocs + 1));
 					UdmDocInit(&Res->Doc[ndocs]);
 					UdmDocFromTextBuf(&Res->Doc[ndocs], tok);
 					
@@ -226,12 +226,12 @@ static int do_client(UDM_ENV * Conf, int client){
 					ulen=strlen(textbuf)+2;
 					olen=dlen;
 					dlen=dlen+ulen;
-					dinfo=(char*)realloc(dinfo,dlen+1);
+					dinfo=(char*)UdmRealloc(dinfo,dlen+1);
 					dinfo[olen]='\0';
 					sprintf(dinfo+olen,"%s\r\n",textbuf);
 				}
 				
-				if(!dinfo) dinfo = (char*)strdup("nodocinfo");
+				if(!dinfo) dinfo = (char*)UdmStrdup("nodocinfo");
 				
 				hdr.cmd=UDM_SEARCHD_CMD_DOCINFO;
 				hdr.len=strlen(dinfo);
@@ -259,6 +259,7 @@ static int do_client(UDM_ENV * Conf, int client){
 			      done = 1;
 			      break;
 			    }
+			    buf[hdr.len] = 0;
 			    sscanf(buf, "%i %i", &crc32, &origin_id);
 			    D = UdmDocInit(NULL);
 			    UdmVarListAddInt(&D->Sections, "crc32", crc32);
@@ -285,12 +286,12 @@ static int do_client(UDM_ENV * Conf, int client){
 					ulen=strlen(textbuf)+2;
 					olen=dlen;
 					dlen=dlen+ulen;
-					dinfo=(char*)realloc(dinfo,dlen+1);
+					dinfo=(char*)UdmRealloc(dinfo,dlen+1);
 					dinfo[olen]='\0';
 					sprintf(dinfo+olen,"%s\r\n",textbuf);
 			    }
 			    
-			    if (!dinfo) dinfo = (char*)strdup("nocloneinfo");
+			    if (!dinfo) dinfo = (char*)UdmStrdup("nocloneinfo");
 				
 			    hdr.cmd = UDM_SEARCHD_CMD_DOCINFO;
 			    hdr.len = strlen(dinfo); 
@@ -315,7 +316,7 @@ static int do_client(UDM_ENV * Conf, int client){
 				UdmCatAction(Agent, &Cat, cmd);
 				
 				dlen = Cat.ncategories * 1024;
-				dinfo = (char*)malloc(dlen);
+				dinfo = (char*)UdmMalloc(dlen);
 				UdmCatToTextBuf(&Cat, dinfo, dlen);
 				
 				hdr.cmd = UDM_SEARCHD_CMD_CATINFO;
@@ -348,7 +349,7 @@ static int do_client(UDM_ENV * Conf, int client){
 			  break;
 				
 			case UDM_SEARCHD_CMD_WORDS:
-			        words = (char*)realloc(words, hdr.len + 1);
+			        words = (char*)UdmRealloc(words, hdr.len + 1);
 				if (words == NULL) {
 					udm_snprintf(buf, sizeof(buf)-1, "Can't alloc memory for query");
 					UdmLog(Agent, verb, "Can't alloc memory for query");
@@ -483,7 +484,7 @@ static void usage(void){
 	fprintf(stderr,
 "\n\
 searchd from %s-%s-%s\n\
-http://www.mnogosearch.org (C) 1998-2002, LavTech Corp.\n\
+http://www.mnogosearch.org (C)1998-2004, LavTech Corp.\n\
 \n\
 Usage: searchd [OPTIONS]\n\
 \n\
