@@ -86,19 +86,13 @@ static int cmpword(UDM_URL_CRD *s1,UDM_URL_CRD *s2){
 	return 0;
 }
 
-static int cmpurlid(UDM_URL_CRD *s1,UDM_URL_CRD *s2){
-	if (s1->url_id > s2->url_id) return 1;
-	if (s1->url_id < s2->url_id) return -1;
-	{
-	  int pos1, pos2;
-	  if ((pos1 = UDM_WRDPOS(s1->coord)) > (pos2 = UDM_WRDPOS(s2->coord))) return 1;
-	  if (pos1 < pos2) return -1;
-	}
-	return 0;
+static int cmpurlid (UDM_URL_CRD *s1, UDM_URL_CRD *s2) {
+	if (s1->url_id > s2->url_id) return(1);
+	if (s1->url_id < s2->url_id) return(-1);
+	return(UDM_WRDPOS(s1->coord) - UDM_WRDPOS(s2->coord));
 }
 
 static int cmpsiteid(UDM_URLCRDLIST *L, UDM_URL_CRD *C, UDM_URLDATA *D, long j){
-
         if (D->site_id > L->Data[j].site_id) return 1;
 	if (D->site_id < L->Data[j].site_id) return -1;
 	if (C->coord > L->Coords[j].coord) return 1;
@@ -109,7 +103,6 @@ static int cmpsiteid(UDM_URLCRDLIST *L, UDM_URL_CRD *C, UDM_URLDATA *D, long j){
 }
 
 static int cmppattern(UDM_URLCRDLIST *L, UDM_URL_CRD *C, UDM_URLDATA *D, long j, const char *pattern) {
-  
   for(; *pattern != '\0'; pattern++) {
     switch(*pattern) {
     case 'R':
@@ -153,15 +146,15 @@ void UdmSortSearchWordsByURL(UDM_URL_CRD *wrd,size_t num){
 static size_t UdmH[] = {1, 5, 19, 41, 109, 209, 505, 929, 2161, 3905, 8929, 16001, 36289, 64769};
 
 void UdmSortSearchWordsBySite(UDM_URLCRDLIST *L, size_t num){
-        long h, i, j;
+        register ssize_t h, i, j;
 	int s = 13;
 	UDM_URL_CRD Crd;
 	UDM_URLDATA Dat;
 
-	while(num / 3 < UdmH[s]) s--;
+	while((s > 0) && ((num / 3) < UdmH[s])) s--;
 	while(s >= 0) {
-	  h = (long)UdmH[s];
-	  for (j = h; j < (long)num; j++) {
+	  h = UdmH[s];
+	  for (j = h; j < (ssize_t)num; j++) {
 	    Crd = L->Coords[j];
 	    Dat = L->Data[j];
 
@@ -185,16 +178,16 @@ DD6:
 
 
 void UdmSortSearchWordsByPattern(UDM_RESULT *Res, UDM_URLCRDLIST *L, size_t num, const char *pattern) {
-        long h, i, j;
+        register ssize_t h, i, j;
 	int s = 13;
 	UDM_URL_CRD Crd;
 	UDM_URLDATA Dat;
 	size_t Cnt = 1;
 
-	while(num / 3 < UdmH[s]) s--;
+	while( (s > 0) && ((num / 3) < UdmH[s])) s--;
 	while(s >= 0) {
-	  h = (long)UdmH[s];
-	  for (j = h; j < (long)num; j++) {
+	  h = UdmH[s];
+	  for (j = h; j < (ssize_t)num; j++) {
 	    Crd = L->Coords[j];
 	    Dat = L->Data[j];
 	    if (Res->PerSite) Cnt = Res->PerSite[j];
@@ -276,8 +269,9 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 	int * ustr, * lt, * lex;
 	size_t ulen;
 	int word_match   = UdmVarListFindInt(&query->Conf->Vars,"wm",UDM_MATCH_FULL);
-	int search_mode = UdmSearchMode(UdmVarListFindStr(&query->Conf->Vars, "m", "all"));
+/*	int search_mode = UdmSearchMode(UdmVarListFindStr(&query->Conf->Vars, "m", "all")); */
 	const char * txt = UdmVarListFindStr(&query->Conf->Vars,"q","");
+	const char * qlang = UdmVarListFindStr(&query->Conf->Vars, "g", NULL);
 	char *ltxt;
 	size_t i, wlen, llen, nphrasecmd = 0;
 	char *wrd, *clex;
@@ -285,8 +279,8 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 	UDM_CONV uni_lc, bc_uni;
 	const char *lang;
 	
-	if ((wrd = (char*)malloc(query->Conf->WordParam.max_word_len * 12 + 1)) == NULL) return 0;
-	if ((uwrd = (int*)malloc(sizeof(int) * (query->Conf->WordParam.max_word_len + 1))) == NULL) { UDM_FREE(wrd); return 0; }
+	if ((wrd = (char*)UdmMalloc(query->Conf->WordParam.max_word_len * 12 + 1)) == NULL) return 0;
+	if ((uwrd = (int*)UdmMalloc(sizeof(int) * (query->Conf->WordParam.max_word_len + 1))) == NULL) { UDM_FREE(wrd); return 0; }
 
 
 	if (!(browser_cs = query->Conf->bcs)) {
@@ -305,13 +299,13 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 	UdmConvInit(&uni_lc,sys_int,local_cs,UDM_RECODE_HTML);
 	
 	ulen=strlen(txt);
-	ustr=(int*)(malloc((sizeof(int))*(ulen+1)));
+	ustr=(int*)(UdmMalloc((sizeof(int))*(ulen+1)));
 	UdmConv(&bc_uni,(char*)ustr,sizeof(ustr[0])*(ulen+1),txt,ulen+1);
 	for(i = 0; ustr[i] && (i < ulen); i++) if (ustr[i] == '"') ustr[i] = '\'';
 	
 	/* Create copy of query, converted into LocalCharset (for UdmTrack) */
 	llen = ulen * 14 + 1;
-	ltxt=(char*)malloc(llen);
+	ltxt=(char*)UdmMalloc(llen);
 	UdmConv(&uni_lc,ltxt,llen,(char*)ustr,bc_uni.obytes);
 	ltxt[uni_lc.obytes]='\0';
 	UdmVarListReplaceStr(&query->Conf->Vars,"q",ltxt);  /* "q-lc" was here */
@@ -367,7 +361,7 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 			Res->ncmds++;
 			if (Res->nitems >= Res->mitems) {
 			  Res->mitems += UDM_MAXSTACK;
-			  Res->items = (UDM_STACK_ITEM*)realloc(Res->items, Res->mitems * sizeof(UDM_STACK_ITEM));
+			  Res->items = (UDM_STACK_ITEM*)UdmRealloc(Res->items, Res->mitems * sizeof(UDM_STACK_ITEM));
 			}
 		}
 	  } else {
@@ -378,13 +372,13 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 			Res->nitems++;
 			if (Res->nitems >= Res->mitems) {
 			  Res->mitems += UDM_MAXSTACK;
-			  Res->items = (UDM_STACK_ITEM*)realloc(Res->items, Res->mitems * sizeof(UDM_STACK_ITEM));
+			  Res->items = (UDM_STACK_ITEM*)UdmRealloc(Res->items, Res->mitems * sizeof(UDM_STACK_ITEM));
 			}
 
 			if(word_match==UDM_MATCH_FULL){
 				/* Check stopword only when full word         */
 				/* Substring searches should not exclude them */
-				if(UdmStopListFind(&query->Conf->StopWords, wrd, NULL)||
+				if(UdmStopListFind(&query->Conf->StopWords, wrd, qlang) ||
 						(query->Conf->WordParam.min_word_len>wlen)||
 						(query->Conf->WordParam.max_word_len<wlen)){
 				  UDM_WIDEWORD OWord;
@@ -403,6 +397,16 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 			if(Res->WWList.nuniq >= UDM_MAXWORDPERQUERY-1){
 				addwrd=0;
 			}
+
+			if (addwrd) {
+				size_t nw;
+				for (nw = 0; nw < Res->WWList.nwords; nw++) {
+					if (UdmUniStrCmp(Res->WWList.Word[nw].uword, uwrd)) continue;
+					addwrd = 0;
+					break;
+				}
+			}
+
 			if(addwrd){
 				UDM_WIDEWORD OWord;
 				UDM_WIDEWORDLIST * forms;
@@ -414,8 +418,8 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 				OWord.word=wrd;
 				OWord.uword=uwrd;
 				OWord.origin = UDM_WORD_ORIGIN_QUERY;
+
 				UdmWideWordListAdd(&Res->WWList, &OWord);
-				
 				if((forms=UdmAllForms(query,&OWord))){
 					UDM_WIDEWORD FWord;
 					size_t frm;
@@ -430,12 +434,12 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 						FWord.word=wrd;
 						FWord.uword=forms->Word[frm].uword;
 						FWord.origin = forms->Word[frm].origin;
-						
+
 						UdmWideWordListAdd(&Res->WWList,&FWord);
 /*						UdmLog(query, UDM_LOG_DEBUG, "Word form: [%d] %s", FWord.origin, wrd);*/
 					}
 					UdmWideWordListFree(forms);
-					UDM_FREE(forms);
+	    				UDM_FREE(forms);
 				}
 				Res->WWList.nuniq++;
 			}
@@ -449,7 +453,7 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 	  Res->ncmds++;
 	  if (Res->nitems >= Res->mitems) {
 	    Res->mitems += UDM_MAXSTACK;
-	    Res->items = (UDM_STACK_ITEM*)realloc(Res->items, Res->mitems * sizeof(UDM_STACK_ITEM));
+	    Res->items = (UDM_STACK_ITEM*)UdmRealloc(Res->items, Res->mitems * sizeof(UDM_STACK_ITEM));
 	  }
 	}
 	UDM_FREE(ustr); UDM_FREE(uwrd); UDM_FREE(wrd);
@@ -459,20 +463,26 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res){
 
 static uint4 UdmCalcCosineWeight(unsigned int *R, unsigned int *D, size_t nw, size_t ns, size_t nadd) {
   double x = 0.0, y = 0.0, xy = 0.0;
-  size_t i, j;
+  size_t i, j, c;
+  double TR, TD;
 
   for (i = 0; i < nadd; i++) {
-    x += (double) R[i] * (double)R[i];
-    y += (double) D[i] * (double)D[i];
-    xy += (double)R[i] * (double)D[i];
+    TR = (double)R[i]; TD = (double)D[i];
+    x += TR * TR;
+    y += TD * TD;
+    xy += TR * TD;
   }
 
-  for (j = 0; j < ns; j++)
+  for (j = 0; j < ns; j++) {
+    c = nadd + j;
     for (i = 0; i < nw; i++) {
-      x += (double) R[nadd + i * 256 + j] * (double)R[nadd + i * 256 + j];
-      y += (double) D[nadd + i * 256 + j] * (double)D[nadd + i * 256 + j];
-      xy += (double)R[nadd + i * 256 + j] * (double)D[nadd + i * 256 + j];
+      TR = (double)R[c]; TD = (double)D[c];
+      x += TR * TR;
+      y += TD * TD;
+      xy += TR * TD;
+      c += 256;
     }
+  }
 
 /*  fprintf(stderr, "R: %.4f  D: %.4f  RD: %.4f   cos: %.8f [%d]\n", x, y, xy,
 	   xy / sqrt(x) / sqrt(y),
@@ -504,6 +514,11 @@ static unsigned int CalcAverageWordDistance(UDM_PHR_DAT *phr, size_t num) {
   return (np) ? (sum / np) : 0;
 }
 
+
+/*
+#define DEBUG_REL 1
+*/
+
 void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res){	
 	UDM_STACK_ITEM *temp_items;
 	size_t	i, j = 0, D_size, nitems = Res->nitems;
@@ -520,16 +535,176 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res){
 	
 	if(!Res->CoordList.ncoords) return;
 
+#ifdef NEW_REL
+  {
+    urlid_t prev_id;
+    size_t prev_num= 0;
+    size_t N= 0;
+    size_t M= Res->WWList.nuniq;
+    size_t ndoc= 0; 
+    
+    float  TFi[UDM_MAXWORDPERQUERY+1];
+    float   Ci[UDM_MAXWORDPERQUERY+1];
+    float   Vi[UDM_MAXWORDPERQUERY+1];    
+    float   Wi[UDM_MAXWORDPERQUERY+1];
+    float   Di[UDM_MAXWORDPERQUERY+1];
+    
+    bzero(TFi, sizeof(TFi));
+    bzero(Ci, sizeof(Ci));
+    bzero(Vi, sizeof(Vi));
+    bzero(Di, sizeof(Di));
+    
+    
+    /* N - number of documents */
+    for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
+    {
+      if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
+      {
+        N++;
+        if (i < Res->CoordList.ncoords)
+        {
+          prev_id= Crd[i].url_id;
+          prev_num= i;
+        }
+      }
+
+#ifdef DEBUG_REL
+      if (i < Res->CoordList.ncoords)
+      {
+        fprintf(stderr,"[%d]doc=%d sec=%d wrd=%d pos=%d\n",
+          i,Crd[i].url_id,
+          UDM_WRDSEC(Crd[i].coord),
+          UDM_WRDNUM(Crd[i].coord),
+          UDM_WRDPOS(Crd[i].coord));
+      }
+#endif
+    }
+    
+    
+    /* TFi */
+    for (i=0; i < Res->CoordList.ncoords; i++)
+    {
+      TFi[UDM_WRDNUM(Crd[i].coord)]++;
+    }
+    
+    /* Add fictious word */
+    TFi[M]= ((float) Res->CoordList.ncoords) / M;
+    M++;
+    
+    
+    
+    /* Ci */
+    for (i=0; i < M; i++)
+    {
+       Ci[i]= ((float) TFi[i]) / N;
+#ifdef DEBUG_REL
+       fprintf(stderr,"[%d] TDi=%d Ci=%f\n",i,TFi[i], Ci[i]);
+#endif
+    }
+    
+    
+    /* Vi */
+    for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
+    {
+      size_t k;
+      
+      if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
+      {
+        Di[M-1]= TFi[M-1]/N; /* fictious word */
+        for (k=0; k < M; k++)
+        {
+          Vi[k]+= (Di[k] - Ci[k]) * (Di[k] - Ci[k]);
+        }
+        bzero(Di, sizeof(Di));
+        
+        if (i < Res->CoordList.ncoords)
+        {
+          prev_id= Crd[i].url_id;
+          prev_num= i;
+        }
+      }
+      if (i < Res->CoordList.ncoords)
+        Di[UDM_WRDNUM(Crd[i].coord)]++;
+    }
+    
+    
+    for (i=0; i < M; i++)
+    {
+      Vi[i]/= (N-1);
+      Wi[i]= Ci[i] ? TFi[i]*Vi[i]/Ci[i]/Ci[i] : 0;
+#ifdef DEBUG_REL
+      fprintf(stderr,"[%d] Vi=%f Wi=%f\n", i, Vi[i],Wi[i]);
+#endif
+    }
+    
+    
+    bzero(Di,sizeof(Di));
+    for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
+    {
+      size_t k;
+      
+      if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
+      {
+        float sumWiDi= 0;
+        float sumDi2= 0;
+        float sumWi2= 0;
+        float res;
+        size_t n= 1;
+        
+        Di[M-1]= TFi[M-1]/N;  /* fictious word */
+        
+        for (k=0; k < M; k++)
+        {
+          sumWiDi+= Wi[k] * Di[k] / (float)M;
+          sumDi2+=  Di[k] * Di[k] / (float)M;
+          sumWi2+=  Wi[k] * Wi[k];
+          if (k + 1 < M)
+            n*= Di[k];
+        }
+        
+        res= sumWiDi / sqrt(sumDi2 * sumWi2);
+        Crd[ndoc].url_id= prev_id;
+        Crd[ndoc].coord= res * 100000;
+        
+#ifdef DEBUG_REL
+        fprintf(stderr,"RES[%d]=%f %f %f\n",
+          prev_id, res, sumWiDi, sqrt(sumDi2 * sumWi2));
+#endif 
+        
+        if (n)
+          ndoc++;
+        
+        bzero(Di, sizeof(Di));
+        if (i < Res->CoordList.ncoords)
+        {
+          prev_id= Crd[i].url_id;
+          prev_num= i;
+        }
+      }
+      
+      if (i <= Res->CoordList.ncoords)
+        Di[UDM_WRDNUM(Crd[i].coord)]++;
+    }
+    
+    //fprintf(stderr,"  : %d\n", M);
+    //fprintf(stderr,"  : %d\n", N);
+    
+    Res->CoordList.ncoords= ndoc;
+    return;
+    
+  }
+#endif
+
 	UdmWeightFactorsInit(UdmVarListFindStr(&query->Conf->Vars, "wf", ""), wf);
-	phr = (UDM_PHR_DAT*)malloc(phr_m * sizeof(UDM_PHR_DAT));
-	count = (size_t*)malloc(count_size);
+	phr = (UDM_PHR_DAT*)UdmMalloc(phr_m * sizeof(UDM_PHR_DAT));
+	count = (size_t*)UdmMalloc(count_size);
 	D_size = (Res->WWList.nwords * 256 + 1) * sizeof(unsigned int) + 1;
-	if ((R = (unsigned int*)malloc(D_size)) == NULL) return;
-	if ((D = (unsigned int*)malloc(D_size)) == NULL) {
+	if ((R = (unsigned int*)UdmMalloc(D_size)) == NULL) return;
+	if ((D = (unsigned int*)UdmMalloc(D_size)) == NULL) {
 	  UDM_FREE(R);
 	  return;
 	}
-	if ((temp_items = (UDM_STACK_ITEM*)malloc((Res->nitems + 1) * sizeof(UDM_STACK_ITEM))) == NULL) {
+	if ((temp_items = (UDM_STACK_ITEM*)UdmMalloc((Res->nitems + 1) * sizeof(UDM_STACK_ITEM))) == NULL) {
 	  UDM_FREE(R); UDM_FREE(D);
 	  return;
 	}
@@ -580,7 +755,7 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res){
 			count[phr[phr_n].order]++;
 			if (++phr_n >= phr_m) {
 			  phr_m += 256;
-			  phr = (UDM_PHR_DAT*)realloc(phr, phr_m * sizeof(UDM_PHR_DAT));
+			  phr = (UDM_PHR_DAT*)UdmRealloc(phr, phr_m * sizeof(UDM_PHR_DAT));
 			}
 
 			D[1 + 256 * UDM_WRDNUM(Crd[i].coord) + UDM_WRDSEC(Crd[i].coord)] = 
@@ -706,7 +881,7 @@ void UdmGroupBySite(UDM_AGENT *query, UDM_RESULT *Res){
 	UDM_URLDATA *Dat = Res->CoordList.Data;
 	
 	if(!Res->CoordList.ncoords) return;
-	if ((Res->PerSite = (size_t*)malloc(sizeof(*Res->PerSite) * Res->CoordList.ncoords)) == NULL) return;
+	if ((Res->PerSite = (size_t*)UdmMalloc(sizeof(*Res->PerSite) * Res->CoordList.ncoords)) == NULL) return;
 
 	Doc_site = Dat[0].site_id;
 
@@ -738,13 +913,63 @@ typedef struct{
 	uint4	weight;
 } UDM_PHR_PAR;
 
+/******** Convert category string into 32 bit number *************/
+void UdmDecodeHex8Str(const char *hex_str, uint4 *hi, uint4 *lo, uint4 *fhi, uint4 *flo){
+  char str[33],str_hi[17],str_lo[17], *s = str;
 
+        strncpy(str, hex_str, 13);
+	str[12] = '\0';
+	strcat(str,"000000000000");
+	while(*s == '0') *s++ = ' ';
+	strncpy(str_hi,&str[0],6); str_hi[6]=0;
+	strncpy(str_lo,&str[6],6); str_lo[6]=0;
+	
+	*hi = (uint4)strtoul(str_hi, (char **)NULL, 36);
+	*lo = (uint4)strtoul(str_lo, (char **)NULL, 36);
+
+	if ((fhi != NULL) && (flo != NULL)) {
+	  strncpy(str, hex_str, 13);
+	  str[12] = '\0';
+	  strcat(str,"ZZZZZZZZZZZZ");
+	  strncpy(str_hi, &str[0], 6); str_hi[6] = 0;
+	  strncpy(str_lo, &str[6], 6); str_lo[6] = 0;
+	
+	  *fhi = strtoul(str_hi, (char **)NULL, 36);
+	  *flo = strtoul(str_lo, (char **)NULL, 36);
+
+	}
+}
+
+int __UDMCALL UdmAddSearchLimit(UDM_AGENT *Agent, int type, const char *file_name, const char *val){
+	uint4 hi, lo, f_hi, f_lo;
+	
+	if(Agent->nlimits == MAX_SEARCH_LIMIT - 1) return(1);
+	
+	Agent->limits[Agent->nlimits].type = type;
+	strcpy(Agent->limits[Agent->nlimits].file_name, file_name);
+	switch(type){
+		case 0: UdmDecodeHex8Str(val, &hi, &lo, &f_hi, &f_lo); break;
+		case 1: f_hi = hi = 0; f_lo = lo = 0; break;
+		case 2: hi=atoi(val); lo=0; f_hi = hi; f_lo = lo; break;
+		case 3: hi=UdmStrHash32(val); lo = 0; f_hi = hi; f_lo = 0; break;
+	}	
+	Agent->limits[Agent->nlimits].hi = hi;
+	Agent->limits[Agent->nlimits].lo = lo;
+	Agent->limits[Agent->nlimits].f_hi = f_hi;
+	Agent->limits[Agent->nlimits].f_lo = f_lo;
+	
+	Agent->nlimits++;
+
+	UdmLog(Agent, UDM_LOG_DEBUG, "val: %s  %x %x   %x %x", val, hi, lo, f_hi, f_lo);
+	
+	return(0);
+}
 
 int UdmParseQueryString(UDM_AGENT * Agent,UDM_VARLIST * vars,char * query_string){
 	char * tok, *lt;
 	size_t len;
-	char *str = (char *)malloc((len = strlen(query_string)) + 7);
-	char *qs = (char*)strdup(query_string);
+	char *str = (char *)UdmMalloc((len = strlen(query_string)) + 7);
+	char *qs = (char*)UdmStrdup(query_string);
 	char qname[256];
 
 	if ((str == NULL) || qs == NULL) {
@@ -825,8 +1050,8 @@ char * UdmHlConvert(UDM_WIDEWORDLIST *List,const char * src, UDM_CHARSET * lcs,
 	if(!src)return NULL;
 	
 	if ((len = strlen(src)) == 0) return NULL;
-	hpart = (char*)malloc(len * 14 + 10);
-	htxt = (char*)malloc(len * 14 + 10);
+	hpart = (char*)UdmMalloc(len * 14 + 10);
+	htxt = (char*)UdmMalloc(len * 14 + 10);
 	htxt[0]='\0';
 	
 	sys_int=UdmGetCharSet("sys-int");
@@ -834,7 +1059,7 @@ char * UdmHlConvert(UDM_WIDEWORDLIST *List,const char * src, UDM_CHARSET * lcs,
 	UdmConvInit(&uni_bc,sys_int,bcs,UDM_RECODE_HTML);
 	
 	/* Convert to unicode */
-	uni = (int *)malloc((len + 10) * sizeof(int));
+	uni = (int *)UdmMalloc((len + 10) * sizeof(int));
 	UdmConv(&lc_uni,(char*)uni,sizeof(uni[0])*(len+10),src,len+1);
 	
 	/* Parse unicode string */
@@ -859,8 +1084,8 @@ char * UdmHlConvert(UDM_WIDEWORDLIST *List,const char * src, UDM_CHARSET * lcs,
 		/* Check that it is word to be marked */
 		if (List != NULL) {
 		  for(uw=0;uw<List->nwords;uw++){
-			slen = List->Word[uw].len;
-			if((flen==slen)&&(!UdmUniStrNCaseCmp(tok,List->Word[uw].uword,slen))){
+			slen = List->Word[uw].ulen;
+			if(flen == slen && !UdmUniStrNCaseCmp(tok,List->Word[uw].uword,flen)){
 				found=1;
 				break;
 			}
@@ -895,7 +1120,7 @@ int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res, UDM_CHARSET *lcs, UDM_CHARSET *bc
 		UDM_WIDEWORD	*W=&Res->WWList.Word[i];
 
 		len = strlen(W->word);
-		newval = (char*)malloc(len * 12 + 1);
+		newval = (char*)UdmMalloc(len * 12 + 1);
 		UdmConv(&lc_bc,newval,len*12+1,W->word,len+1);
 		UDM_FREE(W->word);
 		W->word=newval;
@@ -910,7 +1135,7 @@ int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res, UDM_CHARSET *lcs, UDM_CHARSET *bc
 			Var = &D->Sections.Var[sec];
 
 			len = strlen(Var->val);
-			newtxt = (char*)malloc(len * 12 + 1);
+			newtxt = (char*)UdmMalloc(len * 12 + 1);
 
 			UdmConv(&lc_bc_text, newtxt, len * 12 + 1, Var->val, len + 1);
 			newval = UdmHlConvert(&Res->WWList, Var->val, lcs, bcs);
@@ -925,8 +1150,8 @@ int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res, UDM_CHARSET *lcs, UDM_CHARSET *bc
 	  for (i = 0; i < Conf->Vars.nvars; i++) {
 	        Var = &Conf->Vars.Var[i];
 	        len = strlen(Var->val);
-		newtxt = (char*)malloc(len * 12 + 1);
-		newval = (char*)malloc(len * 12 + 1);
+		newtxt = (char*)UdmMalloc(len * 12 + 1);
+		newval = (char*)UdmMalloc(len * 12 + 1);
 
 /*		if (db->DBDriver != UDM_DB_SEARCHD)*/  /* FIXME: need unification in charset from different UDM_DB */
 		  UdmConv(&lc_bc, newval, len * 12 + 1, Var->val, len + 1);
@@ -944,7 +1169,7 @@ int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res, UDM_CHARSET *lcs, UDM_CHARSET *bc
 
 char* UdmRemoveHiLightDup(const char *s){
 	size_t	len=strlen(s)+1;
-	char	*res=malloc(len);
+	char	*res = (char*)UdmMalloc(len);
 	char	*d;
 	
 	for(d=res;s[0];s++){
@@ -994,14 +1219,14 @@ int UdmCatFromTextBuf(UDM_CATEGORY *C, char *textbuf) {
 	if(!htok || tag.type != UDM_HTML_TAG)
 	  return UDM_OK;
 
-	C->Category = (UDM_CATITEM*)realloc(C->Category, sizeof(UDM_CATITEM) * ((c = C->ncategories) + 1));
+	C->Category = (UDM_CATITEM*)UdmRealloc(C->Category, sizeof(UDM_CATITEM) * ((c = C->ncategories) + 1));
 	bzero((void*)&C->Category[c], sizeof(UDM_CATITEM));
 	
 	for(i = 1; i < tag.ntoks; i++){
 		size_t	nlen = tag.toks[i].nlen;
 		size_t	vlen = tag.toks[i].vlen;
-		char	*name = strndup(tag.toks[i].name, nlen);
-		char	*data = strndup(tag.toks[i].val, vlen);
+		char	*name = UdmStrndup(tag.toks[i].name, nlen);
+		char	*data = UdmStrndup(tag.toks[i].val, vlen);
 
 		if (!strcmp(name, "id")) {
 		  C->Category[c].rec_id = atoi(data);
@@ -1024,7 +1249,8 @@ int UdmCatFromTextBuf(UDM_CATEGORY *C, char *textbuf) {
 int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang) {
 	UDM_CHARSET     *tis_cs;
 	UDM_CONV        uni_tis, tis_uni;
-	char            *tistr, *tistr_seg;
+	/* FIXME: Remove if not needed: */
+/*	char            *tistr, *tistr_seg; */
 #ifdef CHASEN
 	UDM_CHARSET     *eucjp_cs;
 	UDM_CONV        uni_eucjp, eucjp_uni;
@@ -1038,7 +1264,7 @@ int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang) {
 #ifdef HAVE_CHARSET_gb2312
 	UDM_CHARSET     *gb_cs;
 	UDM_CONV        uni_gb, gb_uni;
-	char            *gbstr, *gbstr_seg;
+/*	char            *gbstr, *gbstr_seg; */
 #endif
 	UDM_CHARSET	*sys_int;
 	size_t          reslen, dstlen = UdmUniLen(ustr);
@@ -1072,7 +1298,7 @@ int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang) {
 #ifdef CHASEN
 
 	if (lang == NULL || !strncasecmp(lang, "ja", 2)) {
-	  eucstr = (char*)malloc(12 * dstlen + 1);
+	  eucstr = (char*)UdmMalloc(12 * dstlen + 1);
 	  UdmConv(&uni_eucjp, eucstr, 12 * dstlen + 1, (char*)ustr, dstlen + 1);
 
 	  UDM_GETLOCK(Indexer, UDM_LOCK_SEGMENTER);
@@ -1080,7 +1306,7 @@ int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang) {
 	  UDM_RELEASELOCK(Indexer, UDM_LOCK_SEGMENTER);
 
 	  reslen = strlen(eucstr_seg) + 1;
-	  ustr = (int*)realloc(ustr, reslen * sizeof(int));
+	  ustr = (int*)UdmRealloc(ustr, reslen * sizeof(int));
 	  UdmConv(&eucjp_uni, (char*)ustr, reslen * sizeof(int), eucstr_seg, reslen);
 	  UDM_FREE(eucstr);
 	  dstlen = UdmUniLen(ustr);
@@ -1088,8 +1314,8 @@ int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang) {
 #endif
 #ifdef MECAB
 	if (lang == NULL || !strncasecmp(lang, "ja", 2)) {
-	  sjisstr = (char*)malloc(12 * dstlen + 1);
-	  UdmConv(&uni_sjis, sjisstr, 12 * dstlen + 1, (char*)ustr, dstlen + 1);
+	  sjisstr = (char*)UdmMalloc(12 * dstlen + 1);
+	  UdmConv(&uni_sjis, sjisstr, 12 * dstlen + 1, (char*)ustr, sizeof(*ustr) * (dstlen + 1));
 
 	  UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
 #ifdef HAVE_PTHREADS
@@ -1102,45 +1328,39 @@ int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang) {
 	  UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
 
 	  reslen = strlen(sjisstr_seg) + 1;
-	  ustr = (int*)realloc(ustr, reslen * sizeof(int));
+	  ustr = (int*)UdmRealloc(ustr, reslen * sizeof(int));
 	  UdmConv(&sjis_uni, (char*)ustr, reslen * sizeof(int), sjisstr_seg, reslen);
 	  UDM_FREE(sjisstr);
 	  dstlen = UdmUniLen(ustr);
 	}
 #endif
-#ifdef HAVE_CHARSET_gb2312
 #if defined(CHASEN) || defined(MECAB)
 	else
 #endif
 	  if (lang == NULL || !strncasecmp(lang, "zh", 2)) {
-	    gbstr = (char*)malloc(12 * dstlen + 1);
-	    UdmConv(&uni_gb, gbstr, 12 * dstlen + 1, (char*)ustr, dstlen + 1);
+	    int *seg_ustr;
 	    UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
-	    if ((gbstr_seg = UdmSegmentByFreq(&Indexer->Conf->Chi, gbstr)) != NULL) {
-	      ustr = (int*)realloc(ustr, 12 * dstlen + 1);
-	      UdmConv(&gb_uni, (char*)ustr, 12 * dstlen + 1, gbstr_seg, 12 * dstlen + 1);
-	      UDM_FREE(gbstr_seg);
-	    }
+	    seg_ustr = UdmSegmentByFreq(&Indexer->Conf->Chi, ustr);
 	    UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
-	    UDM_FREE(gbstr);
+	    if (seg_ustr != NULL) {
+	      UDM_FREE(ustr);
+	      ustr = seg_ustr;
+	    }
 	    dstlen = UdmUniLen(ustr);
 	  }
-#endif
 	
-#if defined(CHASEN) || defined(MECAB) || defined(HAVE_CHARSET_gb2312)
+#if defined(CHASEN) || defined(MECAB)
 	  else
 #endif
 	    if (lang == NULL || !strncasecmp(lang, "th", 2)) {
-	      tistr = (char*)malloc(12 * dstlen + 1);
-	      UdmConv(&uni_tis, tistr, 12 * dstlen + 1, (char*)ustr, dstlen + 1);
+	      int *seg_ustr;
 	      UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
-	      if ((tistr_seg = UdmSegmentByFreq(&Indexer->Conf->Thai, tistr)) != NULL) {
-		ustr = (int*)realloc(ustr, 12 * dstlen + 1);
-		UdmConv(&tis_uni, (char*)ustr, 12 * dstlen + 1, tistr_seg, 12 * dstlen + 1);
-		UDM_FREE(tistr_seg);
-	      }
+	      seg_ustr = UdmSegmentByFreq(&Indexer->Conf->Thai, ustr);
 	      UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
-	      UDM_FREE(tistr);
+	      if (seg_ustr != NULL) {
+		UDM_FREE(ustr);
+		ustr = seg_ustr;
+	      }
 	      dstlen = UdmUniLen(ustr);
 	    }
 
