@@ -54,7 +54,7 @@ __C_LINK int __UDMCALL UdmServerAdd(UDM_AGENT *A, UDM_SERVER *srv){
 	/* Copy URL to temp string    */
 	/* to keep srv->url unchanged */
 	len = strlen(UDM_NULL2EMPTY(srv->Match.pattern)) + 4;
-	if ((urlstr = (char*)malloc(len)) == NULL) return UDM_ERROR;
+	if ((urlstr = (char*)UdmMalloc(len)) == NULL) return UDM_ERROR;
 	strcpy(urlstr, UDM_NULL2EMPTY(srv->Match.pattern));
 	
 	from.freeme = 0;
@@ -152,7 +152,7 @@ __C_LINK int __UDMCALL UdmServerAdd(UDM_AGENT *A, UDM_SERVER *srv){
 	
 	UdmVarListReplaceLst(&new->Vars,&srv->Vars,NULL,"*");
 	
-	new->Match.pattern = (char*)strdup(urlstr);
+	new->Match.pattern = (char*)UdmStrdup(urlstr);
 	new->Match.nomatch=srv->Match.nomatch;
 	new->Match.case_sense=srv->Match.case_sense;
 	new->Match.match_type=srv->Match.match_type;
@@ -207,7 +207,7 @@ UDM_SERVER * UdmServerFind(UDM_ENV *Conf, UDM_SERVERLIST *List, const char *url,
 	/* If it's a robot.txt, cut to hostinfo and find result */
 	if((robots=strstr(url,"/robots.txt"))){
 		if(!strcmp(robots,"/robots.txt")){
-			robots = (char*)strdup(url);
+			robots = (char*)UdmStrdup(url);
 			robots[strlen(url)-10]='\0';
 		}else{
 			robots=NULL;
@@ -244,7 +244,7 @@ UDM_SERVER * UdmServerFind(UDM_ENV *Conf, UDM_SERVERLIST *List, const char *url,
 			Res=srv;
 			if((aliastr != NULL) && (alias != NULL)) {
 			        aliastrlen = 128 + strlen(url) + strlen(alias) + strlen(srv->Match.pattern);
-				*aliastr = (char*)malloc(aliastrlen);
+				*aliastr = (char*)UdmMalloc(aliastrlen);
 				if (*aliastr != NULL)
 				  UdmMatchApply(*aliastr, aliastrlen, url, alias, &srv->Match, 10, P);
 			}
@@ -304,7 +304,7 @@ urlid_t UdmServerGetSiteId(UDM_AGENT *Indexer, UDM_SERVER *srv, UDM_URL *url) {
         (follow == UDM_FOLLOW_SITE)) {
     return srv->site_id;
   }
-  if((urlstr = (char*)malloc(strlen(UDM_NULL2EMPTY(url->schema)) + strlen(UDM_NULL2EMPTY(url->hostname)) + 10)) == NULL) {
+  if((urlstr = (char*)UdmMalloc(strlen(UDM_NULL2EMPTY(url->schema)) + strlen(UDM_NULL2EMPTY(url->hostname)) + 10)) == NULL) {
     return 0;
   }
   sprintf(urlstr, "%s://%s/", UDM_NULL2EMPTY(url->schema), UDM_NULL2EMPTY(url->hostname));
