@@ -99,9 +99,9 @@ static int UdmUniRegComp(UDM_UNIREG_EXP *reg, const int *pattern){
 	tok=UdmUniRegTok(pattern,&lt);
 	while(tok){
 		size_t len;
-		reg->Token=(UDM_UNIREG_TOK*)realloc(reg->Token,sizeof(*reg->Token)*(reg->ntokens+1));
+		reg->Token=(UDM_UNIREG_TOK*)UdmRealloc(reg->Token,sizeof(*reg->Token)*(reg->ntokens+1));
 		len=lt-tok;
-		reg->Token[reg->ntokens].str=(int*)malloc((len+1)*sizeof(int));
+		reg->Token[reg->ntokens].str=(int*)UdmMalloc((len+1)*sizeof(int));
 		memcpy(reg->Token[reg->ntokens].str,tok,len*sizeof(int));
 		reg->Token[reg->ntokens].str[len]=0;
 		
@@ -230,9 +230,17 @@ void UdmUniRegCompileAll(UDM_ENV *Conf) {
 
 static int cmpspellword(int *w1, const int *w2) {
     int u1[BUFSIZ], u2[BUFSIZ];
+    int res;
+#ifdef USE_PARANOIA
+    void *paran = UdmViolationEnter();
+#endif
     UdmUniStrCpy(u1, w1); *u1 &= 255;
     UdmUniStrCpy(u2, w2); *u2 &= 255;
-      return(UdmUniStrCmp(u1, u2));
+    res = UdmUniStrCmp(u1, u2);
+#ifdef USE_PARANOIA
+    UdmViolationExit(paran);
+#endif
+    return res; 
 }
 
 static int cmpspell(const void *s1,const void *s2){
@@ -246,24 +254,60 @@ static int cmpspell(const void *s1,const void *s2){
 
 static int cmpaffix(const void *s1,const void *s2){
   int lc;
-  if (((const UDM_AFFIX*)s1)->type < ((const UDM_AFFIX*)s2)->type) return -1;
-  if (((const UDM_AFFIX*)s1)->type > ((const UDM_AFFIX*)s2)->type) return 1;
+  int res;
+  int u1[BUFSIZ], u2[BUFSIZ];
+#ifdef USE_PARANOIA
+  void *paran = UdmViolationEnter();
+#endif
+  if (((const UDM_AFFIX*)s1)->type < ((const UDM_AFFIX*)s2)->type) {
+#ifdef USE_PARANOIA
+    UdmViolationExit(paran);
+#endif
+    return -1;
+  }
+  if (((const UDM_AFFIX*)s1)->type > ((const UDM_AFFIX*)s2)->type) {
+#ifdef USE_PARANOIA
+    UdmViolationExit(paran);
+#endif
+    return 1;
+  }
   lc = strcmp(((const UDM_AFFIX*)s1)->lang,((const UDM_AFFIX*)s2)->lang);
   if (lc == 0) {
-    if ( (((const UDM_AFFIX*)s1)->replen == 0) && (((const UDM_AFFIX*)s2)->replen == 0) ) return 0;
-    if (((const UDM_AFFIX*)s1)->replen == 0) return -1;
-    if (((const UDM_AFFIX*)s2)->replen == 0) return 1;
-    {
-      int u1[BUFSIZ], u2[BUFSIZ];
-      UdmUniStrCpy(u1,((const UDM_AFFIX*)s1)->repl); 
-      UdmUniStrCpy(u2,((const UDM_AFFIX*)s2)->repl); 
-      if (((const UDM_AFFIX*)s1)->type == 'p') {
-	*u1 &= 255; *u2 &= 255;
-	return(UdmUniStrCmp(u1, u2));
-      } else {
-	u1[((const UDM_AFFIX*)s1)->replen - 1] &= 255; u2[((const UDM_AFFIX*)s2)->replen -1] &= 255;
-	return(UdmUniStrBCmp(u1, u2));
-      }
+    if ( (((const UDM_AFFIX*)s1)->replen == 0) && (((const UDM_AFFIX*)s2)->replen == 0) ) {
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
+      return 0;
+    }
+    if (((const UDM_AFFIX*)s1)->replen == 0) {
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
+      return -1;
+    }
+    if (((const UDM_AFFIX*)s2)->replen == 0) {
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
+      return 1;
+    }
+    
+    UdmUniStrCpy(u1,((const UDM_AFFIX*)s1)->repl); 
+    UdmUniStrCpy(u2,((const UDM_AFFIX*)s2)->repl); 
+    if (((const UDM_AFFIX*)s1)->type == 'p') {
+      *u1 &= 255; *u2 &= 255;
+      res = UdmUniStrCmp(u1, u2);
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
+      return res;
+    } else {
+      u1[((const UDM_AFFIX*)s1)->replen - 1] &= 255; u2[((const UDM_AFFIX*)s2)->replen -1] &= 255;
+      res = UdmUniStrBCmp(u1, u2);
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
+      return res;
     }
   }
   return lc;
@@ -294,12 +338,12 @@ __C_LINK int __UDMCALL UdmImportDictionary(UDM_ENV * Conf, const char *lang, con
 	UDM_CONV touni;
 	UDM_CONV fromuni;
 
-	if ((str = (char*) malloc(1024)) == NULL) return 1; 
-	if ((lstr = (char*) malloc(2048)) == NULL) {
+	if ((str = (char*) UdmMalloc(1024)) == NULL) return 1; 
+	if ((lstr = (char*) UdmMalloc(2048)) == NULL) {
 	  UDM_FREE(str);
 	  return 1; 
 	}
-	if ((ustr = (int*) malloc(8192)) == NULL) {
+	if ((ustr = (int*) UdmMalloc(8192)) == NULL) {
 	  UDM_FREE(str);
 	  UDM_FREE(lstr);
 	  return 1; 
@@ -495,14 +539,31 @@ FILE *affix;
 #ifdef DEBUG_UNIREG
  UDM_CONV fromuni;
 #endif
+#ifdef USE_PARANOIA
+ void *paran = UdmViolationEnter();
+#endif
 
-	    if(!(affix=fopen(filename,"r")))
-		    return(1);
+            if(!(affix=fopen(filename,"r"))) {
+#ifdef USE_PARANOIA
+	      UdmViolationExit(paran);
+#endif
+	      return 1;
+	    }
 
 	    affix_charset = UdmGetCharSet(charset);
-	    if (affix_charset == NULL) return 1;
+	    if (affix_charset == NULL) {
+#ifdef USE_PARANOIA
+	      UdmViolationExit(paran);
+#endif
+	      return 1;
+	    }
 	    sys_int = UdmGetCharSet("sys-int");
-	    if (sys_int == NULL) return 1;
+	    if (sys_int == NULL) {
+#ifdef USE_PARANOIA
+	      UdmViolationExit(paran);
+#endif
+	      return 1;
+	    }
 	    
 	    UdmConvInit(&touni,affix_charset,sys_int,0);
 #ifdef DEBUG_UNIREG
@@ -584,7 +645,10 @@ FILE *affix;
 	    }
 	    fclose(affix);
 	    
-	    return(0);
+#ifdef USE_PARANOIA
+	    UdmViolationExit(paran);
+#endif
+	    return 0;
 }
 
 __C_LINK void __UDMCALL UdmSortDictionary(UDM_SPELLLIST * List){
@@ -678,12 +742,21 @@ static void CheckSuffix(const int *word, size_t len, UDM_AFFIX *Affix, int *res,
   int newword[2*MAXNORMLEN] = {0};
   int err;
 /*3.1  int curlang, curspellang;*/
+#ifdef USE_PARANOIA
+  void *paran = UdmViolationEnter();
+#endif
   
   *res = UdmUniStrBNCmp(word, Affix->repl, Affix->replen);
   if (*res < 0) {
+#ifdef USE_PARANOIA
+    UdmViolationExit(paran);
+#endif
     return;
   }
   if (*res > 0) {
+#ifdef USE_PARANOIA
+    UdmViolationExit(paran);
+#endif
     return;
   }
   UdmUniStrCpy(newword, word);
@@ -693,6 +766,9 @@ static void CheckSuffix(const int *word, size_t len, UDM_AFFIX *Affix, int *res,
     err = UdmUniRegComp(&(Affix->reg), Affix->mask);
     if(err){
       UdmUniRegFree(&(Affix->reg));
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
       return;
     }
     Affix->compile = 0;
@@ -710,9 +786,15 @@ static void CheckSuffix(const int *word, size_t len, UDM_AFFIX *Affix, int *res,
       Indexer->curlang = curlang;
       Indexer->spellang = curspellang;
 */
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
       return;
     }
   }
+#ifdef USE_PARANOIA
+  UdmViolationExit(paran);
+#endif
   return;
 }
 
@@ -722,9 +804,15 @@ static int CheckPrefix(const int *word, UDM_AFFIX *Affix, UDM_AGENT *Indexer, in
   int err, ls, rs, lres,rres, res;
   size_t newlen;
   UDM_AFFIX *CAffix = Indexer->Conf->Affixes.Affix;
+#ifdef USE_PARANOIA
+  void *paran = UdmViolationEnter();
+#endif
   
   res = UdmUniStrNCaseCmp(word, Affix->repl, Affix->replen);
   if (res != 0) {
+#ifdef USE_PARANOIA
+    UdmViolationExit(paran);
+#endif
     return res;
   }
   UdmUniStrCpy(newword, Affix->find);
@@ -734,6 +822,9 @@ static int CheckPrefix(const int *word, UDM_AFFIX *Affix, UDM_AGENT *Indexer, in
     err = UdmUniRegComp(&(Affix->reg),Affix->mask);
     if(err){
       UdmUniRegFree(&(Affix->reg));
+#ifdef USE_PARANOIA
+      UdmViolationExit(paran);
+#endif
       return (0);
     }
     Affix->compile = 0;
@@ -755,7 +846,9 @@ static int CheckPrefix(const int *word, UDM_AFFIX *Affix, UDM_AGENT *Indexer, in
       rs--;
     }
   }
-
+#ifdef USE_PARANOIA
+  UdmViolationExit(paran);
+#endif
   return 0;
 }
 
@@ -918,8 +1011,8 @@ static void UdmAllFormsWord (UDM_AGENT *Indexer, UDM_SPELL *word, UDM_WIDEWORDLI
 	     )  {
 	  
 	  w.len = UdmUniLen(word->word) - Affix[i].findlen + Affix[i].replen;
-	  if ( ( (w.word = realloc(w.word, 8 * w.len + 1)) == NULL) ||
-	       ( (w.uword = realloc(w.uword, (w.len + 1) * sizeof(int))) == NULL)) 
+	  if ( ( (w.word = UdmRealloc(w.word, 8 * w.len + 1)) == NULL) ||
+	       ( (w.uword = UdmRealloc(w.uword, (w.len + 1) * sizeof(int))) == NULL)) 
 	    return;
 
 	  bzero((void*)w.uword, (w.len + 1) * sizeof(int));
@@ -978,8 +1071,8 @@ __C_LINK UDM_WIDEWORDLIST * __UDMCALL UdmAllForms (UDM_AGENT *Indexer, UDM_WIDEW
     while (*cur != NULL) {
 
       w.len = UdmUniLen((*cur)->word);
-      if ( ( (w.word = realloc(w.word, 8 * w.len + 1)) == NULL) ||
-	   ( (w.uword = realloc(w.uword, (w.len + 1) * sizeof(int))) == NULL)) 
+      if ( ( (w.word = UdmRealloc(w.word, 8 * w.len + 1)) == NULL) ||
+	   ( (w.uword = UdmRealloc(w.uword, (w.len + 1) * sizeof(int))) == NULL)) 
 	return NULL;
       UdmUniStrCpy(w.uword,(*cur)->word); 
       UdmConv(&fromuni,w.word,8*w.len+1,(char*)w.uword,sizeof(w.uword[0])*(w.len+1));
