@@ -68,14 +68,14 @@ int UdmStopListAdd(UDM_STOPLIST *List,UDM_STOPWORD * stopword){
 	for(j=0;j<List->nstopwords;j++){
 		if(!strcmp(List->StopWord[j].word,stopword->word)){
 			UDM_FREE(List->StopWord[j].lang);
-			List->StopWord[j].lang = (char*)strdup("");
+			List->StopWord[j].lang = (char*)UdmStrdup("");
 			return 0;
 		}
 	}
 
-	List->StopWord=(UDM_STOPWORD *)realloc(List->StopWord,(List->nstopwords+1)*sizeof(UDM_STOPWORD));
-	List->StopWord[List->nstopwords].word = (char*)strdup(stopword->word);
-	List->StopWord[List->nstopwords].lang = (char*)strdup(stopword->lang?stopword->lang:"");
+	List->StopWord=(UDM_STOPWORD *)UdmRealloc(List->StopWord,(List->nstopwords+1)*sizeof(UDM_STOPWORD));
+	List->StopWord[List->nstopwords].word = (char*)UdmStrdup(stopword->word);
+	List->StopWord[List->nstopwords].lang = (char*)UdmStrdup(stopword->lang?stopword->lang:"");
 	List->nstopwords++;
 
 	return(1);
@@ -106,7 +106,7 @@ __C_LINK int __UDMCALL UdmStopListLoad(UDM_ENV * Conf,const char *fname){
 		sprintf(Conf->errstr,"Can't open stopwords file '%s' (%s)", fname, strerror(errno));
 		return UDM_ERROR;
 	}
-	if ((lwrd = (char*)malloc(Conf->WordParam.max_word_len + 1)) == NULL) return UDM_ERROR;
+	if ((lwrd = (char*)UdmMalloc(Conf->WordParam.max_word_len + 1)) == NULL) return UDM_ERROR;
 
 	bzero((void*)&stopword, sizeof(stopword));
 
@@ -118,13 +118,13 @@ __C_LINK int __UDMCALL UdmStopListLoad(UDM_ENV * Conf,const char *fname){
 			UDM_FREE(charset);
 			charset = udm_strtok_r(str + 8, " \t\n\r", &lasttok);
 			if(charset){
-				charset = (char*)strdup(charset);
+				charset = (char*)UdmStrdup(charset);
 			}
 		}else
 		if(!strncmp(str,"Language:",9)){
 			UDM_FREE(stopword.lang);
 			stopword.lang = udm_strtok_r(str + 9, " \t\n\r", &lasttok);
-			if(stopword.lang)stopword.lang = (char*)strdup(stopword.lang);
+			if(stopword.lang)stopword.lang = (char*)UdmStrdup(stopword.lang);
 		}else
 		if((stopword.word = udm_strtok_r(str, "\t\n\r", &lasttok))) {
 			
