@@ -31,6 +31,7 @@
 #include "udm_host.h"
 #include "udm_contentencoding.h"
 #include "udm_utils.h"
+#include "udm_wild.h"
 
 /******************* Template functions ********************/
 
@@ -65,7 +66,7 @@ static char * HiLightDup(const char * src,const char * beg, const char * end){
 				len++;
 		}
 	}
-	res=(char*)malloc(len);
+	res = (char*)UdmMalloc(len);
 	for(s=src,d=res;*s;s++){
 		switch(*s){
 			case '\2':
@@ -181,7 +182,7 @@ static size_t PrintTextTemplate(UDM_AGENT *A, FILE * stream, char * dst, size_t
 			    }
 			  }
 			}
-			if ((newvalue = (char*)malloc((size_t)(p - value) + 8))) {
+			if ((newvalue = (char*)UdmMalloc((size_t)(p - value) + 8))) {
 			  char *c2, *c3;
 			  strncpy(newvalue, value, (size_t)(p - value));
 			  newvalue[(p - value)] = '\0';
@@ -208,7 +209,7 @@ static size_t PrintTextTemplate(UDM_AGENT *A, FILE * stream, char * dst, size_t
 				UDM_FREE(eval);
 				break;
 			case '%':
-				eval=(char*)malloc(strlen(value)*3);
+				eval =( char*)UdmMalloc(strlen(value) * 3);
 				UdmEscapeURL(eval,value);
 				dlen+=out_string(stream, dst + dlen, dst_len - dlen, eval);
 				UDM_FREE(eval);break;
@@ -238,11 +239,11 @@ static char * GetHtmlTok(const char * src,const char ** lt){
 	}
 	if(!(*lt)){
 		/* Last token */
-		res = (char*)strdup(src);
+		res = (char*)UdmStrdup(src);
 	}else{
 		/* Token in the middle */
 		len=(*lt)-src;
-		res=(char*)malloc(len+2);
+		res = (char*)UdmMalloc(len + 2);
 		strncpy(res,src,len);
 		res[len]='\0';
 	}
@@ -263,27 +264,27 @@ static void TemplateInclude(UDM_AGENT *Agent,FILE *stream,UDM_VARLIST *vars,cons
 	
 	size_t max_doc_size = (size_t)UdmVarListFindInt(vars,"MaxDocSize",UDM_MAXDOCSIZE);
 
-	if(!Inc->Buf.buf)Inc->Buf.buf=malloc(max_doc_size);
+	if(!Inc->Buf.buf) Inc->Buf.buf = (char*)UdmMalloc(max_doc_size);
 	Inc->Buf.maxsize=max_doc_size;
 
 	UdmHTMLTOKInit(tag);
 	UdmHTMLToken(tok, &last, tag);
 	for(i = 0; i < ltag.ntoks; i++) {
 		if (ISTAG(i, "content")) {
-			tag_content = strndup(ltag.toks[i].val, ltag.toks[i].vlen);
+			tag_content = UdmStrndup(ltag.toks[i].val, ltag.toks[i].vlen);
 			break;
 		}
 	}
 	if(tag_content){
 		const char *ce;
 		size_t vurlen = 256 + 4 * strlen(tag_content);
-		char  *vurl = (char*)malloc(vurlen);
+		char  *vurl = (char*)UdmMalloc(vurlen);
 					
 		PrintTextTemplate(Agent, NULL, vurl, vurlen, vars, tag_content);
 		UdmURLParse(&Inc->CurURL,vurl);
 		UDM_FREE(vurl);
 		UdmVarListReplaceStr(&Inc->RequestHeaders, "Host", UDM_NULL2EMPTY(Inc->CurURL.hostname));
-		Inc->connp.hostname = (char*)strdup(UDM_NULL2EMPTY(Inc->CurURL.hostname));
+		Inc->connp.hostname = (char*)UdmStrdup(UDM_NULL2EMPTY(Inc->CurURL.hostname));
 		Inc->connp.port = Inc->CurURL.port ? Inc->CurURL.port : Inc->CurURL.default_port;
 		
 		if(UdmHostLookup(&Agent->Conf->Hosts, &Inc->connp)){
@@ -324,21 +325,21 @@ static size_t TemplateTag(UDM_AGENT *Agent,FILE *stream,char *dst,size_t dst_len
 	char * vname = NULL, *value = NULL;
 	size_t i, res = 0;
 	
-	opt=(char*)malloc(strlen(tok)+200);
+	opt = (char*)UdmMalloc(strlen(tok) + 200);
 	UdmHTMLTOKInit(tag);
 	UdmHTMLToken(tok, &last, tag);
 	sprintf(opt, "<");
 
 	for (i = 0; i < ltag.ntoks; i++) {
 		if (ISTAG(i, "selected")) {
-			vname = strndup(ltag.toks[i].val, ltag.toks[i].vlen);
+			vname = UdmStrndup(ltag.toks[i].val, ltag.toks[i].vlen);
 		} else if (ISTAG(i, "value")) {
-			value = strndup(ltag.toks[i].val, ltag.toks[i].vlen);
+			value = UdmStrndup(ltag.toks[i].val, ltag.toks[i].vlen);
 			sprintf(UDM_STREND(opt), "VALUE=\"%s\" ", value);
 		} else {
-			char *tname = strndup(ltag.toks[i].name, ltag.toks[i].nlen);
+			char *tname = UdmStrndup(ltag.toks[i].name, ltag.toks[i].nlen);
 			if (ltag.toks[i].vlen) {
-				char *tval = strndup(ltag.toks[i].val, ltag.toks[i].vlen);
+				char *tval = UdmStrndup(ltag.toks[i].val, ltag.toks[i].vlen);
 				sprintf(UDM_STREND(opt), "%s=\"%s\" ", tname, tval);
 				UDM_FREE(tval);
 			} else {
@@ -403,8 +404,8 @@ static void HTMLTokToVarList(UDM_VARLIST *vars,UDM_HTMLTOK *tag){
 	size_t	toks;
 	
 	for(toks=0;toks<tag->ntoks;toks++){
-		char *vr=tag->toks[toks].name ? strndup(tag->toks[toks].name,tag->toks[toks].nlen) : (char*)strdup("");
-		char *vl=tag->toks[toks].val ? strndup(tag->toks[toks].val,tag->toks[toks].vlen) : (char*)strdup("");
+		char *vr=tag->toks[toks].name ? UdmStrndup(tag->toks[toks].name,tag->toks[toks].nlen) : (char*)UdmStrdup("");
+		char *vl=tag->toks[toks].val ? UdmStrndup(tag->toks[toks].val,tag->toks[toks].vlen) : (char*)UdmStrdup("");
 		UdmVarListAddStr(vars,vr,vl);
 		UDM_FREE(vr);
 		UDM_FREE(vl);
@@ -466,6 +467,9 @@ static void TemplateCondition(UDM_AGENT *Agent,UDM_VARLIST *vars,const char *tok
 	if(!(strncasecmp(tok,"<!IFNOT",7))){
 		it->condition=strcasecmp(var,val);
 	}else
+	if( !(strncasecmp(tok, "<!IFLIKE", 8)) || !(strncasecmp(tok, "<!ELIKE", 7)) || !(strncasecmp(tok, "<!ELSELIKE", 10)) ) {
+		it->condition = !UdmWildCaseCmp(var, val);
+	}else
 	if( !(strncasecmp(tok, "<!IF", 4)) || !(strncasecmp(tok, "<!ELIF", 6)) || !(strncasecmp(tok, "<!ELSEIF", 8)) ) {
 		it->condition=!strcasecmp(var,val);
 	}
@@ -517,18 +521,24 @@ static void PrintHtmlTemplate(UDM_AGENT * Agent, FILE * stream, char * dst, size
 		if(!(strncasecmp(tok,"<!SET",5))){
 			TemplateSet(Agent,vars,tok,&is);
 		}else
-		if(!(strncasecmp(tok,"<!COPY",6))){
-			TemplateCopy(Agent,vars,tok,&is);
-		}else
 		if(!(strncasecmp(tok,"<!IF",4))){
 			TemplateIf(Agent,vars,tok,&is);
 		}else
-		if(!(strncasecmp(tok,"<!ELSEIF",8))){
-			TemplateElseIf(Agent,vars,tok,&is);
+		if(!(strncasecmp(tok,"<!IFLIKE", 8))) {
+			TemplateIf(Agent,vars,tok,&is);
+		}else
+		if(!(strncasecmp(tok, "<!ELSEIF", 8))) {
+			TemplateElseIf(Agent, vars, tok, &is);
 		}else
 		if(!(strncasecmp(tok,"<!ELIF",6))){
 			TemplateElseIf(Agent,vars,tok,&is);
 		}else
+		if(!(strncasecmp(tok, "<!ELIKE", 7))) {
+			TemplateElseIf(Agent, vars, tok, &is);
+		}else
+		if(!(strncasecmp(tok, "<!ELSELIKE", 10))) {
+			TemplateElseIf(Agent, vars, tok, &is);
+		}else
 		if(!(strncasecmp(tok,"<!ELSE",6))){
 			TemplateElse(Agent,vars,tok,&is);
 		}else
@@ -542,6 +552,9 @@ static void PrintHtmlTemplate(UDM_AGENT * Agent, FILE * stream, char * dst, size
 			if(!(strncasecmp(tok,"<OPTION",7))){
 				dlen += TemplateTag(Agent,stream,dst+dlen,dst_len-dlen,vars,tok);
 			}else
+			if(!(strncasecmp(tok,"<!COPY",6))){
+				TemplateCopy(Agent,vars,tok,&is);
+			}else
 			if(!(strncasecmp(tok,"<INPUT",6))){
 				dlen += TemplateTag(Agent,stream,dst+dlen,dst_len-dlen,vars,tok);
 			}else
@@ -633,12 +646,21 @@ static int ParseVariable(UDM_ENV *Env,UDM_VARLIST *vars,char *str){
 			UdmVarListReplaceInt(&Env->Vars, "IspellUsePrefixes", val);
 		}else
 		if(!strcasecmp(str,"StopwordFile")){
-			char aname[1024];
-			
+			char aname[1024];			
+#ifdef WIN32	/* Space is valid symbol for WIN32 versions */
+			arg = udm_strtok_r(NULL, "\t\r\n", &lt);
+#else
 			arg = udm_strtok_r(NULL, " \t\r\n", &lt);
+#endif
 			if(arg){
-				if(arg[0]=='/')strncpy(aname,arg,sizeof(aname)-1);
-				else	udm_snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,arg);
+				if(arg[0]=='/')
+					strncpy(aname,arg,sizeof(aname)-1);
+				else
+#ifdef WIN32
+					udm_snprintf(aname,sizeof(aname)-1,"%s", arg);
+#else
+					udm_snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,arg);
+#endif
 				if(UdmStopListLoad(Env,aname)){
 					return UDM_ERROR;
 				}
@@ -649,10 +671,20 @@ static int ParseVariable(UDM_ENV *Env,UDM_VARLIST *vars,char *str){
 		}else
 		if(!strcasecmp(str,"Synonym")){
 			char aname[1024];
+#ifdef WIN32	/* Space is valid symbol for WIN32 versions */
+			arg = udm_strtok_r(NULL, "\t\r\n", &lt);
+#else
 			arg = udm_strtok_r(NULL, " \t\r\n", &lt);
+#endif
 			if(arg){
-				if(arg[0]=='/')strncpy(aname,arg,sizeof(aname)-1);
-				else	udm_snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,arg);
+				if(arg[0]=='/')
+					strncpy(aname,arg,sizeof(aname)-1);
+				else
+#ifdef WIN32
+					udm_snprintf(aname,sizeof(aname)-1,"%s",arg);
+#else
+					udm_snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,arg);
+#endif
 				if(UdmSynonymListLoad(Env,aname)){
 					return UDM_ERROR;
 				}
@@ -661,10 +693,19 @@ static int ParseVariable(UDM_ENV *Env,UDM_VARLIST *vars,char *str){
 				return UDM_ERROR;
 			}
 		}else
+		if(!strcasecmp(str,"ImportEnv")){
+			const char *val;
+			arg = udm_strtok_r(NULL, " \t\r\n", &lt);
+			if ((val=getenv(arg)))
+				UdmVarListReplaceStr(vars,arg,val);
+		}else
 		if(!strcasecmp(str,"DBAddr")){
-			if((arg = udm_strtok_r(NULL, " \t\r\n", &lt))) {
-			  UdmDBListAdd(&Env->dbl, arg, UDM_OPEN_MODE_READ);
+			char *p;
+			if((arg = udm_strtok_r(NULL, " \t\r\n", &lt)) &&
+			    (p = UdmParseEnvVar(Env, arg))) {
+			  UdmDBListAdd(&Env->dbl, p, UDM_OPEN_MODE_READ);
 /*			  UdmVarListReplaceStr(vars,tok,arg);*/
+			  UdmFree(p);
 			}
 		}else
 		if((str[0]=='R'||str[0]=='r')&&(str[1]>='0')&&(str[1]<='9')){
@@ -692,7 +733,7 @@ static int ParseVariable(UDM_ENV *Env,UDM_VARLIST *vars,char *str){
 			arg = udm_strtok_r(NULL, " \t\r\n", &lt);
 			if((sc=strchr(arg,':'))){
 				*sc='\0';sc++;
-				nm=(char*)malloc(strlen(arg)+8);
+				nm=(char*)UdmMalloc(strlen(arg)+8);
 				sprintf(nm,"Limit-%s",arg);
 				UdmVarListReplaceStr(vars,nm,sc);
 				UDM_FREE(nm);
@@ -712,14 +753,14 @@ static int ParseVariable(UDM_ENV *Env,UDM_VARLIST *vars,char *str){
 				Alias.arg=arg1;
 				Alias.match_type=UDM_MATCH_BEGIN;
 				Alias.case_sense=0;
-				UdmMatchListAdd(&Env->Aliases,&Alias,Env->errstr,sizeof(Env->errstr));
+				UdmMatchListAdd(NULL, &Env->Aliases, &Alias, Env->errstr, sizeof(Env->errstr), 0);
 			}
 		}else
-		if(!strcasecmp(str,"MaxWordLen")){
+		if(!strcasecmp(str,"MaxWordLength")){
 			arg = udm_strtok_r(NULL, " \t\r\n", &lt);
 			if(arg)Env->WordParam.max_word_len=atoi(arg);
 		}else
-		if(!strcasecmp(str,"MinWordLen")){
+		if(!strcasecmp(str,"MinWordLength")){
 			arg = udm_strtok_r(NULL, " \t\r\n", &lt);
 			if(arg)Env->WordParam.min_word_len=atoi(arg);
 		}else
@@ -827,11 +868,11 @@ int UdmTemplateLoad(UDM_ENV * Env,UDM_VARLIST * vars,const char * tname, UDM_VAR
 				if(s[0]=='/'){
 					if(!strcasecmp(s+1,cursection) && cursection[0]){
 						UDM_VAR *I;
-						tmpl->Var=(UDM_VAR*)realloc(tmpl->Var,(tmpl->nvars+1)*sizeof(UDM_VAR));
+						tmpl->Var=(UDM_VAR*)UdmRealloc(tmpl->Var,(tmpl->nvars+1)*sizeof(UDM_VAR));
 						I=&tmpl->Var[tmpl->nvars];
-						I->name = (char*)strdup(cursection);
-						I->val = (char*)strdup(cur?cur:"");
-						I->txt_val = (char*)strdup(cur?cur:"");
+						I->name = (char*)UdmStrdup(cursection);
+						I->val = (char*)UdmStrdup(cur?cur:"");
+						I->txt_val = (char*)UdmStrdup(cur?cur:"");
 						tmpl->nvars++;
 						cursection[0]='\0';
 						UDM_FREE(cur);
@@ -850,9 +891,9 @@ int UdmTemplateLoad(UDM_ENV * Env,UDM_VARLIST * vars,const char * tname, UDM_VAR
 			continue;
 		
 		if(!cur){
-			cur = (char*)strdup(ostr);
+			cur = (char*)UdmStrdup(ostr);
 		}else{
-			cur=(char*)realloc(cur,strlen(cur)+strlen(ostr)+1);
+			cur=(char*)UdmRealloc(cur,strlen(cur)+strlen(ostr)+1);
 			strcat(cur,ostr);
 		}
 	}
