@@ -15,10 +15,6 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */
 
-#ifdef WIN32
-#include "stdafx.h"
-#endif
-
 #include "udm_config.h"
 
 #include <stdio.h>
@@ -56,12 +52,106 @@
 #endif
 #include <signal.h>
 
-#include "udm_common.h"
-#include "udm_utils.h"
-#include "udm_xmalloc.h"
 #include "udm_signals.h"
+#include "udm_utils.h"
 
-int tz_offset=0;	/* EXT */
+#ifdef WIN32
+#include <assert.h>
+#endif
+
+long tz_offset=0;	/* EXT */
+
+
+char *UdmStrStore (char *dest, const char *src) {
+	size_t d_size = (dest ? strlen(dest) : 0);
+	size_t s_size = strlen(src) + 1;
+	char *d = UdmRealloc(dest, d_size + s_size);
+
+	if (d) memcpy(d + d_size, src, s_size);
+	else UDM_FREE(dest);
+	return(d);
+}
+
+UDM_DSTR *UdmDSTRInit (UDM_DSTR *dstr, size_t size_page) {
+	if (! size_page) return(NULL);
+
+	if (! dstr) {
+		dstr = UdmMalloc(sizeof(UDM_DSTR));
+		if (! dstr) return(NULL);
+		dstr->free = 1;
+	} else {
+		dstr->free = 0;
+	}
+
+	dstr->data = UdmMalloc(size_page);
+	if (! dstr->data) {
+		if (dstr->free) UdmFree(dstr);
+		return(NULL);
+	}
+	dstr->size_page = size_page;
+	dstr->size_data = 0;
+	dstr->size_total = size_page;
+	return(dstr);
+}
+
+void UdmDSTRFree (UDM_DSTR *dstr) {
+	UdmFree(dstr->data);
+	if (dstr->free) UdmFree(dstr);
+}
+
+size_t UdmDSTRAppend (UDM_DSTR *dstr, const char *data, size_t size_data) {
+	size_t bytes_left = dstr->size_total - dstr->size_data;
+	size_t asize;
+	char *tmp;
+
+	if (! data || ! size_data) return(0);
+
+	if (bytes_left <= size_data) {
+		asize = dstr->size_total + ((size_data - bytes_left) / dstr->size_page + 1) * dstr->size_page;
+		tmp = UdmRealloc(dstr->data, asize);
+		if (! tmp) return(0);
+		dstr->data = tmp;
+		dstr->size_total = asize;
+	}
+	memcpy(dstr->data + dstr->size_data, data, size_data);
+	dstr->size_data += size_data;
+	dstr->data[dstr->size_data] = 0;
+	return(size_data);
+}
+
+size_t UdmDSTRAppendSTR (UDM_DSTR *dstr, const char *data) {
+	return(UdmDSTRAppend(dstr, data, strlen(data)));
+}
+
+int UdmDSTRAppendf (UDM_DSTR *dstr, const char *fmt, ...) {
+	va_list ap;
+	size_t bytes_left;
+	int nc;
+	char *tmp;
+	size_t asize;
+
+	while (1) {
+		bytes_left = dstr->size_total - dstr->size_data;
+		va_start(ap, fmt);
+		nc = vsnprintf(dstr->data + dstr->size_data, bytes_left, fmt, ap);
+		va_end(ap);
+		if (nc > -1 && nc + 1 < bytes_left) break;
+
+		if (nc < 0 || nc + 1 == bytes_left) asize = dstr->size_total + dstr->size_page;
+		else asize = dstr->size_total + ((nc - bytes_left) / dstr->size_page + 1) * dstr->size_page;
+
+		tmp = UdmRealloc(dstr->data, asize);
+		if (! tmp) {
+			nc = 0;
+			break;
+		}
+		dstr->size_total = asize;
+		dstr->data = tmp;
+	}
+
+	dstr->size_data += nc;
+	return(nc);
+}
 
 /*** str functions implementation which may be missing on some OS ***/
 
@@ -103,16 +193,17 @@ int strncasecmp(const char *s1, const char *s2, size_t n)
 }
 #endif
 
-#ifndef HAVE_STRNDUP
-char * strndup(const char * str,size_t len){
+#if !defined(HAVE_STRNDUP) || defined(EFENCE)
+char * UdmStrndup(const char * str,size_t len){
 	char * res;
-	res=(char*)malloc(len+1);
+	res=(char*)UdmMalloc(len+1);
 	strncpy(res,str,len);
 	res[len]='\0';
 	return res;
 }
 #endif
 
+
 #ifndef HAVE_STRCASESTR
 char * strcasestr(register const char *s, register const char *find) {
         register char c, sc;
@@ -136,13 +227,19 @@ char * strcasestr(register const char *s, register const char *find) {
 /**************** RFC 1522 ******************************************/
 static char base64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
-int UdmHex2Int(int h){
+int __UDMCALL UdmHex2Int(int h){
 	if((h>='0')&&(h<='9'))return(h-'0');
 	if((h>='A')&&(h<='F'))return(h-'A'+10);
 	if((h>='a')&&(h<='f'))return(h-'a'+10);
 	return(0);
 }
 
+int __UDMCALL UdmInt2Hex(int i){
+	if((i>=0)&&(i<=9))return(i+'0');
+	if((i>=10)&&(i<=15))return(i+'A'-10);
+	return '0';
+}	
+
 char * udm_rfc1522_decode(char * dst, const char *src){
 	const char *s;
 	char  *d;
@@ -237,9 +334,10 @@ char * udm_rfc1522_decode(char * dst, const char *src){
 /************* Base 64 *********************************************/
 /* BASE64 encoding converts  3x8 bits into 4x6 bits                */
 
-__C_LINK void __UDMCALL udm_base64_encode (const char *s, char *store, size_t length){
+__C_LINK size_t __UDMCALL udm_base64_encode (const char *src, char *store, size_t length){
 	size_t i;
 	unsigned char *p = (unsigned char *)store;
+	unsigned const char *s = (unsigned const char *)src;
 
 	for (i = 0; i < length; i += 3){
 		*p++ = base64[s[0] >> 2];
@@ -252,6 +350,7 @@ __C_LINK void __UDMCALL udm_base64_encode (const char *s, char *store, size_t le
 	if (i == length + 1) *(p - 1) = '=';
 	else if (i == length + 2) *(p - 1) = *(p - 2) = '=';
 	*p = '\0';
+	return p- (unsigned char*)store;
 }
 
 
@@ -441,7 +540,7 @@ char * __UDMCALL UdmEscapeURL(char *d,const char *s){
 	dd=d;
 	while(*s){
 		if((*s & 0x80) || strchr("%&<>+[](){}/?#'\"\\;,",*s)) {
-			sprintf(d,"%%%X",(int)*s);
+			sprintf(d,"%%%X",(int) (unsigned char)*s);
 			d+=2;
 		}else
 		if(*s==' '){
@@ -985,40 +1084,43 @@ void UdmTime_t2HttpStr(time_t t, char *str){
 /* Find out tz_offset for the functions above
  */
 int UdmInitTZ(void){
+
+/*
         time_t tt;
 	struct tm *tms;
 
-	tms=localtime(&tt);
+	tzset();
+	tms=localtime(&tt);*/
 	/* localtime sets the external variable timezone */
-	tz_offset=(int)timezone;
+/*	tz_offset = tms->tm_gmtoff;
 	return 0;
-#if 0
-
+*/
 #ifdef HAVE_TM_GMTOFF
-	time_t clock;
+	time_t tclock;
 	struct tm *tim;
         
-	clock=time(NULL);
-	tim=localtime(&clock);
-        tz_offset=tim->tm_gmtoff;
+	tzset();
+	tclock = time(NULL);
+	tim=localtime(&tclock);
+        tz_offset = (long)tim->tm_gmtoff;
 	return 0;
-#else                                                                           
-	struct tm *t;
+#else      
+	time_t tt;
+	struct tm t, gmt;
 	int days, hours, minutes;
 
-	tt= time(NULL);
+	tzset();
+	tt = time(NULL);
         gmt = *gmtime(&tt);
-        t = localtime(&tt);
-        days = t->tm_yday - gmt.tm_yday;
+        t = *localtime(&tt);
+        days = t.tm_yday - gmt.tm_yday;
         hours = ((days < -1 ? 24 : 1 < days ? -24 : days * 24)
-                + t->tm_hour - gmt.tm_hour);
-        minutes=hours * 60 + t->tm_min - gmt.tm_min;
-        tz_offset=minutes*60;
+                + t.tm_hour - gmt.tm_hour);
+        minutes = hours * 60 + t.tm_min - gmt.tm_min;
+        tz_offset = 60L * minutes;
 	return 0;
 #endif /* HAVE_TM_GMTOFF */
 
-#endif /* if 0 */
-
 /* ONE MORE VARIANT - how many do we have?
     struct timezone tz;
 
@@ -1353,7 +1455,7 @@ __C_LINK int __UDMCALL udm_snprintf(char *buf, size_t len, const char *fmt, ...)
 
        need = udm_vsnprintf_len(fmt, ap);
 
-       if ((tmpbuf = (char *) malloc(need + 1)) == NULL)
+       if ((tmpbuf = (char *) UdmMalloc(need + 1)) == NULL)
                strncpy(buf, "Oops! Out of memory in snprintf", len-1);
        else {
                vsprintf(tmpbuf, fmt, ap);
@@ -1371,18 +1473,6 @@ __C_LINK int __UDMCALL udm_snprintf(char *buf, size_t len, const char *fmt, ...)
 
 
 
-#ifndef HAVE_HSTRERROR
-const char *h_errlist[] = {
-    "Error 0",
-    "Unknown host",
-    "Host name lookup failure",
-    "Unknown server error",
-    "No address associated with name"
-};
-#endif
-
-
-
 
 #ifdef WIN32
 int UdmBuild(char *path, int omode) {
@@ -1577,7 +1667,7 @@ int UdmSetEnv(const char * name,const char * value){
 #ifdef HAVE_PUTENV
 	int res;
 	char * s;
-	s=(char*)malloc(strlen(name)+strlen(value)+3);
+	s=(char*)UdmMalloc(strlen(name)+strlen(value)+3);
 	sprintf(s,"%s=%s",name,value);
 	res=putenv(s);
 	UDM_FREE(s);
@@ -1706,14 +1796,17 @@ ssize_t UdmRecvall(int s, void *buf, size_t len){
 	if (len == 0) return 0;
 	return recv(s, buf, len, MSG_WAITALL);
 #else
-
 	ssize_t received = 0, r=0;
 	char *b = buf;
 	time_t start = time(NULL);
 	if (len == 0) return received;
 	while ( ((size_t)received < len) && ((r = recv(s, b+received, len-received,0)) >= 0)){
 		received+=r;
-		if (have_sigint || have_sigterm || have_sigpipe) break;
+		if (have_sigint || have_sigterm 
+#ifndef WIN32
+		    || have_sigpipe
+#endif
+		    ) break;
 		if (time(NULL) - start > 300) break;
 	}
 	return (r < 0) ? r : received;
@@ -1743,21 +1836,6 @@ ssize_t UdmSend(int s, const void *msg, size_t len, int flags) {
 
 #ifndef WIN32
 
-void (*UdmSignal(int signo, void (*handler)(int)))(int) {
-  struct sigaction act, oact;
-
-  act.sa_handler = handler;
-  sigemptyset(&act.sa_mask);
-  act.sa_flags  = 0;
-  if (signo == SIGCHLD) act.sa_flags |= SA_NOCLDSTOP;
-  act.sa_flags |= SA_RESTART;
-  if (sigaction(signo, &act, &oact) < 0) {
-    return(SIG_ERR);
-  }
-  return(oact.sa_handler);
-}
-
-/***********/
 
 static struct flock* file_lock(struct flock *ret, int type, int whence) {
     ret->l_type = type ;
@@ -1778,7 +1856,7 @@ void UdmWriteLock(int fd) {  /* an exclusive lock on an entire file */
     fcntl(fd, F_SETLKW, file_lock(&ret, F_WRLCK, SEEK_SET));
 
 #else
-	// Windows routine
+	/* Windows routine */
 	
 	DWORD filelen_high;
 	DWORD filelen_low;
@@ -1805,7 +1883,7 @@ void UdmUnLock(int fd) { /* ulock an entire file */
     fcntl(fd, F_SETLKW, file_lock(&ret, F_UNLCK, SEEK_SET));
 
 #else
-	// Windows routine
+	/* Windows routine */
 
 	DWORD filelen_high;
 	DWORD filelen_low;
@@ -1827,7 +1905,7 @@ void UdmReadLock(int fd) {   /* a shared lock on an entire file */
     fcntl(fd, F_SETLKW, file_lock(&ret, F_RDLCK, SEEK_SET));
 
 #else
-	// Windows routine
+	/* Windows routine */
 	
 	DWORD filelen_high;
 	DWORD filelen_low;
@@ -1854,7 +1932,7 @@ void UdmReadLockFILE(FILE *f) {   /* a shared lock on an entire file */
     fcntl(fileno(f), F_SETLKW, file_lock(&ret, F_RDLCK, SEEK_SET));
 
 #else
-	// Windows routine
+	/* Windows routine */
 	int file_desc = fileno(f);
 	UdmReadLock(file_desc);
 
@@ -1870,7 +1948,7 @@ void __UDMCALL UdmWriteLockFILE(FILE *f) {  /* an exclusive lock on an entire fi
     fcntl(fileno(f), F_SETLKW, file_lock(&ret, F_WRLCK, SEEK_SET));
 
 #else
-	// Windows routine
+	/* Windows routine */
 
 	int file_desc = fileno(f);
 	UdmWriteLock(file_desc);
@@ -1887,7 +1965,7 @@ void __UDMCALL UdmUnLockFILE(FILE *f) { /* ulock an entire file */
     fcntl(fileno(f), F_SETLKW, file_lock(&ret, F_UNLCK, SEEK_SET));
 
 #else
-	// Windows routine
+	/* Windows routine */
 
 	int file_desc = fileno(f);
 	UdmUnLock(file_desc);
@@ -1895,3 +1973,146 @@ void __UDMCALL UdmUnLockFILE(FILE *f) { /* ulock an entire file */
 #endif
 
 }
+
+
+#ifdef USE_PARANOIA
+
+/* NOTE: gcc optimization should be switched off */
+
+#define NSAVE 10
+/*#define PRINTF_DEBUGINFO*/
+
+#ifdef USE_SYSLOG
+extern char * __progname;
+#endif
+
+typedef struct {
+  unsigned long save[NSAVE];
+  unsigned long saveip[NSAVE];
+  unsigned invflag;
+
+#ifdef PARANOIDAL_ROOT
+  int paranoidal_check;
+#endif
+
+} UDM_PARANOIA_PARAM;
+
+static unsigned 
+getbp()
+{ 
+	__asm__("movl %ebp,%eax");
+	__asm__("movl (%eax),%eax");
+}
+
+#ifdef USE_SYSLOG
+#define SUICIDE \
+	openlog(__progname,LOG_NDELAY|LOG_PERROR|LOG_PID|LOG_CONS,LOG_USER);\
+        syslog(LOG_ERR,"Stack violation - exiting");\
+        closelog();\
+        kill(SIGSEGV,getpid());\
+        exit(1) ;
+#else
+#define SUICIDE \
+        kill(SIGSEGV,getpid());\
+        exit(1) ;
+#endif
+
+
+
+
+void * UdmViolationEnter(void) { 
+	unsigned bp = getbp();
+	int i = 0;
+	UDM_PARANOIA_PARAM *p = (UDM_PARANOIA_PARAM*)UdmMalloc(sizeof(UDM_PARANOIA_PARAM));
+
+#if defined(PRINTF_DEBUGINFO)
+	printf("\nEnter p: %lx\n", p);
+#endif
+	if (p == NULL) exit(2);
+
+	bzero(p, sizeof(UDM_PARANOIA_PARAM));
+#ifdef PARANOIDAL_ROOT
+	p->paranoidal_check = -1;
+#endif
+
+#ifdef PARANOIDAL_ROOT
+	if(p->paranoidal_check == -1) 
+	  p->paranoidal_check = issetugid() || (!geteuid()) || (!getuid());
+	if(!p->paranoidal_check) return (void*)p; 
+#endif
+
+	p->invflag++;
+	if(p->invflag > 1) return (void*)p;
+	bzero(p->save, sizeof(p->save));
+	bzero(p->saveip, sizeof(p->saveip));
+	bp = *((unsigned*)bp);
+	for(i = 0; i < NSAVE; i++) { 
+#if defined(PRINTF_DEBUGINFO)
+		printf("saving %lx (%i,%d)\n", bp, p->invflag, i);
+#endif
+		p->save[i] = bp;
+		if(!bp) break;
+		/* this is bogus, but... Sometimes we got stack entries
+		 * points to low addresses. 0x20000000 - is a shared 
+		 * library maping start */ 
+		if(bp<0x20000000) {
+			p->save[i] = 0;
+			break;
+		};
+		p->saveip[i] = *((long unsigned*)bp + 1);
+#if defined(PRINTF_DEBUGINFO)
+		printf("storing   ip %lx : %x \n", p->saveip[i], *((unsigned*)bp + 1));
+#endif
+		bp = *(long unsigned*)bp;
+	}
+	return (void *)p;
+};
+
+
+void UdmViolationExit(void *v) { 
+	unsigned bp = getbp();
+	int i = 2;
+	UDM_PARANOIA_PARAM *p = (UDM_PARANOIA_PARAM*)v;
+
+#if defined(PRINTF_DEBUGINFO)
+	printf("Exit p: %lx\n", p);
+#endif
+#ifdef PARANOIDAL_ROOT
+	/* at exit_violation we always have paranoidal_check set to 0 or 1 */
+	if(!p->paranoidal_check) {
+	  UDM_FREE(p);
+	  return; 
+	}
+#endif
+
+	if(p->invflag > 1) { 
+		p->invflag--;
+		return;
+	};
+	bp = *((unsigned*)bp);
+	for(i = 0; i < NSAVE; i++) { 
+#if defined(PRINTF_DEBUGINFO)
+		printf("processing %lx (%i, %d)\n", bp, p->invflag, i);
+#endif
+		if(!p->save[i]) break;
+		if(bp != p->save[i]) { 
+			SUICIDE;
+		};
+		if(!bp) break;
+#if defined(PRINTF_DEBUGINFO)
+		printf("restoring ip %lx : %x \n", p->saveip[i], *((unsigned*)bp + 1));
+#endif
+		if(p->saveip[i] != *((unsigned*)bp+1)) { 
+			SUICIDE;
+		};
+		bp = *(unsigned*)bp;
+	};
+	p->invflag--;
+	if (p->invflag == 0) UDM_FREE(p);
+	return;
+};
+
+
+#endif /* USE_PARANOIA */
+
+
